program fpcfix(input, output);
(* fpcfix: remove comments from a program

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1;

const
(* begin module version *)
   version = 1.06; (* of fpcfix.p 2016 Jul 20
2016 Jul 20, 1.06: testing
2016 Jul 20, 1.05: build code - functional!!!
2016 Jul 20, 1.04: build code (almost there!)
2016 Jul 20, 1.03: build code
2016 Jul 20, 1.02: build code
2016 Jul 20, 1.01: build code
2016 Jul 20, 1.00: origin from nocom - functional copy
*)
(* end module version *)

(* begin module describe.fpcfix *)
(*
name

   fpcfix: add assign to pascal programs

synopsis
   fpcfix(input: stdin; output: out)

files

   input:  a Pascal or Delila program

   output: the same program ready to be compiled by fpc

description

   The Free Pascal Compiler (fpc) compiler does not have a --transparent-file-names flag like
   gpc.  This connects the file name in the program statement to the
   local external file by the same name.  It is an essential feature
   for Delila programs since they use so many files at once.

   This program inserts

   assign(afile,'afile');

   in front of each reset(afile) and rewrite(afile).
   The fpc

see also

   {Free Pascal Compiler website:} http://www.freepascal.org/

   {Description of program that removes comments: }
   nocom.p

   {Other comment manipulation programs: }
   decom.p
   codecomments.p

   {fpcfix removes extra lines in the code.  To keep extra lines, 
   remove the line with 'NO EQUALLINES' on it.  A script to do this
   is} mkfpcfixequal { which generates the fpcfixequal.p program.}

author

   Thomas Dana Schneider

bugs

technical notes

   Some programs have comment starts inside quotes.  fpcfix
   is now smart enough to avoid changing these.

   In an earlier version of the progrm (before 2012 Nov 14) the
   comments were left as 'nibbons' that have no contents.  This allowed
   one to see exactly where the comments were.  One can remove these
   using Unix sed:

   set asterisk = '*'
   fpcfix < commented.p |\
   sed -e 's/{}//g' | \
   sed -e "s/(\$asterisk\$asterisk)//g" | \
   cat > commentless.p

   Since most of the time one doesn't want the extra empty, the
   program was changed to remove the nibbons.

*)
(* end module describe.fpcfix *)

(* trigger definitions follow to make program compilable.
The prgmod.p program has the triggers *)
(* LOCKED module filler.const *)
{ original:
      fillermax = 50; (* the size of the filler array for a string *)
}
      fillermax = 20; (* the size of the filler array for a string *)
(* LOCKED module filler.const from prgmod.p 4.20 *)

(* begin module interact.const *)
(* begin module string.const *)
      maxstring = 2000; (* the maximum string *)
(* end module string.const version = 5.54; (@ of prgmod.p 2016 Jul 20 *)
(* end module interact.const version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

type

(* begin module interact.type *)
(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      previous: stringptr; (* the previous string in a series *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 5.54; (@ of prgmod.p 2016 Jul 20 *)
(* end module interact.type version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

(* module filler.type *)
      (* the following is an array used to fill a string.
      it is convenient to have it much shorter than the maxstring, so that
      it is easy to fill the string using procedure fillstring.
      the user must declare the value of constant fillermax. *)
      filler = packed array[1..fillermax] of char;
(* module filler.type from prgmod.p 4.20 *)

(* module trigger.type *)
      trigger = record (* an object to be searched for *)
         seek: string; (* the characters looked for *)
         state: integer; (* how close to triggering we are *)
         skip: boolean; (* trigger not found- skip the line *)
         found: boolean (* the trigger was found *)
      end;
(* module trigger.type from prgmod.p 4.20 *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

(* begin module interact.clearstring *)
(* begin module clearstring *)
(* These modules clear strings in various ways *)
(* ---- *)
procedure emptystring(var ribbon: string);
(* empty the contents of the string but do NOT
remove the pointer.  This is useful for clearing
one string within a linked list of them. *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
   with ribbon do begin
      for index := 1 to maxstring do letters[index] := ' ';
      length := 0;
      current := 0;
   end
end; (* emptystring *)
(* ---- *)
procedure clearstring(var ribbon: string);
(* empty the string and remove the pointer.
As of 2010 Dec 04 there is a previous pointer that
allows proper removal of the linked list using disposestring.
This clearstring routine is for the initial clearing
of the variables. *)
begin (* clearstring *)
   with ribbon do begin
      emptystring(ribbon);
      previous := nil;
      next := nil;
   end
end; (* clearstring *)
(* ---- *)
procedure disposestring(var ribbon: stringptr);
(* remove the linked list from the string in revere order *)
var
   t: stringptr; (* a link in the string *)
begin
   t := ribbon;
   (* find the end *)
   while t^.next <> nil do t := t^.next;
   (* move backwards and dispose links *)
   while t^.previous <> nil do begin
      t := t^.previous;
      dispose(t^.next);
   end;
   dispose(t);
end; (* disposestring *)
(* ---- *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer.  This is now deprecated, do not use it
since clearstring still clears the next pointer. *)
begin (* initializestring *)
   writeln(output,'remove initializestring routine!');
   writeln(output,'replace it with clearstring routine!');
   halt; (* to force deprecation *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 5.54; (@ of prgmod.p 2016 Jul 20 *)
(* end module interact.clearstring version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

(* begin module interact.prompt *)
procedure prompt(var afile: text);
(* prompt a file.
   the prompt is sent to the output file, and
   a line is read into the pascal line buffer.
   (for the cyber system this means to readln afile.)
   guarantee no bomb *)
begin (* prompt *)
      if eof(afile) then reset(afile);
      readln(afile);
end; (* prompt *)
(* end module interact.prompt version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

(* begin module interact.readchar *)
procedure readchar(var afile:text; var ch: char);
(* read a character from afile, guarantee no bomb *)
begin (* readchar *)
      if eof(afile) then prompt(afile);
      read(afile,ch);
(*writeln(output,'"',ch,'"') *)
end; (* readchar *)
(* end module interact.readchar version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a line (as a string) from a file not using string calls.  this lets
one obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;
         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;
         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

(* begin module interact.writestring *)
(* begin module writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module writestring version = 5.54; (@ of prgmod.p 2016 Jul 20 *)
(* end module interact.writestring version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler);
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
(*   fillstring(s, 'this-is-the-string                                ');
the two comments make it easy to line the characters up. also, for this
example, it was assumed that the length of filler as defined by the
constant fillermax was 50. *)
var
      length: integer; (* of the string without trailing blanks *)
      index: integer; (* of s *)
begin (* fillstring *)
      clearstring(s);
      length := fillermax;
      while (length > 1) and (a[length] = ' ') do length := pred(length);
      if (length = 1) and (a[length] = ' ') then begin
         writeln(output, 'fillstring: the string is empty');
         halt
      end;

      for index := 1 to length do s.letters[index] := a[index];
      s.length := length;
      s.current := 1
end; (* fillstring *)
(* end module filler.fillstring version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

(* begin module filler.filltrigger *)
procedure filltrigger(var t: trigger;
                      a: filler);
(* fill the trigger t *)
begin (* filltrigger *)
         fillstring(t.seek,a)
end; (* filltrigger *)
(* end module filler.filltrigger version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

(* begin module trigger.proc *)
(* this module allows one to scan a series of characters, as from
an array or a file, and to "trigger" or detect a simple string
in the series.  the advantage of the trigger is that several triggers
can "observe" a stream of characters at once, each looking for a
different thing.
some other modules required: interact.const, interact.type *)

procedure resettrigger(var t: trigger);
(* reset the trigger to ground state *)
begin (* resettrigger *)
      with t do begin
         state := 0;
         skip := false;
         found := false
      end
end; (* resettrigger *)

procedure testfortrigger(ch: char; var t: trigger);
(* look at the character ch.
   if it is part of the trigger (at the current trigger state),
       then the trigger state goes higher.
   if it is not part of the trigger then the trigger state is reset,
      skip is true and one should skip onward to find the trigger.
   if the trigger is found, found is true.
1996 Sep 12: Bug found!  In the case of a trigger "ab", the program
used to miss it for situations like "aab".  This was because at the
first a it would step up.  Then it would see the second a and recognize
that was not part of ab.  It would fail to realize that it could be
the start of a new one.  The code now accounts for that possibility. *)
begin (* testfortrigger *)
      with t do begin
         state := succ(state);
{
writestring(list,seek);
writeln(list,'testfortrigger seek.letters[',state:1,']:',
seek.letters[state],' ch:',ch);
}
         if seek.letters[state] = ch
         then begin
            skip := false;
            if state = seek.length then found := true
                                   else found := false
         end
         else begin
            (* it failed.  But wait!  It could be the beginning
            of a NEW trigger string! *)
            if seek.letters[1] = ch then begin
               state := 1;
               skip := false;
               found := false
            end
            else begin (* reset trigger *)
               state := 0;
               skip := true;
               found := false
            end
         end
      end
end; (* testfortrigger *)
(* end module trigger.proc version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

(* begin module copystring *)
procedure copystring(a: string; var b: string);
(* copy string a to b.  The length, characters and
where the string points to are all copied. *)
var l: integer; (* index to the string *)
begin
   b.length := a.length;
   for l := 1 to a.length do b.letters[l] := a.letters[l];
   b.current := a.current;
   b.next := a.next;
end;
(* end module copystring version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

(* begin module concatstring *)
procedure concatstring(var a, b: string; var c: string);
(* concatenate a and b and put in c *)
var   abindex: integer; (* index to both strings *)
      cindex: integer; (* index to both strings *)
begin (* concatstring *)
   emptystring(c);
   c.length := a.length + b.length;
   if c.length > maxstring then begin
      writeln(output,'string too long in concatstring');
      halt
   end;

   cindex := 0;
   for abindex := 1 to a.length do begin
      cindex := succ(cindex);
      c.letters[abindex] := a.letters[abindex]
   end;

   for abindex := a.length+1 to a.length+b.length do begin
      cindex := succ(cindex);
      c.letters[abindex] := b.letters[abindex-a.length]
   end;
end; (* concatstring *)
(* end module concatstring version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

(* begin module attachstring *)
procedure attachstring(var a: string; b: string);
(* attach the string b onto the end of a *)
var
   hold: string; (* intermediate for doing the attachment *)
begin
   copystring(a, hold);
   concatstring(hold,b,a)
end;
(* end module attachstring version = 5.54; (@ of prgmod.p 2016 Jul 20 *)

procedure themain(var fin, fout: text);
var
   c: char; (* the current character just read *)
   inquote: boolean; (* true if within a quote *)
   incomment1: boolean; (* true if within a parenthesis star comment *)
   incomment2: boolean; (* true if within a brackets comment *)
   p: char; (* the character previous to c *)

   aline: string; (* a line from the file *)
   bline: string; (* copy of aline with possible modifications *)
   a, b: integer; (* indicies to aline and bline strings *)

   resetFind: trigger; (* trigger for reset *)
   rewriteFind: trigger; (* trigger for rewrite *)
   beginreset:   integer; (* beginning of a 'reset' *)
   beginrewrite: integer; (* beginning of a 'rewrite' *)
   beginresetrewrite:   integer; (* beginning of a 'reset' or 'rewrite' *)

   insert: string; (* the material to insert in front of reset
                      and rewrite *)
   acopy: integer; (* copying location from aline *)
   d: boolean; (* set true to debug *)
   docopy: boolean; (* copy the text.  If false, wait for the next ';'
                       to remove the file name *)

begin
   d := false;

(* these cases exist:
   if new > old then rewrite(mailgotlarger)
   if eof(afile) then reset(afile);
So I have to identify the rewrite/reset and then surround it
with a begin-end pair ...
*)

                         (* 12345678 *)
(*                                   1         2 *)
(*                          12345678901234567890 *)
   filltrigger(resetFind,  'reset(              ');
   filltrigger(rewriteFind,'rewrite(            ');
   fillstring(insert,      'assign(             ');
{
   fillstring(insert,      'Xssign(             ');
}
(* writeln(output,'length of insert is=',insert.length:1);halt; *)

   resettrigger(resetFind);
   resettrigger(rewriteFind);


   (* Pass 1: remove the contents of the file *)
   incomment1 := false;
   incomment2 := false;
   c := ' ';
   while not eof(fin) do begin
      inquote := false; (* quotes can only be on one line *)

      clearstring(aline);
      clearstring(bline);
      with aline do begin
         length := 0;
         while not eoln(fin) do begin
            length := succ(length);
            read(fin, aline.letters[length]);
         end;
      end;
      readln(fin);
{
writestring(output,aline);halt;
}

      b := 0;
      docopy := true;
      for a := 1 to aline.length do begin
         p := c;
         c := aline.letters[a];
         testfortrigger(c, resetFind);
         testfortrigger(c, rewriteFind);

         if not (incomment1 or incomment2)
         then if c = '''' (* This is a single quote mark in Pascal. *)
         then inquote := not inquote;

         if not inquote then begin (* skip comments inside quotes! *)
            if not incomment2 then begin
               if (p = '(') and (c = '*')
               then incomment1 := true;

               if (p = '*') and (c = ')') and incomment1
               then begin
                  incomment1 := false;
               end;
            end;
            if not incomment1 then begin
               if (c = '{')
               then incomment2 := true;
               if (c = '}') and incomment2
               then begin
                  incomment2 := false;
               end;
            end
         end;

         if not ((incomment1 or incomment2) or inquote)
         then begin
            if resetFind.found or rewriteFind.found then begin
               (* the reset begins a few characters before this point *)
               if resetFind.found
               then beginresetrewrite := a - resetFind.seek.length + 1;
               if rewriteFind.found
               then beginresetrewrite := a - rewriteFind.seek.length + 1;

{zzzqqq}
if d then begin
writeln(output, 'FOUND resetFind, beginresetrewrite = ', beginresetrewrite:1);
writestring(output,aline); writeln(output);
write(output,'1.bline:"'); writestring(output,bline); writeln(output,'"');
end;
               (* Construct "assign(afile,'afile');" *)
               (* back up b to find the insertion point: *)
{
               b := a - resetFind.seek.length + 1;
}
               b := beginresetrewrite;
               bline.length := b - 1;
               (* put the insert into b: *)
               attachstring(bline, insert);
{
               bline.length := bline.length + insert.length;
               b := bline.length + insert.length;
}

if d then begin
 write(output,'2.bline:"'); writestring(output,bline); writeln(output,'"');
{writeln(output,'length of insert is=',insert.length:1);}
end;

               (* copy the name up to ')' *)
               acopy := a+1;
               b := bline.length;
               while (aline.letters[acopy] <> ')')
                     and (acopy < maxstring)
               do begin
                  b := succ(b);
                  bline.letters[b] := aline.letters[acopy];
                  acopy := succ(acopy);
               end;

               bline.length := b+1;
if d then begin
 write(output,'3.bline:"'); writestring(output,bline); writeln(output,'"');
end;
               b := bline.length; bline.letters[b] := ',';
               b := succ(b); bline.letters[b] := '''';
               bline.length := b;
if d then begin
 write(output,'4.bline:"'); writestring(output,bline); writeln(output,'"');
end;

               (* copy the name up to ')' *)
               acopy := a+1;
               while (aline.letters[acopy] <> ')')
                     and (acopy < maxstring)
               do begin
                  b := succ(b);
                  bline.letters[b] := aline.letters[acopy];
                  acopy := succ(acopy);
               end;
               bline.length := b;
if d then begin
 write(output,'5.bline:"'); writestring(output,bline); writeln(output,'"');
end;

               b := succ(b); bline.letters[b] := ''''; bline.length := b;
               b := succ(b); bline.letters[b] := ')'; bline.length := b;
               b := succ(b); bline.letters[b] := ';'; bline.length := b;
d := true;
d := false;
if d then begin
 write(output,'6.bline:"'); writestring(output,bline); writeln(output,'"');
end;

{
}
               (* copy through the ';' *)
               acopy := beginresetrewrite;
               while (aline.letters[acopy] <> ';')
                     and (acopy < maxstring)
               do begin
                  b := succ(b);
                  bline.letters[b] := aline.letters[acopy];
                  acopy := succ(acopy);
               end;
               bline.length := b;
               b := succ(b); bline.letters[b] := ';'; bline.length := b;

if d then begin
 write(output,'7.bline:"'); writestring(output,bline); writeln(output,'"');
end;
{ clearstring(bline);}

{
               aline.length := a - resetFind.seek.length + 1;
               aline.length := acopy;
}
               docopy := false;

            end
            else begin
{if d then c := 'x';}
{
if a < acopy then
}
               if docopy then begin
                  b := succ(b);
                  bline.letters[b] := c;
                  bline.length := b;
               end
               else begin
                  if c = ';' then docopy := true;
               end
            end
         end
         else begin
{
writeln(output,'b=',b:1);
write(output,'X');
}
             b := succ(b);
             bline.letters[b] := c;
             bline.length := b;
         end
      end;
{write(output,'+');}
      writestring(output,bline);
      writeln(output);

   end;

end;

begin
  themain(input, output);
1: end.
