program histan(hst,cmp,chisq,output);
(* histan: analysis of a histogram; uses a composition to determine
      expected values and calculates chi-squared for the histogram columns.
      (this program was previously called 'achsq').
      by Gary Stormo
modules needed: delman, auxmods, delmods *)

label 1;  (* end of program *)

const
(* begin module version *)
version = 4.22;  (* of histan.p 1994 sep 5
origin: before 1985 feb 16 *)
(* end module version *)

(* begin module describe.histan *)
(*
name
      histan: histogram analysis.

synopsis
      histan(hst: in, cmp: in, chisq: out, output: out)

files
      hst: the histogram input; is the output of program hist;
      cmp: a composition input; is the output of program comp;
      chisq: the chi-squared analysis output;
      output: for user messages;

description
      histan determines the chi-squared values at each position for a
      histogram.  the observed values come from the histogram.  if a
      composition is provided the expected values come from that, otherwise
      the expected values assume equal frequencies of all bases.  the
      chisquared is calculated for each level of oligonucleotide (i.e.,
      monos, dis, tris) for which the histogram data exists.

see also
      hist.p, comp.p

author
      gary stormo

bugs
      none known
*)
(* end module describe.histan *)

(* begin module hist.const *)
      maxhistwidth = 401;
(* end module hist.const version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

type base = (a,c,g,t); (* this is declared so we do not need book.type *)
(* begin module hist.type *)
      histptr = ^histarray;
      histarray = array[1..maxhistwidth] of integer;
      histogram = record
         zero : histptr;
         mono : array[base] of histptr;
         di : array[base,base] of histptr;
         tri : array[base,base,base] of histptr;
      end;
(* end module hist.type version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

(* begin module comp.type *)
            (* the composition is stored in a tree of these nodes *)
      compnodeptr = ^compnode; (* points to a node of the tree *)
      compnode = record    (* a node of the composition tree *)
         count : integer; (* the number of oligos for this node *)
         son : array[base] of compnodeptr; (* the pointers to 'descendants'
                                           of this node of the tree *)
      end;

            (* spiders are used to make the composition tree *)
      spiderptr = ^spider; (* points to a 'spider' *)
      spider = record  (* a spider climbs the composition tree, its path
                        determined by the sequences, and increments 'count' at
                        all the nodes it passes, thereby determining the
                        composition *)
         depth : integer; (* the level of the node now at *)
         place : compnodeptr; (* a pointer to the current node *)
         next : spiderptr; (* the next spider in the collection *)
      end;

            (* the total number of composition entries at a given level
               is stored in the linked list of type comptotal *)
      compptr = ^comptotal;
      comptotal = record
         count: integer;  (* the number at a given level *)
         next: compptr;  (* pointer to the next level totals *)
      end;

           (* 'path' is used in printing the tree *)
      pathptr = ^path; (* pointer into a path on the tree *)
      path = record  (* the path of bases to get to a particular node *)
         bas : base;
         next : pathptr;
      end;
(* end module comp.type version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

var   cmp, (* output of comp program *)
      hst,  (* a histogram program output *)
      chisq: text;  (* the output of this program *)

         (* variables for the composition data *)
      root: compnodeptr;  (* of the composition tree *)
      compmax,  (* read from the composition file *)
      readmax: integer;  (* the limit to read from the composition *)
      monocomptotal: compptr; (* the beginning of the string of composition
                   totals *)

         (* variables for histogram data *)
      histo: histogram;
      histmin, (* shortest oligo in histogram *)
      histmax, (* longest oligo in histogram *)
      numseqs, (* number of sequences in histogram *)
      histbegin,  (* leftmost position in histogram *)
      histwidth, (* width of histogram *)
      i : integer;  (* an index *)

(* begin module package.primitive *)
(* ************************************************************************ *)
(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module copylines *)
function copylines(var fin, fout: text; n: integer): integer;
(* copy n lines of file fin to file fout.
the actual number of lines copied is returned. *)
var
      index: integer; (* the current line number *)
begin (* copylines *)
      index := 0;
      while (not eof(fin)) and (index < n) do begin
         copyaline(fin, fout);
         index := succ(index)
      end;

      copylines := index
end; (* copylines *)
(* end module copylines version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* ************************************************************************ *)
(* end module package.primitive version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module comp.readcomp *)

(* begin module skipoligo *)
procedure skipoligo(var thefile: text);
(* this procedure is used to skip over an oligonucleotide string.
   it reads until it comes to a base, and then continues to read until
   it comes to a blank. no checking is done for non-base characters
   in between. *)
var ch : char;
begin
      repeat read(thefile,ch)
      until (ch in ['a','c','g','t']);
      repeat read(thefile,ch)
      until (ch = ' ');
end;
(* end module skipoligo version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

procedure readcomp(var comp: text; var compmax: integer; readmax: integer;
                   var root: compnodeptr; var monocomptotal: compptr);
(* this procedure requires modules: comp.type, skipoligo, halt;
   this procedure reads from file 'comp' a composition and puts it into
   the tree pointed to by the 'root' pointer.  'compmax' is the depth
   of the composition tree which is stored.  it is the minimum of
   'readmax', the requested depth, and 'detcomp', the depth for which
   the input file composition was determined.  'monocomptotal' points
   to the beginning (for the monos) of a linked list which gives the
   totals for each level of the composition. *)
type  (* for efficient reading  of the information into the tree, each
         node that has a successor is stored in a queue as well as put
         into the tree.  the variables of the type list are the queue
         elements. *)
    listptr = ^list;
    list = record
      item: compnodeptr;  (* points to the composition tree node *)
      next: listptr;  (* points to the next list item in the queue *)
    end;

var  freeitem,  (* the list of unused list pointers *)
     listitem, (* an item in the queue *)
     first,  (* the first item in the queue *)
     last: listptr;  (* the last item in the queue *)
     comptot,  (* the comp total for a given level *)
     newcomptot: compptr;  (* for adding to the string of comptot"s *)
     detcomp,  (* the level to which the input composition was determined *)
     level,  (* of the composition being read, i.e., monos, dis, ... *)
     number: integer;  (* read from the 'comp' file *)
     ch: char;  (* for reading from 'comp' *)
     ba: base;  (* an index *)

(* getitem and clearitem provide efficient use of linked list storage by
   keeping a list of unused pointers that can be allocated instead of
   always creating 'new' ones  *)
procedure getitem(var l: listptr);
(* obtain a listitem from the free list or by making a new one *)
begin
      if freeitem<>nil
         then begin
            l:=freeitem;
            freeitem:=freeitem^.next
         end
         else new(l);
      l^.next:=nil
end;


procedure clearitem(var l: listptr);
(* return a listitem to the free list *)
var   lptr: listptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freeitem;
         freeitem:=lptr
      end
end;

begin  (* readcomp *)
      reset(comp);
      if eof(comp) then begin
         writeln(output,' error: no composition file provided');
         halt;
      end;
      readln(comp); (* skip the program identification *)
      readln(comp); (* skip the book identification *)
      readln(comp,detcomp); (* obtain the determined composition *)
      readln(comp); (* skip the blank line *)

            (* determine the level of composition to be stored *)
      if (readmax < 1) then begin
         writeln(output);
         writeln(output,' warning: 0 or negative oligo length requested');
         writeln(output,' composition used is depth ',detcomp:1);
         writeln(output);
         compmax := detcomp
      end
      else if (readmax > detcomp) then begin
         writeln(output);
         writeln(output,' warning: requested composition oligo',
                        ' length (',readmax:1,')');
         writeln(output,' is larger than the determined composition',
                        ' oligo length (',detcomp:1,').');
         writeln(output,' composition used is to depth ',detcomp:1);
         writeln(output);
         compmax := detcomp
      end
      else compmax := readmax;

      new(root);
      new(monocomptotal);
      new(comptot);
      new(first);
      new(last);
      first^.item := root;
      first^.next := nil;
      last := first;
      new(freeitem);
      freeitem^.next := nil;

         (* read in the total number of bases from the composition *)
      readln(comp); (* skip the * *)
      readln(comp); (* skip the information line *)
      readln(comp,number);
      root^.count := number;

      repeat
         read(comp,ch); (* skip the space *)
         read(comp,ch); (* the ch determines what to do next *)
         if (ch = '*') then begin  (* determine the level about to be read *)
            readln(comp);
            readln(comp,level);
            if (level = 1) then monocomptotal := comptot
            else begin
               new(newcomptot);
               comptot^.next := newcomptot;
               comptot := newcomptot;
            end;
            comptot^.count := 0;
            comptot^.next := nil
         end
         else begin  (* read the composition values *)
            for ba := a to t do begin
               skipoligo(comp);
               read(comp,number);
               if (number <> 0) then begin
                  new(first^.item^.son[ba]);
                  first^.item^.son[ba]^.count := number;
                  comptot^.count := comptot^.count + number;
                  getitem(listitem);
                  last^.next := listitem;
                  last := listitem;
                  last^.next := nil;
                  last^.item := first^.item^.son[ba]
               end
               else first^.item^.son[ba] := nil;
            end;
            clearitem(first);
            readln(comp);
         end;
      until level = compmax;
         (* for the last level of compositions to be read we don"t
            need to store the nodes in the queue because their successors
            will not be read in. *)
      repeat
         for ba := a to t do begin
            skipoligo(comp);
            read(comp,number);
            if (number <> 0) then begin
               new(first^.item^.son[ba]);
               first^.item^.son[ba]^.count := number;
               comptot^.count := comptot^.count + number;
            end
            else first^.item^.son[ba] := nil;
         end;
         clearitem(first);
         readln(comp);
      until (first = nil);
end;  (* readcomp *)

function getcount(root: compnodeptr; start: pathptr): integer;
(* this function follows the tree from the node 'root' (which
   may be the root of a subtree) along the path initiated with
   'start', and returns the count of the resulting node.  if the
   path through the tree ever hits a null node in the process
   the count is returned to be zero. *)
var place: compnodeptr;  (* a place in the composition tree *)
    point: pathptr;  (* a point in the path through the tree *)
begin
      point:= start;
      place := root;
      while ((place <> nil) and (point <> nil)) do begin
         place := place^.son[point^.bas];
         point := point^.next;
      end;
      if (place = nil) then getcount := 0
      else getcount := place^.count;
end;

(* end module comp.readcomp version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

(* begin module hist.readhist *)
procedure readhist(var hist: text; var histmin, histmax, histwidth,
                  histbegin, numseqs: integer; var histo: histogram);
(* this procedure requires modules: hist.const, hist.type, skipoligo;
   this procedure takes a file 'hist', which is the output of the hist
   program, and returns the information in that file.  the variables are:
      histmin: the shortest oligo counted in the histogram;
      histmax: the longest oligo counted in the histogram;
      histwidth: the width of the histogram;
      histbegin: the first position in the histogram;
      numseqs: the number of sequences counted in the histogram.
   the actual histogram data is returned in the variable 'histo'. *)
var b1,b2,b3 : base;
      i : integer;
begin
            (* get the histogram parameters *)
      reset(hist);
      readln(hist);
      readln(hist);
      readln(hist,histmin);
      readln(hist,histmax);
      readln(hist,numseqs);
      readln(hist,histwidth);
      readln(hist);
      readln(hist,histbegin);
      readln(hist);

{zzz}
writeln(output,' saw parameters: ');
writeln(output,histmin:1);
writeln(output,histmax:1);
writeln(output,numseqs:1);
writeln(output,histwidth:1);
writeln(output,histbegin:1);

            (* get the histogram data from histmin to histmax *)
      if (histmin = 0) then begin
         new(histo.zero);
         for i := 1 to histwidth do read(hist,histo.zero^[i])
      end
      else histo.zero := nil;

      if (histmin <= 1) and (histmax >= 1) then
         for b1 := a to t do begin
            skipoligo(hist);
            new(histo.mono[b1]);
            for i := 1 to histwidth do read(hist,histo.mono[b1]^[i])
         end
      else for b1 := a to t do histo.mono[b1] := nil;

      if ((histmin <= 2) and (histmax >= 2)) then
         for b1 := a to t do
            for b2 := a to t do begin
               skipoligo(hist);
               new(histo.di[b1,b2]);
               for i := 1 to histwidth do read(hist,histo.di[b1,b2]^[i])
            end
      else for b1 := a to t do
         for b2 := a to t do histo.di[b1,b2] := nil;

      if (histmax >= 3) then
         for b1 := a to t do
            for b2 := a to t do
               for b3 := a to t do begin
                  skipoligo(hist);
                  new(histo.tri[b1,b2,b3]);
                  for i := 1 to histwidth do read(hist,histo.tri[b1,b2,b3]^[i])
               end
      else for b1 := a to t do
         for b2 := a to t do
            for b3 := a to t do histo.tri[b1,b2,b3] := nil;
end;

(* end module hist.readhist version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

procedure header;
(* write the header to the chisq file *)
begin
      writeln(chisq,'* histan version ',version:4:2,
            '; chi-squared histogram analysis using composition');
      if not eof(hst) then begin
         if copylines(hst,chisq,2)<>2 then begin
            writeln(output,' hst file too short');
            halt
         end
      end
      else begin
         writeln(output,'empty histogram file');
         halt;
      end;
      if not eof(cmp) then begin
         if copylines(cmp,chisq,2)<>2 then begin
            writeln(output,' cmp file too short');
            halt
         end;
      end
      else begin
         writeln(chisq,'* no composition provided');
         writeln(chisq,'* all oligos assumed equally frequent;');
      end;

      writeln(chisq,'*');
      write(chisq,'* position');
      write(chisq,'     mono');
      write(chisq,'        di');
      writeln(chisq,'       tri');
      writeln(chisq,'*');
end;

function expectedmono(ba :base; pos : integer) : real;
(* calculate the expected ocurrences of the mononucleotide ba *)
var  b1 : base;
      columnsum : integer; (* sum of a histogram column *)
      fractionexpected: real; (* bases on the composition *)
      start: pathptr; (* the start of the path through the tree *)
begin
      if (histmin = 0) then columnsum := histo.zero^[pos]
      else begin
         columnsum := 0;
         for b1 := a to t do columnsum := columnsum + histo.mono[b1]^[pos];
      end;
      new(start);
      start^.bas := ba;
      start^.next := nil;
      if (monocomptotal = nil) then fractionexpected := 1/4
      else fractionexpected := getcount(root,start) / monocomptotal^.count;
      expectedmono := columnsum * fractionexpected;
end;

function expecteddi(ba1,ba2 :base; pos : integer) : real;
(* calculate the expected occurences of the dinucleotide ba1,ba2 *)
var  b1,b2 : base;
      columnsum : integer;  (* sum of histogram column *)
      fractionexpected: real; (* based on composition *)
      point,  (* in the path through the tree *)
      start: pathptr; (* of the path through the tree *)
begin
      columnsum := 0;
      for b1 := a to t do
         for b2 := a to t do
            columnsum := columnsum + histo.di[b1,b2]^[pos];
      new(start);
      start^.bas := ba1;
      new(point);
      point^.bas := ba2;
      start^.next := point;
      point^.next := nil;
      if (monocomptotal = nil) then fractionexpected := 1/16
      else fractionexpected := getcount(root,start)/monocomptotal^.next^.count;
      expecteddi := columnsum * fractionexpected;;
end;

function expectedtri(ba1,ba2,ba3 : base; pos : integer) : real;
(* calculate the expected occurences of the trinucleotide ba1,ba2,ba3 *)
var b1,b2,b3 :base;
      columnsum : integer; (* sum of histogram column *)
      fractionexpected: real;  (* based on composition *)
      point,  (* in the path through the tree *)
      start: pathptr; (* of the path through the tree *)

begin
      columnsum := 0;
      for b1 := a to t do
         for b2 := a to t do
            for b3 := a to t do
               columnsum := columnsum + histo.tri[b1,b2,b3]^[pos];

      new(start);
      start^.bas := ba1;
      new(point);
      point^.bas := ba2;
      start^.next := point;
      new(point);
      point^.bas := ba3;
      point^.next := nil;
      start^.next^.next := point;
      if (monocomptotal = nil) then fractionexpected := 1/64
      else fractionexpected := getcount(root,start) /
               monocomptotal^.next^.next^.count;
      expectedtri := columnsum * fractionexpected;
end;

procedure chisquared(pos : integer); (* calculate chi squareds for mono,di,tri
                                 at position pos and printout *)
var   b1,b2,b3 : base;
      obs : integer; (* observed number of occurences in histogram *)
      exp, (* expected number of occurences from composition *)
      chi : real;
begin
      write(chisq,(pos + histbegin -1):8); (* convert from internal coordinates
                                            to histogram coordinates *)
      if (histmin <= 1) then begin
         chi := 0;
         for b1 := a to t do begin
            obs := histo.mono[b1]^[pos];
            exp := expectedmono(b1,pos);
            if (exp > 0) then
                  chi := chi + (obs - exp)*(obs - exp)/exp
         end;
         write(chisq,chi:10:2);
      end
      else write(chisq,'          ');

      if ((histmin <= 2) and (histmax >= 2) and (compmax >= 2)) then begin
         chi := 0;
         for b1 := a to t do
            for b2 := a to t do begin
               obs := histo.di[b1,b2]^[pos];
               exp := expecteddi(b1,b2,pos);
               if (exp > 0) then
                     chi := chi + (obs - exp)*(obs - exp)/exp
            end;
         write(chisq,chi:10:2)
      end
      else write(chisq,'          ');

      if ((histmax >= 3) and (compmax >= 3)) then begin
         chi := 0;
         for b1 := a to t do
            for b2 := a to t do
               for b3 := a to t do begin
                  obs := histo.tri[b1,b2,b3]^[pos];
                  exp := expectedtri(b1,b2,b3,pos);
                  if (exp > 0) then
                        chi := chi + (obs - exp)*(obs - exp)/exp
               end;
         writeln(chisq,chi:10:2)
      end
      else writeln(chisq);
end;

begin  (* histan *)
      writeln(output,'histan ',version:4:2);
      rewrite(chisq);
      reset(cmp);
      reset(hst);
      header;
         (* read the histogram data into variable histo *)
      readhist(hst,histmin,histmax,histwidth,histbegin,numseqs,histo);
         (* read the composition data into the tree starting at root *)
      if (histmax > 3) then readmax := 3
      else readmax := histmax;
      if eof(cmp) then monocomptotal := nil
      else readcomp(cmp,compmax,readmax,root,monocomptotal);
      for i := 1 to histwidth do chisquared(i);
1: end.
