program instal(xbook, ybook, shlist, inst, rinst, sinst, list, output);
(* instal: delila instruction alignment
   thomas schneider

module libraries required: delman, delmods, prgmods, auxmods
*)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.49; (* of instal.p 1995 July 26
origin: 1984 feb 11 *)
(* end module version *)

(* begin module describe.instal *)
(*
name
      instal: delila instruction alignment

synopsis
      instal(xbook: in, ybook: in, shlist: in, inst: in,
             rinst: out, sinst: out, list: out, output: out)

files
      xbook: a delila system book containing one piece used to
         align the ybook pieces.
      ybook: a delila system book containing pieces to be aligned by
         the piece in xbook.
      shlist: the output of the sorth program.  these sorted helixes
         must have been generated using helix(xbook,ybook,hlist,...) and
         then sorth(hlist,shlist,...,[.../1/...]).  that is, sorth must
         have been used to select only the top 1 helix from hlist.
      inst: the instructions used to generate ybook (or a comparable set
         of instructions that correspond to the ones for ybook).
      rinst: reduced instructions: those instructions from inst that
         have a unique helix in shlist.  in other words, inst is copied to
         rinst only for instructions that have a unique alignment.
         the other instructions are also copied, but surrounded by
         comment delimiters to neutralize them.  neutrialized
         instructions are followed by a delila instruction that will
         maintain the original piece numbers.
      sinst: shifted instructions: the instructions written to rinst are
         realigned by the helixes of shlist.  the new alignment
         is the coordinate where the 5 prime end of the xpiece
         would lie on each y piece.
      list: progress of the realignemnt.
      output: messages to the user.

description
      the purpose of instal is to automatically realign a set of instructions.
      for example, if one has a set of instructions that define the
      initiation codon of some procaryotic ribosome binding sites,
      one may want another alignment by the shine and dalgarno.  to
      do this, the following steps are needed:
      1. the instructions (inst) are converted to a book (ybook) using
      delila.  instructions that define the 3 prime end of the 16s rrna are
      written and used to create xbook.
      2.  potential helixes between xbook and ybook are found with the
      helix program, making an hlist.
      3. the strongest helixes of the hlist are selected using the
      sorth program.  thus each piece of ybook has a unique (or no)
      helix associated with it.
      4. instal is used to alter the original instructions.  instructions
      (pieces) with no unique helix are neutralized by putting them in
      comments to in both rinst and sinst.  in addition, the instructions
      of sinst are 'shifted' so they are aligned by the 5 prime end of xpiece.

see also
      delila.p, helix.p, sorth.p

author
      thomas dana schneider

bugs
      none known

technical notes
      the largest shift that is recorded is specified by the
      constant absshift

*)
(* end module describe.instal *)

(* begin module instal.const *)
      absshift = 50; (* the largest shift (+ or -) that can be
         recorded by the program *)
      fillermax = 3; (* the size of the filler array for a string *)
(* end module instal.const *)

(* begin module interact.const *)
      maxstring = 150; (* the maximum string *)
(* end module interact.const version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module book.const *)
(* constants needed for book manipulations *)

      dnamax = 3000;  (* length of dna arrays *)
      namelength = 20; (* maximum key name length *)
      linelength = 80; (* maximum line readable in book *)

(* end module book.const version = 'delmod 6.51 85 apr 17 tds/gds' *)

type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module filler.type *)
      (* the following is an array used to fill a string.
      it is convenient to have it much shorter than the maxstring, so that
      it is easy to fill the string using procedure fillstring.
      the user must declare the value of constant fillermax. *)
      filler = packed array[1..fillermax] of char;
(* end module filler.type version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module trigger.type *)
      trigger = record (* an object to be searched for *)
         seek: string; (* the characters looked for *)
         state: integer; (* how close to triggering we are *)
         skip: boolean; (* trigger not found- skip the line *)
         found: boolean (* the trigger was found *)
      end;
(* end module trigger.type version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module book.type *)
(* types needed for book manipulations *)

      chset = set of 'a'..'z';

      (* types defined in book definition *)

      alpha = packed array[1..namelength] of char; (* this is not alfa *)

      (* name is a left justified string with blanks following the
         characters *)
      name = record
         letters: alpha;
         length: 0..namelength (* zero means an unspecified structure *)
      end;

      lineptr = ^line;
      line = record (* a line of characters *)
         letters: packed array [1..linelength] of char;
         length: 0..linelength;
         next: lineptr
      end;

      direction = (plus, minus, dircomplement, dirhomologous);
      configuration = (linear, circular);
      state = (on, off);


      header = record (* header of key *)
         keynam: name; (* key name of structure *)
         fulnam: lineptr; (* full name of structure *)
         note:   lineptr (* note key *)
      end;


      (* base types *)
      base = (a,c,g,t);
      dnaptr = ^dnastring;
      dnarange = 0..dnamax;
      seq = packed array[1..dnamax] of base;
      dnastring = record
         part: seq;
         length: dnarange;
         next: dnaptr
      end;


      orgkey = record (* organism key *)
         hea: header;
         mapunit: lineptr (* genetic map units *)
      end;

      chrkey = record (* chromosome key *)
         hea: header;
         mapbeg: real; (* number of genetic map beginning *)
         mapend: real (* number of genetic map ending *)
      end;

      pieceptr = ^piece;
      piekey = record (* piece key *)
         hea: header;
         mapbeg: real; (* genetic map beginning *)
         coocon: configuration; (* configruation (circular/linear) *)
         coodir: direction; (* direction (+/-) relative to genetic map *)
         coobeg: integer; (* beginning nucleotide *)
         cooend: integer; (* ending nucleotide *)
         piecon: configuration; (* configruation (circular/linear) *)
         piedir: direction; (* direction (+/-) relative to coordinates *)
         piebeg: integer; (* beginning nucleotide *)
         pieend: integer; (* ending nucleotide *)
      end;


      piece = record
         key: piekey;
         dna: dnaptr
      end;

      reference = record
         pienam : name;  (* name of piece referred to *)
         mapbeg : real;  (* genetic map beginning *)
         refdir : direction;  (* direction relative to coordinates *)
         refbeg : integer;  (* beginning nucleotide *)
         refend : integer;  (* ending nucleotide *)
      end;

      genkey = record  (* gene key *)
         hea : header;
         ref : reference;
      end;

      trakey = record  (* transcript key *)
         hea : header;
         ref : reference;
      end;

      markerptr = ^marker;
      markey = record (* marker key *)
         hea : header;
         ref : reference;
         sta : state;
         phenotype : lineptr;
         next : markerptr;
      end;

      marker = record
         key : markey;
         dna : dnaptr;
      end;

(* end module book.type version = 'delmod 6.51 85 apr 17 tds/gds' *)

var
(* begin module instal.var *)
      xbook, ybook, shlist, inst, rinst, sinst, list: text;
(* end module instal.var *)

(* begin module book.var *)
(* ************************************************************************ *)
(* global variables needed for book manipulations *)

      (* free storage: *)
      freeline: lineptr; (* unused lines *)
      freedna: dnaptr; (* unused dnas *)

      readnumber: boolean; (* whether to read a number from the notes, or
                              to read in the notes *)
      number: integer; (* the number of the item just read *)
      numbered: boolean; (* true when the item just read is numbered *)
      skipunnum: boolean; (* a control variable to allow skipping of
                             un-numbered items in the book *)

(* ************************************************************************ *)
(* end module book.var version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module package.primitive *)
(* ************************************************************************ *)
(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module copylines *)
function copylines(var fin, fout: text; n: integer): integer;
(* copy n lines of file fin to file fout.
the actual number of lines copied is returned. *)
var
      index: integer; (* the current line number *)
begin (* copylines *)
      index := 0;
      while (not eof(fin)) and (index < n) do begin
         copyaline(fin, fout);
         index := succ(index)
      end;

      copylines := index
end; (* copylines *)
(* end module copylines version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* ************************************************************************ *)
(* end module package.primitive version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler);
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
(*   fillstring(s, 'this-is-the-string                                ');
the two comments make it easy to line the characters up. also, for this
example, it was assumed that the length of filler as defined by the
constant fillermax was 50. *)
var
      length: integer; (* of the string without trailing blanks *)
      index: integer; (* of s *)
begin (* fillstring *)
      clearstring(s);
      length := fillermax;
      while (length > 1) and (a[length] = ' ') do length := pred(length);
      if (length = 1) and (a[length] = ' ') then begin
         writeln(output, 'fillstring: the string is empty');
         halt
      end;

      for index := 1 to length do s.letters[index] := a[index];
      s.length := length;
      s.current := 1
end; (* fillstring *)
(* end module filler.fillstring version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module trigger.proc *)
(* this module allows one to scan a series of characters, as from
an array or a file, and to "trigger" or detect a simple string
in the series.  the advantage of the trigger is that several triggers
can "observe" a stream of characters at once, each looking for a
different thing.
some other modules required: interact.const, interact.type *)

procedure resettrigger(var t: trigger);
(* reset the trigger to ground state *)
begin (* resettrigger *)
      with t do begin
         state := 0;
         skip := false;
         found := false
      end
end; (* resettrigger *)

procedure testfortrigger(ch: char; var t: trigger);
(* look at the character ch.
   if it is part of the trigger (at the current trigger state),
       then the trigger state goes higher.
   if it is not part of the trigger then the trigger state is reset,
      skip is true and one should skip onward to find the trigger.
   if the trigger is found, found is true. *)
begin (* testfortrigger *)
      with t do begin
         state := succ(state);
(*         if debugging then begin
            writestring(list,seek);
            writeln(list,'testfortrigger seek.letters[',state:1,']:',
                           seek.letters[state],' ch:',ch);
         end;*)
         if seek.letters[state] = ch
         then begin
            skip := false;
            if state = seek.length then found := true
                                   else found := false
         end
         else begin (* reset trigger *)
            state := 0;
            skip := true;
            found := false
         end
      end
end; (* testfortrigger *)
(* end module trigger.proc version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module package.getpiece *)
(* ************************************************************************ *)
(* begin module package.brpiece *)
(* ************************************************************************ *)
(* begin module book.basis *)
(* procedures needed for book manipulations *)

(* get procedures should be used for all linked lists of records *)

procedure getline(var l: lineptr);
(* obtain a line from the free line list or by making a new one *)
begin
      if freeline<>nil
         then begin
            l:=freeline;
            freeline:=freeline^.next
         end
         else new(l);
      l^.length:=0;
      l^.next:=nil
end;

procedure getdna(var l: dnaptr);
begin
      if freedna<>nil
         then begin
            l:=freedna;
            freedna:=freedna^.next
         end
         else new(l);
      l^.length:=0;
      l^.next:=nil
end;

(* clear procedures should be called each time the records are no longer needed
         failure to do this may result in a stack overflow. *)

procedure clearline(var l: lineptr);
(* return a line to the free line list *)
var   lptr: lineptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freeline;
         freeline:=lptr
      end
end;

procedure cleardna(var l: dnaptr);
var   lptr: dnaptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freedna;
         freedna:=lptr
      end
end;

procedure clearheader(var h: header);
(* clear the header h (remove lines to free storage) *)
begin
      with h do begin
         clearline(fulnam);
         while note<>nil do clearline(note)
      end
end;

procedure clearpiece(var p: pieceptr);
(* clear the dna of the piece *)
begin
      while p^.dna<>nil do cleardna(p^.dna);
      clearheader(p^.key.hea)
end;

function chartobase(ch:char):base;
(* convert a character into a base *)
begin
      case ch of
         'a': chartobase:=a;
         'c': chartobase:=c;
         'g': chartobase:=g;
         't': chartobase:=t
      end
end;

function basetochar(ba:base):char;
(* convert a base into a character *)
begin
      case ba of
         a: basetochar:='a';
         c: basetochar:='c';
         g: basetochar:='g';
         t: basetochar:='t';
      end
end;

function complement(ba:base):base;
(* take the complement of ba *)
begin
      case ba of
         a: complement:=t;
         c: complement:=g;
         g: complement:=c;
         t: complement:=a;
      end
end;

function pietoint(p: integer; pie: pieceptr): integer;
(* p is a coordinate on the piece.
   we want to transform p into a number
   from 1 to n: an internal coordinate system for
   easy manipulation of piece coordinates *)
var   i: integer; (* an intermediate value *)
begin
      with pie^.key do begin
         case piedir of
            plus:  if p>=piebeg
                   then i:=p-piebeg+1
                   else i:=(p-coobeg)+(cooend-piebeg)+2;
            minus: if p<=piebeg
                   then i:=piebeg-p+1
                   else i:=(cooend-p)+(piebeg-coobeg)+2
         end;
         pietoint:=i
      end
end;

function inttopie(i: integer; pie: pieceptr):integer;
(* i is in the range 1 to some maximum.  it is an internal coordinate
   system for the program.  we want to do a
   coordinate transformation to obtain
   a value in the range of the piece called pie:
   i=1 corresponds to piebeg and
   i=its maximum corresponds to pieend *)
var   p: integer; (* an intermediate value *)
begin
      with pie^.key do begin
         case piedir of
            plus: begin
               p:=piebeg+(i-1);
               if p>cooend
                  then if coocon=circular
                     then p:=p-(cooend-coobeg+1)
            end;
            minus: begin
               p:=piebeg-(i-1);
               if p<coobeg
                  then if coocon=circular
                     then p:=p+(cooend-coobeg+1)
            end
         end;
         inttopie:=p
      end
end;

function piecelength(pie: pieceptr): integer;
(* return the length of the dna in pie *)
begin
      piecelength:=pietoint(pie^.key.pieend,pie)
end;

(* end module book.basis version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.getto *)
function getto(var thefile: text; ch: chset): char;
(* search the file for a character in the first line which is a
   member of the set ch. *)
var achar: char;
begin
      achar:=' ';
      while (not(achar in ch)) and (not eof(thefile))
         do readln(thefile,achar);
      if (achar in ch) then getto:=achar
                       else getto:=' '
end;
(* end module book.getto version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.skipstar *)
procedure skipstar(var thefile:text);
(* skip start of line (or star = '*'). *)
begin (* skipstar *)
      if thefile^ <> '*' then begin
         writeln(output,' procedure skipstar: bad book');
         writeln(output,' "*" expected as first character on the line, but "',
                        thefile^,'" was found');
         halt
      end;
      get(thefile); (* skip the star *)

      if thefile^ <> ' ' then begin
         writeln(output,' procedure skipstar: bad book');
         writeln(output,' "* " expected on a line but "*',
                        thefile^,'" was found');
         halt
      end;
      get(thefile) (* skip the blank *)
end; (* skipstar *)
(* end module book.skipstar version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brreanum *)
procedure brreanum(var thefile: text; var reanum: real);
(* read a real number from the file *)
begin
      skipstar(thefile);
      readln(thefile,reanum);
end;
(* end module book.brreanum version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brnumber *)
procedure brnumber(var thefile: text; var num: integer);
(* read a number from the file *)
begin
      skipstar(thefile);
      readln(thefile,num)
end;
(* end module book.brnumber version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brname *)
procedure brname(var thefile: text; var nam: name);
(* read a name from the file *)
var   i: integer; (* an index to the name *)
      c: char; (* a character read *)
begin (* brname *)
      skipstar(thefile);
      with nam do begin
         length:=0;
         repeat
            length:=succ(length);
            read(thefile,c);
            letters[length] := c
         until (eoln(thefile)) or
               (length>=namelength) or
               (letters[length]=' ');
         if letters[length]=' ' then length:=length-1;
         if length<namelength
            then for i:=length+1 to namelength do letters[i]:=' '
      end;
      readln(thefile)
end; (* brname *)
(* end module book.brname version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brline *)
procedure brline(var thefile: text; var l: lineptr);
(* read a line from the file *)
var
      i,j: integer;
      acharacter: char;
begin
      skipstar(thefile);
      i:=0;
      while (not eoln(thefile)) do begin
         i:=succ(i);
         read(thefile,acharacter);
         l^.letters[i]:=acharacter
      end;
      if i<l^.length then for j:=i+1 to l^.length do l^.letters[j]:=' ';
      l^.length:=i;
      l^.next:=nil;
      readln(thefile)
end;
(* end module book.brline version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brdirect *)
procedure brdirect(var thefile: text; var direct: direction);
(* read a direction *)
var   ch: char;
begin
      skipstar(thefile);
      readln(thefile,ch);
      if ch='+' then direct:=plus
                else direct:=minus
end;
(* end module book.brdirect version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brconfig *)
procedure brconfig(var thefile: text; var config: configuration);
(* read a configuration *)
var   ch: char;
begin
      skipstar(thefile);
      readln(thefile,ch);
      if ch='l' then config:=linear
                else config:=circular
end;
(* end module book.brconfig version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brnotenumber *)
procedure brnotenumber(var thefile: text; var note: lineptr);
(* book note reading to obtain the number of the object.
the procedure returns the value of the number as a global.
(this is not such a good practice, but we are stuck with it for now.) *)
begin (* brnotenumber *)
      note:=nil;
      numbered := false;
      number := 0; (* force number to zero if there
         is no number at all *)
      (* the next character is n or * depending on whether there are notes *)
      if thefile^ = 'n' then begin
         readln(thefile);
         if thefile^ <> 'n' then begin
            skipstar(thefile);
            if not eoln(thefile) then begin
               if thefile^ = '#' then begin
                  numbered := true;
                  get(thefile); (* move past the number symbol *)
                  read(thefile,number);
               end
            end;
            repeat
               readln(thefile)
            until thefile^ = 'n';
            readln(thefile)
         end
         else readln(thefile)
      end
end; (* brnotenumber *)
(* end module book.brnotenumber version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brnote *)
procedure brnote(var thefile: text; var note: lineptr);
(* read note key *)
var
      newnote: lineptr; (* the new note *)
      previousnote: lineptr; (* the last line of the notes *)
begin (* brnote *)
      note:=nil;
      if thefile^ = 'n' then begin (* enter note *)
         readln(thefile);
         if thefile^ <> 'n' then begin (* abort null note (n/n) *)
            getline(note);
            newnote:=note;
            while thefile^ <> 'n' do begin (* wait until end of note *)
               brline(thefile,newnote);
               previousnote:=newnote;
               (* get next note *)
               getline(newnote^.next);
               newnote:=newnote^.next;
            end;
            (* last note was not used, so: *)
            clearline(newnote);
            previousnote^.next:=nil;
            readln(thefile)
         end
         else readln(thefile)
      end
end; (* brnote *)
(* end module book.brnote version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brheader *)
procedure brheader(var thefile: text; var hea: header);
(* read the header of a key. *)
begin
      with hea do begin
         (* read key name *)
         brname(thefile,keynam);

         (* read full name *)
         getline(fulnam);
         brline(thefile,fulnam);

         (* read note key *)
         if readnumber then brnotenumber(thefile,note)
                       else brnote(thefile,note)
      end
end;
(* end module book.brheader version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brpiekey *)
procedure brpiekey(var thefile: text; var pie: piekey);
(* read piece key *)
begin
      with pie do begin
         brheader(thefile,hea);
         brreanum(thefile,mapbeg);
         brconfig(thefile,coocon);
         brdirect(thefile,coodir);
         brnumber(thefile,coobeg);
         brnumber(thefile,cooend);
         brconfig(thefile,piecon);
         brdirect(thefile,piedir);
         brnumber(thefile,piebeg);
         brnumber(thefile,pieend);
      end
end;
(* end module book.brpiekey version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brdna *)
procedure brdna(var thefile: text; var dna: dnaptr);
(* read in dna from thefile *)
(* note: if the dna were circularized, by linking the last dnastring
to the first, then the cleardna routine could not clear properly,
and would loop forever... there is no reason to do that, since a simple
mod function will allow one to access the circle. *)
var
      ch: char;
      workdna: dnaptr;
begin
      getdna(dna);
      workdna:=dna;
      ch:=getto(thefile,['d']);
      read(thefile,ch); (* skipstar *)
      while (ch = '*') do
      begin
         read(thefile,ch); (* skip blank *)
         repeat
            read(thefile,ch);
            if ch in ['a','c','g','t'] then begin
               if workdna^.length=dnamax then begin
                  getdna(workdna^.next);
                  workdna:=workdna^.next
               end;
               workdna^.length:=succ(workdna^.length);
               workdna^.part[workdna^.length]:=chartobase(ch)
            end
         until eoln(thefile);
         readln(thefile); (* go to next line *)
         read(thefile,ch); (* ch is either '*' or 'd' *)
      end;
      readln(thefile)
end;
(* end module book.brdna version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brpiece *)
procedure brpiece(var thefile: text; var pie: pieceptr);
(* read in a piece *)
begin
      brpiekey(thefile,pie^.key);
      if numbered or (not skipunnum)
         then brdna(thefile,pie^.dna)
end;
(* end module book.brpiece version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brinit *)
procedure brinit(var book: text);
(* check that the book is ok to read, and
   set up the global variables for br routines *)
begin (* brinit *)
      (* halt if the book is bad (first word is 'halt') or the first
         character is not * *)
      reset(book);
      if not eof(book) then begin
         (* check for the date line *)
         if book^ <> '*' then begin
            if book^ <> 'h'
            then writeln(output, ' this is not the first line of a book:')
            else writeln(output, ' bad book:');
            write(output, ' ');

            while not (eoln(book) or eof(book)) do begin
               write(output, book^);
               get(book)
            end;
            writeln(output);
            halt
         end
      end
      else begin
         writeln(output, ' book is empty');
         halt
      end;

      (* initialize free storage *)
      freeline:=nil;
      freedna:=nil;

      readnumber:=true; (* usually we read in numbers for items *)
      number:=0; (* arbitrary value *)
      numbered:=false; (* the piece has no number (none yet read in) *)
      skipunnum:=false;
end; (* brinit *)
(* end module book.brinit version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* ************************************************************************ *)
(* end module package.brpiece version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.getpiece *)
procedure getpiece(var thefile: text; var pie: pieceptr);
(* move to and read in the next piece in the book *)
var   ch: char;
begin
      ch:=getto(thefile,['p']); (* get to the next p(iece) in the book *)
      if ch<>' ' then begin
         brpiece(thefile,pie);
         ch:=getto(thefile,['p']); (* read past closing p *)
      end
end;
(* end module book.getpiece version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* ************************************************************************ *)
(* end module package.getpiece version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module findcolon *)
procedure findcolon(var thefile: text);
(* move the file to the characters just past ': ' (colon, space) *)
var
      found: boolean;
procedure die;
begin writeln(output,' no helix list data'); halt end;
begin
      found:=false;
      while not found do begin
         if eof(thefile)
            then die
            else if thefile^ = ':'
               then begin
                  get(thefile);
                  if eof(thefile)
                     then die
                     else if thefile^ = ' '
                        then begin
                           get(thefile);
                           found:=true
                        end
               end
               else get(thefile)
      end
end;
(* end module findcolon version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

(* begin module gethelix *)
procedure gethelix(var hlist: text;
                   var x,y: integer;
                   var length: integer;
                       getenergy: boolean;
                   var energy: real;
                   var done: boolean);
(* get the (x,y,length) info on some helix listed in hlist.
if the x,y pair changes, then done is true and no values are
returned.  checknames may be used at this point to start the
next pair.  if getenergy is true, then the routine expects that there
will be an energy on the helix line *)
begin (* gethelix *)
      if eof(hlist)
         then done:=true
         else begin
            get(hlist); (* skip the blank *)
            if hlist^ <> ' '
               then done:=true
               else begin
                  done:=false;
                  findcolon(hlist); read(hlist,x);
                  findcolon(hlist); read(hlist,y);
                  findcolon(hlist); read(hlist,length);
                  if getenergy then read(hlist, energy)
                               else energy := 0;
                  readln(hlist)
               end
         end
end; (* gethelix *)
(* end module gethelix version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

(* ************************************************************************ *)

(* begin module instal.readc *)
procedure readc(var inst, rinst, sinst: text;
                var c: char; copying: boolean);
(* read a character c from inst, write to rinst and sinst
if copying is true *)
begin
      if eof(inst) then begin
         writeln(output,'unexpected end of inst file');
         halt
      end;
      if eoln(inst) then begin
         readln(inst);

         (* always copy carriage returns *)
         writeln(rinst);
         writeln(sinst)
      end
      else begin
         read(inst, c);
         if copying then begin
            write(rinst, c);
            write(sinst, c)
         end
      end
end;
(* end module instal.readc *)

(* begin module instal.token *)
procedure token(var inst, rinst, sinst: text);
(* move past next token of inst, copying to rinst and sinst.
a token is any number of blanks and end of lines followed by
neither of these and then either of these.  eg:
         six tokens on this line *)
var
      c: char; (* for reading *)
begin
      repeat (* search for end of blanks *)
         readc(inst, rinst, sinst, c, true)
      until not ((c=' ') or eoln(inst));

      repeat (* copy the token out *)
         readc(inst, rinst, sinst, c, true)
      until (c =' ') or eoln(inst)
end;
(* end module instal.token *)

(* begin module instal.getalignment *)
procedure getalignment(var inst, rinst, sinst: text;
                       var alignment: integer);
(* copy inst to rinst and sinst and obtain the alignment of the get:
      get from 56 -10 to 56 +15;
alignment is 56.  assume we are just past the get, or the '-10'. *)
begin
      token(inst, rinst, sinst); (* go past 'from' or 'to' *)
      read(inst, alignment);
end;
(* end module instal.getalignment *)

(* begin module instal.copytoget *)
procedure copytoget(var inst, rinst, sinst: text;
                        begcom, endcom, getinst: trigger);
(* copy the inst file to rinst and sinst up to (but not including)
       a get command.
skip the contents of any comments (defined by begcom and endcom,
the beginning and end of comments).  search for the word
"get" defined by the trigger getinst. *)
var
      c: char; (* a character in inst *)
      level, l: integer; (* the number of characters into the
         trigger we are currently, and an index used to copy from the
         trigger out if it turns out that we don"t trigger *)
      isnoteoln: boolean; (* is the inst file not at end of line.  if so, then
         we will not want to copy the previous character to rinst and
         sinst as we normally do, since the readc routine does that *)
begin (* copytoget *)
      resettrigger(begcom);
      resettrigger(getinst);
      level := 0;
      while not getinst.found do begin
(*write(rinst,'<copytoget: after while>');*)
         isnoteoln := not eoln(inst);
         readc(inst,rinst,sinst,c,false);
         testfortrigger(c, getinst);

         (* see if we have detected the get *)
         if getinst.state = 0 then begin
            if level <> 0 (* discharge "stored" characters *)
            then with getinst.seek do for l := 1 to level do begin
                  write(rinst,letters[l]);
                  write(sinst,letters[l]);
            end;
            if isnoteoln then begin
               write(rinst,c);
               write(sinst,c);
            end
         end;
         (* keep track of the previous level of the trigger *)
         level := getinst.state;

         if not getinst.found then begin
            testfortrigger(c,begcom);
            if begcom.found then begin
               (* comment located, skip it *)
               resettrigger(endcom);
               while not endcom.found do begin
(*write(rinst,'<copytoget: looking for endcom>'); *)
                  if eof(inst) then begin
                     writeln(output,'no end to a comment in inst');
                     halt
                  end;
                  readc(inst,rinst,sinst,c,true);
                  testfortrigger(c,endcom)
               end;
               resettrigger(begcom)
            end
         end
      end
(*;write(rinst,'<<<<<<out of copytoget>>>>>>'); *)
end; (* copytoget *)
(* end module instal.copytoget *)

(* begin module instal.themain *)
procedure themain(var xbook, ybook, shlist, inst, rinst, sinst, list: text);
(* the main procedure of instal *)
var
      begcom, (* beginning of a comment *)
      endcom, (* ending of a comment *)
      getinst: (* a get instruction *)
         trigger;
      xpiece, ypiece: pieceptr; (* a piece from each of xbook and ybook *)
      alignment, (* the alignment number of a get instruction *)
      secondalignment, (* the second alignment of a get *)
      newalignment: (* the new alignment in sinst *)
         integer;

      (* variables for gethelix *)
      x, y: integer; (* location in external coordinates *)
      length: integer; (* length of helix *)
      energy: real; (* a dummy for the energy of the helix *)
      done: boolean; (* when there are no more helixes in the set *)

      sets: integer; (* count of the helix sets.  this is used to
         change the numbering instructions so that if an instruction
         is skipped (by commenting it out) the pieces will still have
         the same numbering as they did before *)
      c: char; (* a character for reading inst and writing sinst and rinst *)

      (* variables for conversion and recording information about the shifts *)
      shift: integer; (* the current shift in alignment *)
         (* the distribution of shift values is stored: *)
      distribution: array[-absshift..+absshift] of integer;
      minshift, maxshift: integer; (* the minimum and maximum shifts
         used.  these make the distribution display at the end easy *)
      shiftedsets: integer; (* the total number of shifted instructions *)
      frequency: real; (* the frequency of one shift state *)
      information: real; (* the information associated with each possible
         shift 'state' *)
      ln2: real; (* the natural logarithm of 2, used to speed the
         calculation of the information *)
      totalinformation: real; (* the sum of the information over all
         possible states *)
procedure conversion;
(* convert the alignment to the newalignment, using the
variables of themain.  (the procedure is located here for easy
reference to those variables).  the actors are:
      alignment: the alignment value in the get
      x: the 5 prime end of the helix on xpiece
      y: the 5 prime end of the helix on ypiece
      length: the length of the helix
      newalignment: the resulting new alignment value for the get of sinst
         this is the 5 prime end of xpiece as it would lie on ypiece
      shift: how many bases the alignment was shifted
*)
var
      internal: integer; (* the internal coordinate location of
         the new alignment *)
begin (* conversion *)
      internal :=
         (pietoint(y,ypiece) + length - 1) + (pietoint(x,xpiece) - 1);
      shift := internal -  pietoint(alignment, ypiece);
      newalignment := inttopie(internal, ypiece);

      (* record the shift in the distribution *)
      if abs(shift) <= absshift
      then begin
         distribution[shift] := distribution[shift] + 1;
         if shift < minshift then minshift := shift;
         if shift > maxshift then maxshift := shift;
      end
      else begin
         writeln(output, ' shift outside range -maxshift to +maxshift:');
         writeln(output,' ',shift:1,' is outside ',
            -maxshift:1,' to ',+maxshift:1)
      end;

      (* some lines of code for debugging:
      writeln(list,'------------------------');
      writeln(list,' new internal coordinate is ',internal:1,
               ' range of ypiece is 1 to ',piecelength(ypiece):1);
      writeln(list,' pietoint(y,ypiece) = ',pietoint(y,ypiece):1);
      writeln(list,' pietoint(x,xpiece) = ',pietoint(x,xpiece):1);
      writeln(list,' length = ',length:1);
*)
end; (* conversion *)
procedure initialize;
(* set up everything *)
var
      shift: integer; (* used to clear the distribution array *)
begin (* initialize *)
      writeln(output,'instal ',version:4:2);
      reset(shlist);
      reset(inst);
      rewrite(rinst);
      rewrite(sinst);
      rewrite(list);

      fillstring(begcom.seek,'(* ');
      fillstring(endcom.seek,'*) ');
      fillstring(getinst.seek,'get');

      (* the following bit of code is useful for testing the
      instruction reading procedures. *)
(* * )
      while not eof(inst) do begin
         copytoget(inst,rinst,sinst, begcom,endcom,getinst);
         write(rinst,'get');
         write(sinst,'get');
         getalignment(inst, rinst, sinst, alignment);
         write(rinst,alignment:1);
         write(sinst,(alignment+1):1);
         writeln(list,alignment:1);
      end;
( * *)

      (* prepare the xbook and xpiece *)
      brinit(xbook);
      new(xpiece);
      getpiece(xbook,xpiece);

      (* prepare the ybook and ypiece *)
      brinit(ybook);
      new(ypiece);

      (* start up the list file and skip the header of shlist *)
      writeln(list,' instal ',version:4:2, ' instruction alignment');
      writeln(list);
      writeln(list,' shlist: ');
      if copylines(shlist,list,7) <> 7 then begin
         writeln(output,'incomplete header of shlist');
         halt
      end;
      readln(shlist); (* skip first set header *)

      (* copy the first line of inst (hopefully the title) to list *)
      writeln(list);
      writeln(list,' inst: ');
      write(list,' ');
      if copylines(inst,list,1) <> 1 then begin
         writeln(output,'empty inst file');
         halt
      end;
      reset(inst); (* just to be safe *)
      writeln(list);

      sets := 0;
      for shift := -maxshift to +maxshift
         do distribution[shift] := 0;
      minshift := maxint;
      maxshift := -maxint;
      shiftedsets := 0;
end; (* initialize *)
begin (* themain *)
      initialize;
      while not eof(shlist) do begin
         gethelix(shlist,x,y,length,false,energy,done);
         sets := sets + 1;
         write(list,' set ',sets:4,' ');

         if done
         then begin
            if not eof(shlist) then begin (* skip the instructions *)
               write(list,'skip');

               (* comment out next instructions in sinst and rinst *)
               copytoget(inst,rinst,sinst,begcom,endcom,getinst);
               write(sinst,'(* get');
               write(rinst,'(* get');
               c := ' ';
               while c <> ';'
                  do readc(inst, rinst, sinst, c, true);
               writeln(sinst,' *)');
               writeln(rinst,' *)');

               (* advance the numbering *)
               writeln(sinst,'default numbering ',(sets+1):1,';');
               writeln(rinst,'default numbering ',(sets+1):1,';');

               readln(shlist); (* skip header of next set *)

               (* skip the next ypiece *)
               getpiece(ybook,ypiece);
               clearpiece(ypiece);
            end
         end
         else begin (* realign the instructions *)
            write(list,'realign');
            shiftedsets := shiftedsets + 1;

            (* get the inst alignment *)
            copytoget(inst,rinst,sinst,begcom,endcom,getinst);
            write(rinst,'get'); (* fill in missing get *)
            write(sinst,'get');
            getalignment(inst,rinst,sinst,alignment);

            (* regenerate rinst like inst *)
            write(rinst,alignment:1);

            (* get ypiece to allow conversion *)
            getpiece(ybook,ypiece);

            (* convert and shift *)
            conversion;
            write(sinst,newalignment:1);

            (* get second alignment *)
            token(inst,rinst,sinst); (* skip the '-10' *)
            getalignment(inst,rinst,sinst,secondalignment);
            if alignment <> secondalignment then begin
               writeln(output,'instruction ',sets:1,
                  ' from and to alignments are not identical');
               halt
            end;

            (* complete the output files *)
            write(rinst, alignment:1);
            write(sinst, newalignment:1);
            write(list,' ', alignment:6,' to ',newalignment:6);
            write(list,' shift is ',shift:3);

            (* clear the y piece to avoid memory overflow *)
            clearpiece(ypiece);

            (* skip header of next set *)
            gethelix(shlist,x,y,length,false,energy,done);
            if done
            then begin
               if not eof(shlist) then readln(shlist)
            end
            else begin
               writeln(output,' there are at least two helixes',
                  ' in set ',sets:1,', use the program sorth');
               writeln(output,' to take only the top helix of',
                  ' each set');
               halt
            end

         end;
         writeln(list)
      end;

      (* close up by copying the rest of the instructions out *)
      while not eof(inst)
         do readc(inst,rinst,sinst,c,true);

      (* check if there was another xpiece *)
      clearpiece(xpiece);
      getpiece(xbook,xpiece);
      if not eof(xbook)
      then writeln(output,'warning: only the first piece',
         ' of xbook was used');

      (* show the distribution *)
      totalinformation := 0;
      ln2 := ln(2); (* for speed this is precalculated *)
      writeln(list);
      writeln(output,' ',shiftedsets:1,' instructions shifted',
         ' out of ',sets:1,' instructions (sets)');
      writeln(list,' ',shiftedsets:1,' instructions shifted',
         ' out of ',sets:1,' instructions (sets)');
      writeln(list);
      writeln(list,' the distribution of shifts is: ');
      writeln(list,' shift  number  frequency information',
         ' total information');
      for shift := minshift to maxshift  do begin
         write(list,' ',shift:5,'  ',distribution[shift]:6);
         frequency := distribution[shift]/shiftedsets;

         if (frequency <= 0.0) or (frequency >= 1.0)
         then information := 0.0
         else information := -frequency * ln(frequency)/ln2;

         totalinformation := totalinformation + information;
         writeln(list,' ',frequency:10:5, '  ',information:10:5,
               ' ',totalinformation:10:5)
      end
end; (* themain *)
(* end module instal.themain *)

begin (* instal *)
      themain(xbook, ybook, shlist, inst, rinst, sinst, list);
1: end. (* instal *)
