program lenin(lengths, leninp, finst, linst, output);
(* lenin: convert a list of lengths into Delila instructions

  Tom Schneider
  Frederick, Maryland  21701-1013
  toms@alum.mit.edu

  1990
 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.24; (* of lenin.p 1990 August 21
origin 1990 August 15 *)
(* end module version *)

(* begin module describe.lenin *)
(*
name
   lenin: convert a list of lengths into Delila instructions

synopsis
   lenin(lengths: in, leninp: in, finst: out, linst: out, output: out)

files
   lengths: The olength or slength file from dbinst.
      The file is expected to contain comment lines that start with '*'.
      These are followed by columns of LENGTH, FIRST-POSITION,
      LAST-POSITION and PIECE-NAME.
   leninp: parameters to control the program.
       First line: FROM and TO for the finst instructions
       Second line: FROM and TO for the linst instructions
   finst:  Delila instructions constructed from the lengths file
      according to the parameter file, and the FIRST-POSITION of the object
      or space.
   linst:  Delila instructions constructed from the lengths file
      according to the parameter file, and the LAST-POSITION of the object
      or space.
   output: messages to the user

description
   The program allows one to make a set of instructions that correspond
   to the ends of objects that exist in the GenBank entries.
   Dbinst does not do this; and it is easier to do it this way.

   For the finst file, the Delila instructions created are of the form:

   piece PIECE-NAME;
   get from FIRST-POSTION FROM.first to FIRST-POSITION TO.first;

   while for the linst file, the Delila instructions created are of the form:

   piece PIECE-NAME;
   get from LAST-POSTION FROM.last to LAST-POSITION TO.last;

examples

documentation

see also
   dbinst.p

author
   Thomas Dana Schneider

bugs
   Title of finst is not correct.  To correct, don't have dbinst put
the title, and have lenin construct it.

technical notes

*)
(* end module describe.lenin *)

var
   lengths, leninp, finst, linst: text; (* files used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module lenin.gettitle *)
procedure gettitle(var lengths, fout: text);
(* copy the first line of lengths to fout,
surrounded by a Delila title instruction.  The routine looks for:

* dbinst stuff with * as the first character
** marks start of delila instructions
* delila instruction package
** marks end of delila instructions
* other lines to skip

 *)
var
   a,b: char; (* first two characters on a line *)
   found: boolean; (* begin or end of delila instruction packet was found *)
begin
   reset(lengths);
   (* locate the start of the Delila text in the length file *)
   found := false;
   while not found do begin
      if eof(lengths) then begin
         writeln(output,'delila instructions not found in lengths file!');
         halt
      end;
      readln(lengths,a,b);
      if a <> '*' then begin
         writeln(output,'While locating Delila instructions');
         writeln(output,'Missing * in lengths file!  Found "',a,'"');
         halt
      end;
      if b = '*' then found := true
   end;

   (* now copy the instructions *)
   found := false;
   while not found do begin
     read(lengths,a,b);
      if a <> '*' then begin
         writeln(output,'While copying Delila instructions');
         writeln(output,'Missing * in lengths file!  Found "',a,'"');
         halt
      end;
      if b = '*'
      then begin
         found := true;
         readln(lengths);
      end
      else begin
         while not eoln(lengths) do begin
            fout^ := lengths^;
            put(fout);
            get(lengths)
         end;
         readln(lengths);
         writeln(fout);
      end;
   end;

   (* skip the other non-data lines *)
(*
   while lengths^ = '*' do readln(lengths);
*)
end;
(* end module lenin.gettitle *)

(* begin module lenin.copytitle *)
procedure copytitle(var fin, fout: text);
(* copy the first line of fin to fout,
surrounded by a Delila title instruction. *)
begin
   write(fout,'title "');
   while not eoln(fin) do begin
      fout^ := fin^;
      put(fout);
      get(fin)
   end;
   writeln(fout,'";');
   readln(fin);
end;
(* end module lenin.copytitle *)

(* begin module copytotwo *)
procedure copytotwo(var fin, fout1, fout2: text);
(* copy from fin to both fout1 and fout2.  readln is provided for fin,
but no writelns for the fouts. *)
var c: char; (* a character to copy *)
begin
   while not eoln(fin) do begin
      read(lengths,c);
      write(fout1,c);
      write(fout2,c);
   end;
   readln(fin);
end;
(* end module copytotwo *)

(* begin module lenin.sign *)
procedure sign(var thefile: text; number: integer);
(* put a negative sign out or a positive sign *)
begin (* sign *)
   if number >= 0 then write(thefile,'+',number:1)
                  else write(thefile,number:1)
end; (* sign *)
(* end module lenin.sign *)

(* begin module lenin.themain *)
procedure themain(var lengths, leninp, finst, linst: text);
(* the main procedure of the program *)
var
   c: char; (* a character in the piece name *)
   count: integer; (* count of the instructions written so far *)
   firstposition: integer; (* the first position of the object or space *)
   lastposition: integer; (* the last position of the object or space *)
   length: integer; (* the length of the object or space *)
   thefromfirst: integer; (* the FROM base, for the first end of the object *)
   thetofirst: integer; (* the TO base, for the first end of the object *)
   thefromlast: integer; (* the FROM base, for the last end of the object *)
   thetolast: integer; (* the TO base, for the last end of the object *)
begin
   writeln(output,'lenin ',version:4:2);

   reset(lengths);
   reset(leninp);
   rewrite(finst);
   writeln(finst,'(* lenin ',version:4:2,' *)');
   rewrite(linst);
   writeln(linst,'(* lenin ',version:4:2,' *)');

   gettitle(lengths,finst);
   gettitle(lengths,linst);

   readln(leninp,thefromfirst,thetofirst);
   readln(leninp,thefromlast,thetolast);
   writeln(output,'First: From ',thefromfirst:1,' to ',thetofirst:1);
   writeln(output,' Last: From ',thefromlast:1,' to ',thetolast:1);
(*
zzz FINISH PUTTING IN THE NEw variables thefromlast etc......
make it count instructions and put the number out also
*)

   count := 0;
   while not eof(lengths) do begin
      (* handle change of species *)
      while lengths^='*' do begin
         read(lengths,c);
         if lengths^='*' then begin
            (* must be a new species, write the instruction out: *)
            read(lengths,c); (* get past the '*' *)
            copytotwo(lengths,finst,linst);
            writeln(finst);
            writeln(linst);
         end
         else readln(lengths) (* skip the line *)
      end;

      read(lengths,length,firstposition,lastposition);

      writeln(finst);
      writeln(linst);
      write(finst,'piece ');
      write(linst,'piece ');

      (* copy the piece name to both files *)
      copytotwo(lengths,finst,linst);
      writeln(finst,';');
      writeln(linst,';');

      (* write the get instructions *)
      count := count + 1;
      write(finst,'get from',
                    ' ',firstposition:1,' ');
      sign(finst,thefromfirst);
      write(finst,
                    ' to',
                    ' ',firstposition:1,' ');
      sign(finst,thetofirst);
      writeln(finst,'; (* ',count:1,' *)');

      write(linst,'get from',
                    ' ',lastposition:1,' ');
      sign(linst,thefromlast);
      write(linst,
                    ' to',
                    ' ',lastposition:1,' ');
      sign(linst,thetolast);
      writeln(linst,'; (* ',count:1,' *)');
   end;
end;
(* end module lenin.themain *)

begin
   themain(lengths, leninp, finst, linst);
1: end.
