program lidel(fin, lidelp, inst, output);
(* lidel: convert coordinate list to Delila instructions

  Tom Schneider
  network address: toms@alum.mit.edu


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.30; (* of lidel.p 1995 July 14
origin 1995 Jan 31 *)
(* end module version *)

(* begin module describe.lidel *)
(*
name
   lidel: convert coordinate list to Delila instructions

synopsis
   lidel(fin: in, lidelp: in, inst: out, output: out)

files
   fin:  An input file of coordinates.  Lines beginning with "*" are ignored by
       being copied to the inst as comments.  Two columns are required:  the
       piece name and the coordinate.

   lidelp:  parameters to control the program.
      first line: 2 integers, range to use.
      second line: piecol: column number of piece name
      third line: datacol: column number of coordinate data
      fourth line: title of book
      fifth line: organism name
      sixth line: chromosome name
      seventh line: symmetry of instructions.

	 If the first character on the line is 'e' then even symmetry is used
	 and the center of the symmetry will lie between bases 0 and 1.

	    If the first character is 'e' then there are two possible ways that
	    the numbers in fin could be defined:

	       If the second character on the line is 'l' then the number in
	       fin is to the left (5') of the symmetry center.

	       If the second character on the line is 'r' then the number in
	       fin is to the right (3') of the symmetry center.

	 If the first character on the line is 'o' then odd symmetry is used
	 and the center of the symmetry will lie at base 0.

      eighth line: number of instructions per integer in fin.
         1: only one instruction is written.
         2: two instructions are written: same as 1 and the complement
            at the point of symmetry.

   inst:  Delila instructions which correspond to the locations in fin.

   output: messages to the user

description
   
   Frequently one has a list of coordinates that need to be converted
   to Delila instructions.

examples

   An example lidelp file:

-700 700   fromrange torange
1          column number of piece name
3          column number of coordinate data
CDS
Sacc.cerevisiae
Sacc.cerevisiae
er         symmetry (e)ven or (o)dd, (l)eft or (r)ight
2          number of instructions to write to fin

parameters for lidel.p

documentation

see also

author
   Thomas Dana Schneider

bugs

   The coordinates given are for a location (0) and the complement (at 1)
this assumes that the point of symmetry is at 0.5; this must be generalized.

technical notes

*)
(* end module describe.lidel *)

(* begin module interact.const *)
      maxstring = 150; (* the maximum string *)
(* end module interact.const version = 4.15; (@ of prgmod.p 1994 November 12 *)

type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type version = 4.15; (@ of prgmod.p 1994 November 12 *)

var
   fin, (* file used by this program *)
   lidelp, (* file used by this program *)
   inst: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 4.15; (@ of prgmod.p 1994 November 12 *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring version = 4.15; (@ of prgmod.p 1994 November 12 *)

(* begin module interact.writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module interact.writestring version = 4.15; (@ of prgmod.p 1994 November 12 *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a string from a file not using string calls.  this lets one
obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;

         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;

         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 4.15; (@ of prgmod.p 1994 November 12 *)

(* begin module equalstring *)
function equalstring(a, b: string): boolean;
(* test for equality between two strings at current positions *)
var   index: integer; (* index to both strings *)
      equal: boolean; (* are letters in a and b the same? *)
begin (* equalstring *)
      if a.length = b.length
      then begin
         index := 1;
         repeat
            equal := (a.letters[index] = b.letters[index]);
            index := succ(index)
         until (not equal) or (index > a.length);
         equalstring := equal
      end
      else equalstring := false
end; (* equalstring *)
(* end module equalstring version = 4.15; (@ of prgmod.p 1994 November 12 *)

(* begin module copystring *)
procedure copystring(a: string; var b: string);
(* copy string a to b *)
var l: integer; (* index to the string *)
begin
   b.length := a.length;
   for l := 1 to a.length do b.letters[l] := a.letters[l]
end;
(* end module copystring version = 4.15; (@ of prgmod.p 1994 November 12 *)

(* begin module skipblanks *)
procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
      while (thefile^ = ' ') and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
      while (thefile^ <> ' ') and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 4.15; (@ of prgmod.p 1994 November 12 *)

(* begin module grabtoken *)
procedure grabtoken(var thefile: text; var thestring: string);
(* skip any blanks and then grab the next token from the file *)
var
   c: char; (* a character in thefile *)
   done: boolean; (* done finding the name *)
begin
   skipblanks(thefile);
   done := false;
   with thestring do begin
      length := 0;
      while not done do begin
         if eoln(thefile)
         then done := true
         else begin
            read(thefile,c);
            if c = ' '
            then done := true
            else begin
               length := succ(length);
               letters[length] := c;
            end
         end
      end
   end
end;
(* end module grabtoken version = 4.15; (@ of prgmod.p 1994 November 12 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 4.15; (@ of prgmod.p 1994 November 12 *)

(* begin module ncopyaline *)
procedure ncopyaline(var fin, fout: text);
(* no return: copy a line from file fin to file fout without carriage return *)
begin (* ncopyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
end; (* ncopyaline *)
(* end module ncopyaline *)

(* begin module lidel.sign *)
procedure sign(var fout: text; int: integer);
(* write the interger int to the file fout with
the sign + or - *)
begin
   if int < 0
   then write(fout,' ',int:1)
   else write(fout,' +',int:1)
end;
(* end module lidel.sign *)

(* begin module lidel.themain *)
procedure themain(var fin, lidelp, inst: text);
(* the main procedure of the program *)
var
   b: string; (* the name of a piece *)
   bold: string; (* the previous name of a piece *)
   datacol: integer; (* column number of coordinate data *)
   duplex: integer; (* 1: one way instructions, 2: two way instructions *)
   fromrange, torange: integer; (* range for instructions *)
   index: integer; (* index to find columns *)
   l: integer; (* coordinate just beyond junction *)
   piecol: integer; (* column number of piece names *)
   side: char; (* when symmetry = e, the side (l or r) of the fin coordinate *)
   symmetry: char; (* the symmetry of the created instructions, e or o *)
procedure dopiecename(var fin: text);
(* do the piece name *)
begin
   grabtoken(fin, b);
   if not equalstring(b, bold) then begin
      writeln(inst);
      write(inst,'piece ');
      writestring(inst, b);
      writeln(inst,';');
      copystring(b, bold);
   end
end;

procedure makeinst(var inst: text; l:integer; fromfirst: boolean);
(* make the instructions gotten around position l.
If fromfirst is true then use the fromrange first *)
begin
   write(inst,'get from ', l:1);
   if fromfirst then sign(inst,fromrange)
                else sign(inst,torange);
   write(inst,' to ',l:1);
   if fromfirst then sign(inst,torange)
                else sign(inst,fromrange);
   if fromfirst then write(inst,' direction +')
                else write(inst,' direction -');
   writeln(inst,';');
end;
begin
   writeln(output,'lidel ',version:4:2);

   reset(lidelp);
   readln(lidelp, fromrange, torange);
   readln(lidelp, piecol);
   readln(lidelp, datacol);
   if piecol = datacol then begin
      writeln(output,'piece name column can not equal coordinate column');
      halt
   end;
   if (piecol < 1) or (datacol < 1) then begin
      writeln(output,'columns must be postive integers');
      halt
   end;

   reset(fin);
   rewrite(inst);
   write(inst,'title "');
   ncopyaline(lidelp,inst);
   writeln(inst,'";');

   writeln(inst,'(* lidel ',version:4:2,' *)');

   write  (inst,'organism ');
   ncopyaline(lidelp,inst);
   writeln(inst,';');

   write  (inst,'chromosome ');
   ncopyaline(lidelp,inst);
   writeln(inst,';');

   readln(lidelp, symmetry, side);
   if not (symmetry in ['e','o']) then begin
      writeln(output,'symmetry must be e (even) or o (odd)');
      halt
   end;
   if symmetry = 'e' then if not (side in ['l','r']) then begin
      writeln(output,'side of fin coordinates must be l (left) or r (right)');
      halt
   end;

   readln(lidelp, duplex);
   if (duplex < 1) or (duplex > 2) then begin
      writeln(output,'You can have only 1 or 2 instructions per integer');
      halt
   end;

   writeln(inst,'default numbering piece;');
   writeln(inst,'default numbering 1;');
   writeln(inst,'default out-of-range reduce-range;');
{   writeln(inst,'default coordinate zero;');}

   clearstring(b);
   clearstring(bold);
   while not eof(fin) do begin
      if fin^='*' then begin
         write(inst,'(* ');
         ncopyaline(fin,inst);
         writeln(inst,' *)');
      end
      else begin

         if piecol < datacol then begin
            for index := 1 to piecol-1 do skipcolumn(fin);
            dopiecename(fin);
            for index := piecol+1 to datacol-1 do skipcolumn(fin);
            readln(fin,l);
         end
         else begin
            for index := 1 to datacol-1 do skipcolumn(fin);
            read(fin,l);
            for index := datacol+1 to piecol-1 do skipcolumn(fin);
            dopiecename(fin);
            readln(fin);
         end;

         if duplex = 2 then writeln(inst);

         case symmetry of
            'e': begin (* even symmetry *)
               case side of
                  'l': begin (* left side coordinate *)
                     makeinst(inst, l+1, true)
                  end;
                  'r': begin (* right side coordinate *)
                     makeinst(inst, l-1, true)
                  end;
               end;
            end;
            'o': begin (* odd symmetry *)
               makeinst(inst, l, true)
            end;
         end;

         if duplex=2 then begin
            makeinst(inst, l, false)
         end;

      end
   end
end;
(* end module lidel.themain *)

begin
   themain(fin, lidelp, inst);
1: end.
