program linreg(input,output);
(* linear regression
   thomas schneider
*)

label 1; (* end of program *)

const
(* begin module version *)
version = 2.00; (* of linreg.p 1998 feb 7
last previous change: 1985 dec 19
origin 1985 dec 12 *)
(* end module version *)

(* begin module describe.linreg *)
(*
name
      linreg: linear regression

synopsis
      linreg(input: in, output: out)

files
      input:
          first line is which pair of columns to correlate as x then y
          remaining lines are the data in columns, ending with end of file.
      output: regression results

description
      linear regression is performed between the indicated data columns.

author
      thomas schneider

bugs
      none known

*)
(* end module describe.linreg *)
         
(* begin module linear.regression *)  
(* linear regression procedure  
      1980 august 5 
      by thomas schneider 
      copyright 1980 *) 
  
procedure regress ( control: char; (* described below *)  
                    x, y: real;    (* the data pairs *) 
                var sumx, sumy, sumxsqd, sumysqd, sumxy, (* internal records *) 
(* mean, variance and covariance of x and y: *) 
                    ex, ey, varx, vary, covxy,
(* correlation coefficient: *)
                    r,
(* m = slope, b = y intercept as in y = mx + b *) 
                    m, b  
                        : real; 
(* number of data pairs entered: *) 
                var n: integer);
(* linear regression for variables x and y. 
   the control variable has only three acceptable states: 
      c clear all variables to zero 
        (note that r is set to 2.0, so that a program can check whether 
         or not the results have been calculated yet. the recommended test
         is not for equality, since round-off could cause problems.  use  
              if r > 1.5 then (no calculations yet) 
                         else (calculations done)    )
      e enter the x and y values into the internal sums 
      r results are calculated from the sums
*)
begin 
      if not( control in ['c','e','r']) then begin  
         writeln(output,' linear regression control variable value, "',control, 
                        '" is not acceptable.');
         writeln(output,' it must be in ["c","e","r"]');  
         halt 
      end 
  
      else case control of  
         'c': begin (* clear *) 
            x:=0.0; y:=0.0; 
            sumx:=0.0; sumy:=0.0; 
            sumxsqd:=0.0; sumysqd:=0.0; 
            sumxy:=0.0; 
            ex:=0.0; ey:=0.0; 
            varx:=0.0; vary:=0.0; 
            covxy:=0.0; 
            r:=2.0; (* this (impossible) value can act as a flag that 
                       shows that the calculations have not been made. *) 
  
            m:=0.0; b:=0.0; 
            n:=0  
         end; 
  
         'e': begin (* enter data *)
            sumx:=sumx + x; 
            sumy:=sumy + y; 
            sumxsqd:=sumxsqd + x*x; 
            sumysqd:=sumysqd + y*y; 
            sumxy:=sumxy + x*y; 
            n:=succ(n)
         end; 
  
         'r': begin (* calculate results *) 
            ex:=sumx/n; 
            ey:=sumy/n; 
  
            varx:=(sumxsqd/n) - ex*ex;
            vary:=(sumysqd/n) - ey*ey;
  
            covxy:=(sumxy/n) - ex*ey; 
  
            r:=covxy/(sqrt(varx * vary)); 
  
            m:=covxy/varx;
            b:=ey - m*ex
         end
      end 
end;  
(* end module linear.regression *)  

(* begin module themain *)
procedure themain(var infile, outfile: text);
(* the main procedure of the program, reading from infile for data
and producing results to outfile *)
var
      colx,coly: integer; (* column x and column y to regress with *)
      datum: real; (* an item from the data *)
      column: integer; (* the column of the datum *)
      satisfied: boolean; (* true when we have data for all regressions
                             on the current line *)
      xsatisfied: boolean; (* we have the x *)
      ysatisfied: boolean; (* we have the y *)
 
      (* variables for the regress procedure: *)
      x, y: real;    (* the data pairs *) 
          (* internal records of regress: *) 
      sumx, sumy, sumxsqd, sumysqd, sumxy,
          (* mean, variance and covariance of x and y: *) 
      ex, ey, varx, vary, covxy,
          (* correlation coefficient: *)
      r,
          (* m = slope, b = y intercept as in y = mx + b *) 
      m, b: real; 
          (* number of data pairs entered: *) 
      n: integer;
begin
      writeln(output,'linreg ',version:4:2);
      reset(infile);
      readln(infile,colx,coly);
      regress('c',x,y,sumx,sumy,sumxsqd,sumysqd,sumxy,
              ex,ey,varx,vary,covxy,r,m,b,n);

      while not eof(infile) do begin
         (* read in the column data *)
         column := 0;
         satisfied := false;
         xsatisfied := false;
         ysatisfied := false;

         while not satisfied do begin
            read(infile, datum);
            column := column + 1; 
            if column = colx then begin
               x := datum;
               xsatisfied := true
            end;
            if column = coly then begin
               y := datum;
               ysatisfied := true
            end;
            satisfied := xsatisfied and ysatisfied
         end;
         readln(infile);

         regress('e',x,y,sumx,sumy,sumxsqd,sumysqd,sumxy,
                 ex,ey,varx,vary,covxy,r,m,b,n);
      end;

      regress('r',x,y,sumx,sumy,sumxsqd,sumysqd,sumxy,
              ex,ey,varx,vary,covxy,r,m,b,n);

      (* produce the results *)
      writeln(outfile,'columns  x=',colx:5,' and y=',coly:5);
      writeln(outfile,'mean     x=',ex:5:3,' and y=',ey:5:3);
      writeln(outfile,'variance x=',varx:5:3,' and y=',vary:5:3);
      writeln(outfile,'covariance = ',covxy:5:3);
      writeln(outfile,'r = ',r:5:3);
      writeln(outfile,'m = ',m:5:3);
      writeln(outfile,'b = ',b:5:3);
end;
(* end module themain *)

begin
      themain(input,output);
1: end.
