program matmod(encseq, output);
(* mathematics modules
   thomas schneider
*)

label 1; (* end of the program *)

const
(* begin module version *)
version = 'matmod 2.08 2016sep20 tds/gds';
(*   2016 sep 20, 2.08: remove thethe duplicates
     2007 Mar 21, 2.07: add simpson routine
     1994 Sep  5, 2.06: functional *)
(* origin around 1980 *)
(* end module version *)

(* begin module describe.matmod *)
(*
name
      matmod: mathematics modules

synopsis
      matmod(encseq: in, output: out)

files
      encseq: empty or the output of the encode program for testing parameter
         reading routines.
      output: the version of matmod is printed.  successful compilation
         and running of the program indicates that the modules are correct.

description
      Self contained modules for mathematical manipulation.
      Included is a procedure for linear regression analysis of data pairs,
      a random number generator, newton's method to find roots of functions,
      and routines for reading the parameters from the encode program.

see also
      delmod.p, module.p, encode.p

author
      thomas d. schneider and gary d. stormo

bugs
      none known

technical notes

      The constant n in procedure randomtest determines how many times
      the random number generator will be in a series of tests.  If n
      is small, the test will be poor, if it is large then the test
      may take a long time.

*)
(* end module describe.matmod *)

(* begin module matmod.const *)
      debugging = false; (* for debugging purposes *)
(* end module matmod.const *)

(* begin module vector.const *)
      maxvecpart = 64; (* the number of elements in a 'part' of a vector *)
      vpagewidth = 64; (* the width (in characters) of the output vector
         file from procedure writevector *)
(* end module vector.const *)


(* begin module package.encode *)
(* *********************************************************************** *)
(* begin module encode.readencpar *)
(* end module encode.readencpar *)

(* begin module encode.functions *)
(* end module encode.functions *)

(* begin module encode.evprint *)
(* end module encode.evprint *)

(* *********************************************************************** *)
(* end module package.encode *)

type
(* begin module encode.info *)
(* information about the encode modules

definitions

      vector.  a set of integers (elements) that represent the encoding
of one sequence.  a vector is made up of one or more parameter vectors.
the form of the vector is defined by the parameter list (the entire
linked list pointed to by a paramptr).

      parameter vector (pv).  a set of integers (elements) that represent the
encoding due to one parameter record.  each parameter vector is made up of
one or more window vectors.  the integers (elements) represent the
encoding of one region of the sequence.

      window vector (wv).  a set of integers (elements) that represent the
encoding of a sequence over a particular window.  the length of the
window vector is always four raised to the coding level.  (eg. for
mononucleotides the length is 4, for dinucleotides the length is 16, etc).
a window vector is made up of several elements.

element.  one of the integers in a vector.  this is the number of a
particular n-tide (eg a, gxc, etc) found in the window.

*)
(* end module encode.info *)

(* begin module encode.type.param *)
(* the following types allow the user to specify parameters which will
   be used to encode the sequences in the book. *)
(* spaces are allowed between the encoded bases, and the number of
   bases to be skipped between each encoded pair of bases are kept
   in this linked list of integers *)
      spaceptr = ^spacelist;
      spacelist = record
         skips : integer; (* bases skipped to next coded base *)
         next : spaceptr; (* points to next spacing number *)
      end;

(* the encoding parameters for each region are stored in these records,
   and the records for each region are connected into a linked list of
   all the encoding parameters for the entire sequences *)
      endpoints = (start,stop);  (* end points of a coding region *)
      paramptr = ^parameter;
      parameter = record  (* these are the instructions for doing the coding *)
         range : array[endpoints] of integer; (* the bases to be coded by these
                                 instructions, relative to the alignedbase *)
         window : integer; (* the number of bases included in a coding vector *)
         wshift : integer;  (* movement of the window to the next site *)
         coding : integer; (* the 'level' at which the coding is being done,
                             i.e., 1: mononucleotides; 2: dinucleotides; ... *)
         spaces : spaceptr;  (* the spacing between the encoded bases *)
         cshift : integer;  (* shift to the next coding unit in the window *)

         (* values calculated at read-in time *)
         wvlength: integer; (* length of a window vector.  this is coding
            raised to the 4th power *)
         pvlength: integer; (* parameter vector length.  the vector made up
            of a series of window vectors. *)
         (* note: pvlength/wvlength is the number of windows in the
            parameter range *)

         next : paramptr;  (* set of instructions for coding another region *)
      end;

(* end module encode.type.param *)

(* begin module vector.type *)
(* vectors are useful for many applications, including the encoding of
   sequences.  this vector type is designed to be flexible enough to be
   used whenever one needs a vector.  it is designed as a linked list so
   it can contain as many elements as are ever needed.  the actual 'vector'
   is a record containing the length and a pointer to the first 'vectorpart'.
   that vectorpart is a record containing an array of the first 'maxvecpart'
   elements  (maxvecpart is a constant, from module vector.const, which must
   be included) and a pointer to the next 'vectorpart'.  the elements are
   of type real so that either integers or reals may be used.  *)
      partptr = ^vectorpart;
      vectorpart = record
         numbers : array[1..maxvecpart] of real;
         next : partptr;
      end;

      vector = record
         length : integer;
         part : partptr;
      end;
(* end module vector.type version = 'matmod 1.55  83 jul 7 tds/gds'; *)

var
(* begin module matmod.var *)
      encseq: text; (* the output of the encode program *)
(* end module matmod.var *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.44 82 aug 29 tds/gds'; *)

(* begin module linear.regression *)
(* linear regression procedure
origin: 1980 august 5
      by thomas schneider
      copyright 1986 *)

procedure regress ( control: char; (* described below *)
                    x, y: real;    (* the data pairs *)
                var sumx, sumy, sumxsqd, sumysqd, sumxy, (* internal records *)
(* mean, variance and covariance of x and y: *)
                    ex, ey, varx, vary, covxy,
(* correlation coefficient: *)
                    r,
(* m = slope, b = y intercept as in y = mx + b *)
                    m, b
                        : real;
(* number of data pairs entered: *)
                var n: integer);
(* linear regression for variables x and y.
   the control variable has only three acceptable states:
      c clear all variables to zero
        (note that r is set to 2.0, so that a program can check whether
         or not the results have been calculated yet. the recommended test
         is not for equality, since round-off could cause problems.  use
              if r > 1.5 then (no calculations yet)
                         else (calculations done)    )
      e enter the x and y values into the internal sums
      r results are calculated from the sums
*)
begin
      if not( control in ['c','e','r']) then begin
         writeln(output,' linear regression control variable value, "',control,
                        '" is not acceptable.');
         writeln(output,' it must be in ["c","e","r"]');
         halt
      end

      else case control of
         'c': begin (* clear *)
            x:=0.0; y:=0.0;
            sumx:=0.0; sumy:=0.0;
            sumxsqd:=0.0; sumysqd:=0.0;
            sumxy:=0.0;
            ex:=0.0; ey:=0.0;
            varx:=0.0; vary:=0.0;
            covxy:=0.0;
            r:=2.0; (* this (impossible) value can act as a flag that
                       shows that the calculations have not been made. *)

            m:=0.0; b:=0.0;
            n:=0
         end;

         'e': begin (* enter data *)
            sumx:=sumx + x;
            sumy:=sumy + y;
            sumxsqd:=sumxsqd + x*x;
            sumysqd:=sumysqd + y*y;
            sumxy:=sumxy + x*y;
            n:=succ(n)
         end;

         'r': begin (* calculate results *)
            (* check for conditions that would bomb the program *)
            if n = 0 then begin
               writeln(output,'regress: n is 0; no samples recorded');
               halt
            end;
            if n = 1 then begin
               writeln(output,'regress: n is 1; regression impossible');
               halt
            end;

            ex:=sumx/n;
            ey:=sumy/n;

            varx:=(sumxsqd/n) - ex*ex;
            vary:=(sumysqd/n) - ey*ey;
            if varx=0.0 then begin
               writeln(output,'regress: variance of x is zero;',
                              ' regression impossible');
               halt
            end;
            if vary=0.0 then begin
               writeln(output,'regress: variance of y is zero;',
                              ' regression impossible');
               halt
            end;

            covxy:=(sumxy/n) - ex*ey;

            r:=covxy/(sqrt(varx * vary));

            m:=covxy/varx;
            b:=ey - m*ex
         end
      end
end;
(* end module linear.regression *)

(* begin module random *)
procedure random(var seed: real);
(* the seed value is used to generate a random number in the range 0 to 1.
the seed is then altered, to provide the seed of the next random number.
in other words, the seed is the random number.  to obtain the next random
number in the pseudo-random series, simply call random again with the same
seed.  note that the series will not be truely random, since the previous
value determines the next value.  this is "pseudo-random".

an initial seed of 4.0 is suggested.  note that this is not the first
random number.  you can use different seeds to get different pseudo-random
number series.

the expected mean and standard deviation of a flat distribution is
calculated using simple integrals:

       1
mean = / x dx = 0.5
       0
           1           2
variance = / (x - mean)  dx = 1/12;  stdev = sqrt(variance) = .2886751
           0

(the integral for the variance is solved easily by substituting y = x - mean.)

   the algorithm is derived from:
pascal programs for scientists and engineers, alan r. miller
sybex, berkeley ca (1981).
which in turn was an adaptation from hp-35 applications programs *)
const
      pi = 3.141592653589; (* a value used to help mix up the result *)
var
      intermediate: real; (* an intermediate result *)
begin (* random *)
      intermediate := seed + pi;
      intermediate := exp(5.0*ln(intermediate));
      seed := intermediate - trunc(intermediate)
end; (* random *)
(* end module random *)

(* begin module random.test *)
procedure randomtest(var table: text);
(* this procedure tests the random procedure by calling it many
times and printing a table representing the results,  the mean and the
standard deviation for a series of tests.  this random testing procedure
should be run every time random is used on a new computer system. *)
const
      n = 100; (* the number of calls to random per test *)
      tests = 20; (* the number of tests to apply *)
      seed = 4.0; (* the seed for generation *)
var
      test: integer; (* the number of the current test *)
      call: integer; (* the number of the current call *)
      x, (* the random variable *)
      sumx, (* the sum of x during one test *)
      sumxsqd, (* the sum of x squared *)
      mean, (* the mean of x in one test *)
      std: (* the standard deviation of x in one test *)
         real;
begin (* randomtest *)
      writeln(table);
      (* initialize things *)
      x := seed; (* start x.  note that this is not the first random number *)
      writeln(table, ' test of the random number generator:');
      writeln(table);
      writeln(table,'     mean     std dev');
      writeln(table,'    (0.5000)  (0.2887)  (expected values)');
      writeln(table,' ===================== ', n:1, ' calls per test');

      for test := 1 to tests do begin
         (* initialize values for one test *)
         sumx := 0;
         sumxsqd := 0;

         for call := 1 to n do begin
            random(x); (* generate the next random number *)
            sumx := sumx + x;
            sumxsqd := sumxsqd + x*x
         end;

         mean := sumx/n;
         std := sqrt( (n/(n-1)) * (sumxsqd/n - mean * mean) );
         writeln(table,' ', mean:10:4, std:10:4)
      end
end; (* randomtest *)
(* end module random.test *)

(*********************************************************************)
(*********************************************************************)

(* begin module newton.test.functions *)
function f(x: real):real;
(* a function for testing the newton's method function *)
begin (* f *)
      f := ln(x)
end; (* f *)

function fp(x: real):real;
(* the derivative of the function f, for testing the newton's method function *)
begin (* fp *)
      fp := 1/x
end; (* fp *)
(* end module newton.test.functions *)

(* begin module newton *)
function newton(initialx,
                tolerance: real;
                tries: integer;
            var passes: integer): real;
(* this function uses newton's method to find a root of the function
f(x).  f(x) is coded externally from this function, as is fp(x), the
derivative of f(x) with respect to x.  an initial guess for x, initialx,
must be given along with a tolerance.  the returned value will be within
the region [root-tolerance, root+tolerance].  if the root can not be
found within 'tries' tries, 'passes' will equal 'tries' and the current
value of x will be returned.  the user must determine if newton found
the root by comparing 'passes' to 'tries'. *)
var
      deltax: real; (* the change in x at each pass *)
      searching: boolean; (* test to continue the search for the root *)
      x: real; (* the independent variable *)
begin (* newton *)
      x := initialx;
      passes := 0;
      searching := true;
      while searching and (passes < tries) do begin
         (* writeln(output, passes:5, x:10:5);*)
         deltax := f(x)/fp(x);
         if abs(deltax) > tolerance
         then begin
            x := x - deltax;
            passes := passes + 1
         end
         else searching := false
      end;
      newton := x
end; (* newton *)
(* end module newton *)

(* begin module newton.test *)
procedure newtontest(var fout: text);
(* test the newton method function. *)
const
      initialx =  0.00000001; (* makes for a nice demo *)
      maxtries = 100; (* do not get into an infinite loop... *)
      tolerance = 0.00000001; (* you may want to try zero also... *)
var
      answer: real; (* the value returned by newton *)
      passes: integer; (* number of times newton tried *)
      tries: integer; (* how many times we ask newton to try *)
begin (* newtontest *)
      writeln(fout);
      writeln(fout,' test of newton''s method.');
      writeln(fout,' the function is the natural logarithm, for which the');
      writeln(fout,' root should be 1.00.');
      writeln(fout,' the tolerance in the answer is:  ',tolerance:10:8);
      writeln(fout,' the initial guess is:',' ':12,initialx:10:8);

      tries := 0;
      passes := 0;
      while (tries < passes+1) and (tries < maxtries) do begin
         tries := tries + 1;
         answer := newton(initialx, tolerance, tries, passes);
         writeln(fout,' tries: ',tries:4,
                      ' passes: ',passes:4,
                      ' answer: ',answer:10:8);
      end
end; (* newtontest *)
(* end module newton.test *)

(*********************************************************************)
(*********************************************************************)

(* begin module vector.io *)
(* *********************************************************************** *)
(* begin module vector.primitives *)
(* end module vector.primitives *)

(* begin module vector.readvector *)
(* end module vector.readvector *)

(* begin module vector.writevector *)
(* end module vector.writevector *)
(* *********************************************************************** *)
(* end module vector.io *)

(* begin module vector.functions *)
(* *********************************************************************** *)
(* begin module vector.io *)
(* end module vector.io *)

(* begin module vector.dotproduct *)
(* end module vector.dotproduct *)

(* begin module vector.magnitude *)
(* end module vector.magnitude *)

(* begin module vector.normalize *)
(* end module vector.normalize *)
(* *********************************************************************** *)
(* end module vector.functions *)

(* begin module vector.primitives *)
(* these functions and procedures do manipulations on vectors. *)

function vget(v:vector; pos:integer): real;
(* this returns from vector 'v' the value of the element at position 'pos' *)
var i: integer;
begin
      if (pos > v.length) or (pos < 1) then begin
         writeln(output,' error in call to function vget:',
            ' position ',pos:1,' is beyond the end of the vector');
         halt;
      end;

      (* move to the correct 'vectorpart' *)
      for i := 1 to ((pos -1) div maxvecpart) do  v.part := v.part^.next;

      (* get the proper array element from this part *)
      vget := v.part^.numbers[(pos -1) mod maxvecpart + 1];
end;

procedure vput(var v:vector; pos:integer; number: real);
(* this puts into vector 'v' the value of 'number' at position 'pos' *)
var i: integer;
      firstpart: partptr;  (* of the vector *)
begin
      if (pos > v.length) or (pos < 1) then begin
         writeln(output,' error in call to function vput:',
            ' position ',pos:1,' is beyond the end of the vector');
         halt;
      end;

      (* move to the correct 'vectorpart' *)
      firstpart := v.part;
      for i := 1 to ((pos -1) div maxvecpart) do  v.part := v.part^.next;

      (* put the 'number' into the proper array element for this part *)
      v.part^.numbers[(pos -1) mod maxvecpart + 1] := number;
      v.part := firstpart;
end;

procedure makevector(var v: vector; l: integer);
(* create the linked list of vector-parts which are needed for a vector
   of length 'l' *)
var   numparts,  (* number of parts needed for the vector *)
      i: integer; (* index to the parts of the vector *)
      firstpart,  (* of the vector *)
      newpart: partptr;  (* of the vector *)
begin
      if l < 1 then begin
         writeln(output,' makevector: positive length required');
         halt
      end;

      v.length := l;
      new(v.part);
      firstpart := v.part;
      numparts := (v.length - 1) div  maxvecpart + 1;
      for i := 1 to (numparts - 1) do begin
         new(newpart);
         v.part^.next := newpart;
         v.part := newpart;
      end;
      v.part^.next := nil;
      v.part := firstpart;
end;
(* end module vector.primitives *)

(* begin module vector.readvector *)
procedure readvector(var thefile:text; var v: vector);
(* read the elements of v into v from 'thefile'.  v must already be set
   up (all the parts created and linked together) before calling.
   'thefile' must contain, from the cursor point until the end of the
    vector, only numbers, either integers or reals; otherwise it will
    bomb.  the vector ends with an end of line so that
    end of file can be tested for. *)
var i,j: integer;  (* indecies *)
      numparts,  (* number of parts in the vector *)
      lastpart: integer; (* the number of elements in the last vector part *)
      firstpart: partptr;  (* pointer to the first vector part *)
begin
      firstpart := v.part;
      numparts := (v.length - 1) div maxvecpart + 1;
      lastpart := (v.length - 1) mod maxvecpart + 1;
      for i := 1 to (numparts - 1) do begin
         for j := 1 to maxvecpart do read(thefile,v.part^.numbers[j]);
         v.part := v.part^.next;
      end;
      for j := 1 to lastpart do read(thefile,v.part^.numbers[j]);
      readln(thefile);
      v.part := firstpart;
end;
(* end module vector.readvector *)

(* begin module vector.writevector *)
procedure writevector(var thefile: text; v: vector; y,z: integer);
(* writes the elements of 'v' to 'thefile'.  the integers y and z are:
    y: the size of the field for printing the number;
    z: the size of the decimal part of the field,
    if z = 0 then integers will be printed. *)
var   pos,   (* posititon in the vector *)
      i,j: integer;  (* indecies *)
      numparts,  (* number of parts in the vector *)
      lastpart: integer; (* the number of elements in the last vector part *)
      firstpart: partptr;  (* pointer to the first vector part *)
      x: integer; (* the number of elements to write on a line *)
begin
      x := trunc(vpagewidth/(y+1));
      pos := 0;
      firstpart := v.part;
      numparts := (v.length - 1) div maxvecpart + 1;
      lastpart := (v.length - 1) mod maxvecpart + 1;
      if (z = 0) then begin
         for i := 1 to (numparts - 1) do begin
            for j := 1 to maxvecpart do begin
               write(thefile,' ',round(v.part^.numbers[j]):y);
               pos := succ(pos);
               if (pos mod x = 0) then writeln(thefile);
            end;
            v.part := v.part^.next;
         end;
         for j := 1 to lastpart do begin
            write(thefile,' ',round(v.part^.numbers[j]):y);
            pos := succ(pos);
            if ((pos < v.length) and (pos mod x = 0)) then writeln(thefile);
         end
      end
      else begin
         for i := 1 to (numparts - 1) do begin
            for j := 1 to maxvecpart do begin
               write(thefile,' ',v.part^.numbers[j]:y:z);
               pos := succ(pos);
               if (pos mod x = 0) then writeln(thefile);
            end;
            v.part := v.part^.next;
         end;
         for j := 1 to lastpart do begin
            write(thefile,' ',v.part^.numbers[j]:y:z);
            pos := succ(pos);
            if ((pos < v.length) and (pos mod x = 0)) then writeln(thefile);
         end;
      end;
      writeln(thefile);
      v.part := firstpart;
end;
(* end module vector.writevector *)

(* begin module vector.dotproduct *)
function dotproduct(vectora,vectorb: vector): real;
(* this returns the dotproduct of 'vectora' and 'vectorb' *)
var i: integer;  (* an index *)
    j: real;  (* partial products *)
begin
      if vectora.length <> vectorb.length then begin
         writeln(output,' function dotproduct: vector lengths must',
            ' be equal');
         halt
      end;
      j := 0;
      for i := 1 to vectora.length do
         j := j + vget(vectora,i) * vget(vectorb,i);
      dotproduct := j;
end;
(* end module vector.dotproduct *)

(* begin module vector.magnitude *)
function magnitude(var v: vector): real;
(* find the magnitude (length) of the vector v *)
begin
      magnitude := sqrt(dotproduct(v,v));
end;
(* end module vector.magnitude *)

(* begin module vector.normalize *)
procedure normalize(var v: vector);
(* this replaces the vector v with the congruent vector of unit length,
   i.e.,  the resulting vector v has the property that v.v = 1   *)
var i: integer;  (* index *)
      length: real;  (* of the unnormalized vector *)
begin
      length := magnitude(v);
      for i := 1 to v.length do
         vput(v,i,vget(v,i)/length);
end;
(* end module vector.normalize *)

(* begin module vector.vset *)
procedure vset(thevalue: real; var v: vector);
(* set the value thevalue into v *)
var   i: integer; (* an index *)
begin
      for i := 1 to v.length do vput(v,i,thevalue)
end;
(* end module vector.vset *)

(* begin module vector.vadd *)
procedure vadd(a: vector; var b: vector);
(* add the vector a into the vector b *)
var   i: integer; (* an index *)
begin
      if a.length <> b.length then begin
         writeln(output,' function vadd: vector lengths must be equal');
         halt
      end;

      for i := 1 to a.length
         do vput(b,i,vget(b,i)+vget(a,i))
end;
(* end module vector.vadd *)

(* begin module vector.vsub *)
procedure vsub(a: vector; var b: vector);
(* subtract the vector a from the vector b *)
var   i: integer; (* an index *)
begin
      if a.length <> b.length then begin
         writeln(output,' function vsub: vector lengths must be equal');
         halt
      end;

      for i := 1 to a.length
         do vput(b,i,vget(b,i)-vget(a,i))
end;
(* end module vector.vsub *)

(* begin module vector.vmult *)
procedure vmult(k: real; var v: vector);
(* multiply the vector v by the scaler k *)
var   i: integer; (* an index *)
begin
      for i := 1 to v.length
         do vput(v,i,k*vget(v,i))
end;
(* end module vector.vmult *)

(* begin module vector.extract *)
procedure extract(    a: vector; astart, astop: integer;
                  var b: vector; bstart: integer);
(* extract the segment defined by [astart,astop] of vector a
and insert it into vector b starting at bstart, going in the
positive direction.  any segment of a may be used, so one may
reverse the order of the values.  all checking is done by vget
and vput. *)
var   i: integer; (* index *)
      precalc: integer; (* a precalculated value to increase speed *)
begin
      if astart <= astop then begin
         precalc := bstart-astart;
         for i := astart to astop
            do vput(b,i+precalc,vget(a,i))
      end
      else begin
         precalc := bstart+astart;
         for i := astart downto astop
            do vput(b,precalc-i,vget(a,i))
      end
end;
(* end module vector.extract *)

(* begin module arccos *)
function arccos(x: real): real;
(* Calculate the arccosine.  Pascal only supplies arctan(angle),
so we have to calculate it ourselves.  Begin from

      2           2
   sin angle + cos angle = 1

             2
divide by cos angle to obtain a relationship between tan(angle) and
cos(angle).  Rearrange.
   The special case of x=0 must be dealt with, since this form
would give a division by zero.  The value that must be returned is
pi/2, since cos(pi/2)=0.  A nice way to find this is to note that at a
45 degree angle (pi/4) the tangent is 1, so that pi=4arctan(1),
and pi/2 = 2 * arctan(1).
   Angles larger than pi/2 can be had if x is negative. *)
var
   pi: real; (* 3.1415... *)
begin
   if (x < -1.0) or (x > +1.0) then begin
      writeln(output,' function arccos(x):',
         ' x was not -1<=x<=+1');
      halt
   end;
 
   if x=0.0
   then arccos := 2.0 * arctan(1.0)
   else if x > 0.0
        then arccos := arctan(sqrt( (1.0/(x*x)) - 1.0))
        else begin
           pi := 4.0 * arctan(1.0);
           arccos := pi - arctan(sqrt( (1.0/(x*x)) - 1.0))
        end
end;
(* end module arccos *)
 
(* begin module vector.angle *)
function angle(var a,b: vector): real;
(* find the angle between the vectors a and b.
to calculate the angle between the vectors we use
   dotproduct(a,b)=magnitude(a)*magnitude(b)*cos(angle)
which requires arccos(angle), supplied by module arccos. *)
begin
      angle := arccos( dotproduct(a,b)/
                     ( magnitude(a) * magnitude(b)) )
end;
(* end module vector.angle *)

(* begin module vector.test *)
procedure vectortest(var fout: text);
(* test of some of the vector properties *)
var   a,b,c,d,e,f: vector;  (* vectors for trying out the functions *)
      i: integer; (* index for vector *)
      pi: real; (* 3.1415... but as accurate as the machine can calculate *)
begin
      makevector(a,2);
      makevector(b,2);
      makevector(c,2);
      makevector(d,5);
      makevector(e,3);
      makevector(f,2);

      vput(a,1,1); vput(a,2,0);
      vset(1,b);
      for i:=1 to 5 do vput(d,i,i);
      vput(f,1,-1); vput(f,2,1);

      writeln(fout);
      writeln(fout,' test of vector math');
      write(fout,' vector a: '); writevector(fout,a,1,0);
      write(fout,' vector b: '); writevector(fout,b,1,0);
      write(fout,' vector d: '); writevector(fout,d,2,1);
      write(fout,' vector f: '); writevector(fout,f,2,1);

      writeln(fout,' extract parts of d: ');
      extract(d,2,3,c,1);
      write(fout,' from 2 to 3: '); writevector(fout,c,1,0);
      extract(d,5,3,e,1);
      write(fout,' from 5 to 3: '); writevector(fout,e,1,0);
      extract(d,3,5,e,1);
      write(fout,' from 3 to 5: '); writevector(fout,e,1,0);
      extract(d,4,2,e,1);
      write(fout,' from 4 to 2: '); writevector(fout,e,1,0);

      write(fout,' a + b = ');
      vset(0,c); vadd(a,c); vadd(b,c);
      writevector(fout,c,1,0);

      write(fout,' a - b = ');
      vset(0,c);  vadd(a,c); vsub(b,c);
      writevector(fout,c,1,0);

      write(fout,' a is multiplied by 5, a now is ');
      vmult(5,a);
      writevector(fout,a,1,0);

      writeln(fout,' magnitude(a) = ',magnitude(a):5:3);
      writeln(fout,' magnitude(b) = ',magnitude(b):5:3);

      writeln(fout);
      writeln(fout,' test arccos function');
      pi := 2*arccos(0);
      writeln(fout,' pi = ',pi:6:4);
      writeln(fout,
                   'cos(arccos(0.00 * pi/4.0)) * 4/pi = ',
                    cos(arccos(0.00 * pi/4.0)) * 4/pi:6:4);
      writeln(fout,
                   'cos(arccos(0.12 * pi/4.0)) * 4/pi = ',
                    cos(arccos(0.12 * pi/4.0)) * 4/pi:6:4);
      writeln(fout,
                   'cos(arccos(0.25 * pi/4.0)) * 4/pi = ',
                    cos(arccos(0.25 * pi/4.0)) * 4/pi:6:4);
      writeln(fout,
                   'cos(arccos(0.37 * pi/4.0)) * 4/pi = ',
                    cos(arccos(0.37 * pi/4.0)) * 4/pi:6:4);
      writeln(fout,
                   'cos(arccos(0.50 * pi/4.0)) * 4/pi = ',
                    cos(arccos(0.50 * pi/4.0)) * 4/pi:6:4);
      writeln(fout,
                   'cos(arccos(0.75 * pi/4.0)) * 4/pi = ',
                    cos(arccos(0.75 * pi/4.0)) * 4/pi:6:4);
      writeln(fout,
                   'cos(arccos(1.00 * pi/4.0)) * 4/pi = ',
                    cos(arccos(1.00 * pi/4.0)) * 4/pi:6:4);

      writeln(fout,' angle(a,b) = pi/',pi/angle(a,b):6:4);
      writeln(fout,' angle(b,f) = pi/',pi/angle(b,f):6:4);
      writeln(fout,' angle(a,f) = pi*',angle(a,f)/pi:6:4);
end;
(* end module vector.test *)

(*********************************************************************)
(*********************************************************************)

(* begin module encode.readencpar *)
(* read encode parameters *)

function vectorsize(param: paramptr): integer;
(* determine the size of the vector generated by a string of parameter
   records, begin with 'param' *)
var   size: integer;  (* of the vector, so far *)
begin
      size := 0;
      while (param <> nil) do begin
         size := size + param^.pvlength;
         param := param^.next;
      end;
      vectorsize := size;
end;

function paramsize(param: paramptr): integer;
(* determine the size of the vector generated by one parameter record *)
var   rangesize,  (* the number of bases in the range *)
      numwindows: integer; (* the number of windows which start in the
                  range, taking into account the shifting *)
begin
      with param^ do begin
         rangesize := range[stop] - range[start] + 1;
         numwindows := trunc((rangesize -1)/wshift + 1);
         paramsize := numwindows * wvlength
      end
end; (* paramsize *)


procedure readencpar(var thefile: text; var param: paramptr;
                    var regions, vectorlength: integer);
(* read the parameter list from 'thefile' and put the information into the
   linked parameter list, beginning with 'param'; the number of parameter
   records is returned in 'regions'.  the length of each vector is
   vectorlength *)
var   aparam,  (* from the parameter list *)
      newparam: paramptr;  (* the new parameter record from 'thefile' *)
      firstspaces,  (* in the spaces list *)
      newspaces: spaceptr; (* for the spaces list *)
      i,j: integer;  (* indices *)
      ch: char;  (* to check for proper positioning *)
begin
      reset(thefile);
      if eof(thefile) then begin
         writeln(output,' encoded sequence file is empty');
         halt
      end;
      new(aparam);
      if param = nil then new(param);
      aparam := param;
            (* skip the header *)
      for i := 1 to 2 do readln(thefile);

            (* skip a blank line in the header *)
      readln(thefile);

            (* read the number of regions *)
      readln(thefile,regions);

      for i := 1 to regions do begin
         with aparam^ do begin
               (* read the range *)
            read(thefile,range[start]);
            repeat read(thefile,ch) until (ch = 'o');
            readln(thefile,range[stop]);
               (* read the window size and shift *)
            readln(thefile,window);
            readln(thefile,wshift);
               (* read the coding level and arrangement *)
            read(thefile,coding);
            new(spaces);
            if (coding > 1) then begin
               repeat read(thefile,ch) until (ch = ':');
               new(firstspaces);
               spaces := firstspaces;
               read(thefile,spaces^.skips);
               for j := 2 to coding -1 do begin
                  new(newspaces);
                  spaces^.next := newspaces;
                  spaces := newspaces;
                  read(thefile,spaces^.skips);
               end;
               spaces^.next := nil;
               spaces := firstspaces
            end
            else spaces := nil;
            readln(thefile);
            readln(thefile,cshift);

            (* set up wvlength *)
            wvlength := round(exp(coding*ln(4)));
         end;

         (* set up pvlength *)
         aparam ^.pvlength := paramsize(aparam);

         if (i < regions) then begin
            new(newparam);
            aparam^.next := newparam;
            aparam := newparam;
         end;
      end;
      aparam^.next := nil;

      (* read the vector size *)
      readln(thefile,vectorlength);
      if vectorlength <> vectorsize(param)
      then begin
         writeln(output,' vector length in the encoded file');
         writeln(output,' does not correspond to the parameters');
         halt
      end;
end;

(* end module encode.readencpar version = 'matmod 1.55  83 jul 7 tds/gds'; *)

(* begin module copy.dummy *)
(* these routines are needed to keep enchead happy; they
are available from other module libraries *)

procedure copyaline(var a,b:text);
begin (* dummy *) end;

function copylines(var a,b: text; i: integer): integer;
begin copylines := 0 (* dummy *) end;
(* end module copy.dummy *)

(* begin module encode.enchead *)
procedure enchead(var encseq, fout: text; var length: integer);
(* copy the information from the start of the encoded vector
file encseq to fout.  to make fout have the same form as encseq,
it is necessary for the program which calls this procedure to
identify itself to the fout file on one line.
   the vector lengths is returned.  the encseq file is positioned
for reading of the vectors immediately after enchead.
   in contrast, the fout file is left at the end of the
vector length line so the user may add more information to the line.
after completing the line with writeln, a second writeln is
required before the vectors begin. *)
var   regions,  (* number of independently encoded regions *)
      i: integer; (* the number of lines of parameters *)
begin
      reset(encseq);
      readln(encseq); (* skip the previous program identification *)
      copyaline(encseq,fout); (* copy book information *)
      copyaline(encseq,fout); (* copy blank line *)
         (* copy the encoding parameters to the fout *)
      read(encseq,regions);
      write(fout,' ',regions:1);
      i := 6 * regions + 2;  (* 6 lines for each parameter and
         2 lines for the coded regions *)

      if copylines(encseq,fout,i) <> i then begin
         writeln(output,' encseq is missing parameters');
         halt
      end;

      (* now get the length *)
      readln(encseq,length);
      write(fout,' ',length:1,' is the vector length');
end;
(* end module encode.enchead *)

(* begin module encode.functions *)
(* these functions are quite handy to help one know where one is
in an encoded vector.  they all take a parameter list as read by
readencpar, and the number of the element of the vector one is interested
in.  this element value runs from 0 to one less than the length of the
parameter vector.  examples are in evprint. *)

function beginpv(aparam: paramptr; element: integer): boolean;
(* are we at the beginning of a parameter vector? *)
begin (* beginpv *)
      if aparam <> nil (* is there a vector? *)
      then beginpv := (element = 0)
      else beginpv := true (* note circular use of aparam in evprint... *)
end; (* beginpv *)

function endpv(aparam: paramptr; element: integer): boolean;
(* are we at the end of a parameter vector?
the zero element is the first element of the parameter vector. *)
begin (* endpv *)
      if aparam <> nil
      then endpv := (element >= aparam^.pvlength)
      else endpv := true (* past the end of all vectors *)
end; (* endpv *)

function beginwv(aparam: paramptr; element: integer): boolean;
(* are we at the beginning of a window vector in the current parameter
vector?  the zero element is the first element of the parameter vector. *)
begin (* beginwv *)
      if aparam <> nil
      then with aparam^ do beginwv := ((element mod wvlength) = 0)
      else beginwv := false (* no further parameters *)
end; (* beginwv *)

function endwv(aparam: paramptr; element: integer): boolean;
(* are we at the end of a window vector in the current parameter vector?
the zero element is the first element of the parameter vector. *)
begin (* endwv *)
      if aparam <> nil
      then with aparam^ do endwv := (((element + 1) mod wvlength) = 0)
      else endwv := true (* no further parameters *)
(*;writeln(output,'element=',element:1,'   wvlength=',aparam^.wvlength:1); *)
end; (* endwv *)

function encposition(aparam: paramptr; element: integer): integer;
(* this returns the aligned position corresponding to the element in
a parmaeter list (aparam).
the zero element is the first element of the parameter vector. *)
begin (* encposition *)
      with aparam^
         do encposition := range[start] + (element div wvlength) * wshift
end; (* encposition *)
(* end module encode.functions *)

(* begin module encode.evprint *)
(* procedures for encoded vector printing *)

procedure evln(var afile: text; element: integer; curparam: paramptr);
(* see evprint for definitions.  this procedure is used to provide
writelns to afile when the user wants to control the display better. *)
begin (* evln *)
      (* for end of windows: *)
      if beginwv(curparam,element)
      then writeln(afile)
      else if curparam = nil then writeln(afile); (* end of vector *)

      (* for end of parameters: *)
      if beginpv(curparam,element) then writeln(afile)
end; (* evln *)

procedure evheader(var afile: text;
                       posfield, valfield: integer;
                       firstparam: paramptr;
                   var curparam: paramptr);
(* see evprint for definitions of the variables and how to use evheader.
evheader must be called if you want to provide your own writelns
to the evprint display, ie, when ln is false *)
const
      codingmax = 4; (* the maximum coding level allowed *)
var
      bases: array[1..codingmax] of char; (* the bases *)
      i,j,n: integer; (* indices for loops *)
      size: integer; (* the number of characters in the encoded object.
                        a=1, aa=2, axa=3, etc *)
      s: spaceptr; (* for finding size *)
procedure space(i:integer);
(* put out i spaces to afile *)
var
      index: integer; (* for i *)
begin (* space *)
      for index := 1 to i do write(afile,' ');
end; (* space *)
procedure advance(level: integer);
(* advance the bases at the level.  aa -> ac -> ag -> ... -> tt *)
begin (* advance *)
      case bases[level] of
         'a': bases[level] := 'c';
         'c': bases[level] := 'g';
         'g': bases[level] := 't';
         't': begin
            bases[level] := 'a';
            advance(level - 1) (* aha.. *)
         end
      end
end; (* advance *)
begin (* evheader *)
      if curparam = nil then curparam := firstparam; (* start or recycle *)

      with curparam^ do begin
         if coding > codingmax then begin
            writeln(output,' evheader: coding levels must not be higher',
                           ' than ', codingmax:1);
            halt
         end;

         (* find the size of the encoded object *)
         size := coding;
         s := curparam^.spaces;
         while s <> nil do begin
            size := size + s^.skips;
            s := s^.next
         end;

         (* see if there is room for printing *)
         if abs(valfield) < size then begin
            writeln(output,' evheader: valfield (',valfield:1,
                           ') cannot be less than coding level and spaces (',
                           size:1,')');
            halt
         end;

         (* clear bases *)
         for i := 1 to coding do bases[i] := 'a';

         (* print the header *)
         (* room for space, ' to ' and two numbers *)
         space(1 + 4 + 2*posfield);
         for i := 1 to wvlength do begin
            space(1 + abs(valfield) - size);
            s := spaces;
            for n := 1 to coding do begin
               write(afile, bases[n]);
               if s <> nil then begin
                  for j := 1 to s^.skips
                     do write(afile,'x');
                  s := s^.next
               end
            end;

            (* advance the bases except for the last time *)
            if i <> wvlength then advance(coding)
         end
      end
end; (* evheader *)

procedure evprint(var afile: text;         (* the file to print to *)
                      posfield: integer;   (* the number of characters in
                                              which to print the position *)
                      val: real;           (* the value to print *)
                      valfield,            (* the number of characters
                                              in which to print the value *)
                      valdecimal: integer; (* the number of characters
                                              devoted to the digits below the
                                              decimal point of val.  if <= 0
                                              val is printed as an integer *)
                      ln: boolean;         (* control writelns (see below) *)
                      firstparam: paramptr; (* the first parameter *)
                  var curparam: paramptr;  (* the current encoding parameter *)
                  var element: integer);   (* the position in the vector *)
(* evprint is an encoded vector printing routine.  it is designed to make
generation of display of encoded vectors or calculated vectors based on
encoded vectors easy.  the basic idea is that evprint will print all the
display associated with one vector value.  so at the beginning it prints
an appropriate header from the vector parameters, and for each line it
prints the beginning and end positions of the window.
1) all fields printed out are separated by one space.
2) the value arguments are to be set by the user before calling evprint.
evprint is then called sequentially with all the vector values placed in val.
this allows the data to come from any data structure; it is not forced
to come from a particular array or file type.
3) to initialize, read in firstparam, set the element to 0 and curparam to nil.
evprint will restart at firstparam after each pass across the vector.
4) once the printing of the display has begun,
do not modify the element or curparam yourself.  it is also not wise to
modify any value parameters except val and possibly valdecimal.  the element
argument is used internally by evprint to keep track of the display.
the value arguments, after being set by the user initially, control the
form of the display.

control of writelns:  if ln is true, then evprint will take care of
writelns.  otherwise the user must do the writeln, as:
      if beginpv(curparam,element) then begin
         evheader(afile,posfield,valfield,firstparam,curparam);
         writeln(afile,[your headers for each column]);
      end;
      evprint(afile,...false...)
      if beginwv(curparam,element)
         then write(afile,[your own material on the same line]);
      evln(afile,curparam,element);
note that evprint had moved the element to the next position, so the
test is for begin of wv, not end of wv.

if valfield is negative (but abs(valfield) is a reasonable size) then
val will not be printed, allowing the user to do so after evprint.
this allows more flexible displays.
note: ln must be false to use this feature.
 *)
procedure sider;
(* produce the side of the display *)
var
      position: integer; (* the location that a window begins
                            from the aligned base *)
begin (* sider *)
      with curparam^ do begin
         position :=encposition(curparam, element);
         write(afile,' ',position: posfield, ' to ',
                 (position + window - 1):posfield)
      end
end; (* sider *)
procedure middler;
(* produce the middle of the display *)
begin (* middler *)
      write(afile,' ');
      if valfield > 0  (* to display or not to display... *)
      then if valdecimal <= 0 (* chose display *)
           then write(afile, round(val):valfield)
           else write(afile, val:valfield:valdecimal)
      else if ln then begin
         writeln(output,' evprint: if valfield is negative,',
                        ' then ln must be false');
         halt
      end
end; (* middler *)
procedure step;
(* step to the next position *)
begin (* step *)
      element := succ(element); (* move along the vector *)

      if endpv(curparam, element) (* are we off the end of the vector? *)
      then begin (* yes *)
         curparam := curparam^.next;
         element := 0;
      end
end; (* step *)
begin (* evprint *)
      if ln then if beginpv(curparam,element) then begin
         evheader(afile, posfield, valfield, firstparam, curparam);
         writeln(afile)
      end;

      if beginwv(curparam, element) then sider;

      middler;

      step;

      if ln then evln(afile,element,curparam)
end; (* evprint *)
(* end module encode.evprint *)


(* begin module encode.test *)
procedure enctest(var fin, fout: text);
(* test of encode modules:
      encode.type
      encode.vectorsize
      encode.functions
      encode.evprint
the file fin is expected to contain output from the encode program
the file fout is the results of the test *)
var
      firstparam, (* pointer to the parameter list *)
      aparam: (* one of the parameters *)
         paramptr;
      regions, (* number of separate parameter records *)
      vindex: integer; (* index to v *)
      vectorlength: (* length of the vector v *)
         integer;
      v: vector; (* a vector of data *)
      element: integer; (* a value of v *)
begin (* enctest *)
      writeln(fout);
      reset(fin);
      if eof(fin) then begin
         writeln(fout,' if file encseq contains the output from the encode');
         writeln(fout,' program, then parameter reading will be tested.')
      end
      else begin
         new(firstparam);
         readencpar(fin,firstparam,regions,vectorlength);
         writeln(fout,' the encode file contains vectors with the',
                      ' following ',regions:1,' sized components:');
         aparam := firstparam;
         while aparam <> nil do begin
            write(fout,' ',paramsize(aparam):5);
            aparam := aparam^.next
         end;
         writeln(fout);
         writeln(fout, ' the total size of each vector is ',vectorlength:1);
         writeln(fout);

         writeln(fout,' the first vector is:');
         makevector(v,vectorlength);
         readvector(fin,v);
         aparam := nil;
         element := 0;
         for vindex := 1 to vectorlength
            do evprint(fout,4,vget(v,vindex),5,0,true,
                       firstparam,aparam,element);
      end
end; (* enctest *)
(* end module encode.test *)

(* begin module simpson *)
procedure simpson(upper : real; var answer,tol : real);
(* Perform a numerical integration of the normal distribution using
Simpson's rule.  The variable "lower" is the lower bound of the
integration.  The variable "upper" is the upper bound of the
integration.  "answer" is the value of the integration, and "tol"
is the tolerance of the integration procedure, and thus the error
of the calculation.  Written by Mark Shaner, 1992.  Taken from
Miller, Alan R. PASCAL PROGRAMS FOR SCIENTISTS AND ENGINEERS. 
SYBEX Inc., 1981. pgs 260-291

2007 Mar 21:  Tom Schneider:  Using a tolerance of 1/maxint caused an
infinite loop on a 64 bit machine.  Presumably it was attempting to
compute the value to 20 or so decimal places.  To avoid this, the
tolerance is set to a more reasonable value.

*)

var
  i  (* a counter *): integer;

  x,x1,x2, (* the independent variables for calculating the value of
              functions *)
  pi, (* the value of pi *)
  val, (* the value of 1/sqrt(2*pi), it is used in calculating the value of the
          gaussian for any x value, and is defined as a variable in order to
          speed calculations. *)
  deltax, (* the distance between every x value and the subsequent one *)
  evensum, (* the sum of the area under each of the even numbered parabolas *)
  oddsum, (* the sum of the area under each of the odd numbered parabolas *)
  endsum, (* the sum of the area under the first and last parabolas *)
  endcor, (*  the value of the end correction, it is determined using dgauss *)
  sum1 : real; (* a place to store the previous sum so that it can be compared
                with the subsequent sum to determine if the tolerance level has
                been reached *)
  pieces : integer; (* the number of parabolas under the curve *)
  lower, (* the lower bound of the integration *)
  sum :real; (* the value of the area under the curve *)

begin
  pi := 4.0*arctan(1.0);
  val := 1/sqrt (2*pi);
  (* activate both files to be used *)
  lower := 0;
  upper := abs(upper);
{
  tol := 1/maxint;
}
  tol := 1.0E-6;
  pieces := 2;
  deltax := (upper-lower)/pieces;
  x := lower + deltax;
  oddsum := val*exp(-0.5*x*x);
  evensum := 0.0;
  x1 := lower;
  x2 := upper;
  endsum := val*exp(-0.5*x1*x1) + val*exp(-0.5*x2*x2);
  endcor := -(x1*val)*exp(-0.5*x1*x1) - (-(x2*val)*exp(-0.5*x2*x2));
  sum := (endsum + 4.0*oddsum)*deltax/3.0;

  repeat
    pieces := pieces*2;
    sum1 := sum;
    deltax := (upper - lower)/pieces;
    evensum := evensum + oddsum;
    oddsum := 0.0;
    for i := 1 to (pieces div 2) do
    begin
      x := lower + deltax*(2.0*i-1.0);
      oddsum := oddsum + val*exp(-0.5*x*x);
    end;
    sum := (7.0*endsum + 14.0*evensum + 16.0*oddsum + endcor*deltax)
          *deltax/15.0;
  until abs(sum-sum1) <= abs(tol*sum);
  answer := (0.5 - sum);
  if answer < 0.0 then answer := 0.0; (* safety for roundoff errors *)
(*
  writeln (upper:7:5,'      ',answer:7:5,'+/-',tol)
*)
end;
(* end module simpson *)

begin (* matmod *)
(* begin module matmod.main *)
      writeln(output, ' mathematics modules ', version);

      vectortest(output);

      newtontest(output);

      enctest(encseq, output);

      randomtest(output);
(* end module matmod.main *)
1: end. (* matmod *)
