program pascru(pin, pout, pascrup, note, pvar, output);
(* pascru: Crunch pascal source code

  Ryan Shultzaberger and Tom Schneider

  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu
  https://alum.mit.edu/www/toms/

  Module libraries used: prgmod.p makmod.p

 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.31; (* of pascru.p 2000 July 25
2000 July 25; 1.30: bug noted, improve documentation
2000 July 21; 1.29: space after note file added (TDS)
2000 July 20; 1.28: note file added (TDS)
2000 July 19; 1.23: remove extra spaces on output (TDS)
2000 July 19; 1.18: add pascrup (by TDS)
2000 July 19; 1.15: introduce ident type (by TDS)
2000 July 18; 1.08: change findvar to count characters written (by TDS)
2000 June 28; 1.00 origin *)
updateversion = 1.21; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.pascru *)
(*
name

   pascru: Crunch pascal source code

synopsis

   pascru(pin: in, pout: out,
          pascrup: in, note: in,
          pvar: out, output: out)

files

   pin:  Pascal program source code to be crunched and distorted.

   pout:  Manipulated pascal source code, which has no comments
          or structure and has randomized variable names.

   pascrup:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      seed: a real number between 0 and 1 used to start the random
         number generator.  The date and time is used if this
         number is outside 0 to 1.

      domanualpage (char)  copy the manual page (like this one) from the
          program to the start of the pout file if the first character is
          'm', otherwise not.

      protected words:  the remainder of the file is a list of words that
         should remain unchanged during the big crunch.  The words must be
         left justified.  All characters after the first space to the right
         are ignored, so one can put comments there.

   note:  This file is copied to the pout just below the manual page.
      It is an *exact* copy, so it must contain one or more Pascal comment.

   pvar: This file is a list of variable names and their random equivalent.

   output: messages to the user

description

   What should one do when one would like to give somebody source code so
   that they can compile a program on their own computer, but one would still
   like to make it difficult (if not impossible) for them to modify the
   code?  Pascru solves the problem.

   This program can to be used to crunch source code to make it difficult to
   be understood by a user.  It removes all comments and indentation.  It
   also wraps the lines, leaving no structure.  Furthermore, it converts all
   variable names into new variable names, each of which is a random letter
   of the alphabet followed by a 6 digit random number.  Finally, to maintain
   the source code documentation that describes how to run the program, it
   also puts the version line and documentation module at the start of the
   output file.

   Other than these alterations, the code is unchanged so that, for example,
   strings are preserved.  After compilation, the resulting source code will
   give identical results as the input code.

   Some words must not be changed according to the the PASCAL user manual and
   report, ISO Pascal Standard; 4th edition:

   * Reserved words (page 212)
   * Standard identifiers (page 211)

   * Other words that should be on the list
   e - exponentiation (a matter of how things are parsed)
   date

examples

pascrup:

1.21  version of pascru that this parameter file is designed for.
-0.50 seed for the random number generator 0-1: fixed, otherwise timeseed
m     m means do manual page at top, otherwise not
date            Sun Pascal compiler date/time key word
TimeStamp       GPC compiler date/time
DateValid       GPC compiler date/time
TimeValid       GPC compiler date/time
Year            GPC compiler date/time
Month           GPC compiler date/time
Day             GPC compiler date/time
DayOfWeek       GPC compiler date/time
Hour            GPC compiler date/time
Minute          GPC compiler date/time
Second          GPC compiler date/time
MicroSecond     GPC compiler date/time

documentation

@book{Jensen.Wirth1991,
author = "K. Jensen  
 and N. Wirth  
 and A. B. Mickel
 and J. F. Miner",
title = "Pascal User Manual and Report,
ISO Pascal Standard",
edition = "fourth", 
publisher = "Springer-Verlag",
address = "New York",
isbn = "0-387-97649-3",
year = "1991"}    

@book{Jensen1975,
author = "K. Jensen
 and N. Wirth",
title = "Pascal User Manual and Report",
publisher = "Springer-Verlag",
address = "New York",
year = "1975"}

see also

  {The module copymanualpage, which came from} makman.p{,
  extracts the program describe module.}

author

   Ryan Shultzaberger and Tom Schneider

bugs

   1. If a comment contains an open left parenthesis inside a comment, the
   resulting pout program will die, apparently the parenthesis is passed
   through instead of being absorbed inside the comment.

   2. If a line ends with a comment, two key words get crunched together and
   the resulting code dies:

   begin{a comment here}
      writeln(output,'a text there');

   the result is 'beginwriteln'.

technical notes

   The constant maxlength is the length after which the output lines will be
   broken at the next opportunity.

   The program can use the date and time to create random numbers.  These are
   computer system calls and are not likely to function correctly on other
   computers.  Simply delete the calls to disable them if you are having
   trouble compiling.  You can always set different seeds to start the random
   numbers by hand.

   The length of the output variables is set by procedure makerandom.  The
   first character must be a character, the remainder are a number for
   convenience.  A greater length maximizes the difficulty of interpreting
   the code, but if the identifier names get too long, there is a chance that
   the compiler will confuse two.  According to the first definition of
   Pascal (Jensen1975) "Identifiers denoting distinct objects must differ
   over their first 8 characters."  While this may no longer be true, some
   compilers may still be sensitive to it.  Therefore the names were set to 8
   long.  This gives 26e7 possible variables ...

*)
(* end module describe.pascru *)

(* begin module pascru.const *)
      maxlength = 50; (* the length after which the output lines will be
                         broken at the next opportunity *)
(* end module pascru.const *)

(* begin module string.const *)
      maxstring = 150; (* the maximum string *)
(* end module string.const version = 1.43; (@ of makman.p 2000 July 19 *)

(* begin module filler.const *)
      fillermax = 50; (* the size of the filler array for a string *)
(* end module filler.const *)

(* begin module datetime.const *)
   datetimearraylength = 19; (* length of dataarray for dates,
       It is just long enough to include the 4 digit year - solving the
       year 2000 problem:
1980/06/09 18:49:11
123456789 123456789
         1         2
*)
(* end module datetime.const version = 7.40; {of delmod.p 2000 Feb 18} *)

writing = true; {zzznnn} (* are we writing or testing? *)

type

(* begin module datetime.type *)
   (* array for dates *)
   datetimearray = packed array[1..datetimearraylength] of char;
(* end module datetime.type version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 1.43; (@ of makman.p 2000 July 19 *)

(* begin module filler.type *)
      (* the following is an array used to fill a string.
      it is convenient to have it much shorter than the maxstring, so that
      it is easy to fill the string using procedure fillstring.
      the user must declare the value of constant fillermax. *)
      filler = packed array[1..fillermax] of char;
(* end module filler.type version = 1.43; (@ of makman.p 2000 July 19 *)

(* begin module trigger.type *)
      trigger = record (* an object to be searched for *)
         seek: string; (* the characters looked for *)
         state: integer; (* how close to triggering we are *)
         skip: boolean; (* trigger not found- skip the line *)
         found: boolean (* the trigger was found *)
      end;
(* end module trigger.type version = 1.43; (@ of makman.p 2000 July 19 *)

(* begin module pascru.type *)
   identptr = ^ident; (* a pointer to an ident *)
   ident = record
      (* identifier name set up as a trigger *)
      t: trigger; (* a trigger for searching for the name *)
      newvar: string; (* the new variable name *)
      next: identptr; (* pointer to the next identifier *)
   end;
(* end module pascru.type *)

var
   pin, (* file used by this program *)
   pout, (* file used by this program *)
   pascrup, (* file used by this program *)
   note, (* file used by this program *)
   pvar (* file used by this program *)
     : text;

{zzzlll} linenumber: integer; (* count of line number for debugging *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 1.43; (@ of makman.p 2000 July 19 *)

(* begin module crash *)
procedure crash;
(* Crash the program by trying to open a nonexistant file.  This allows
tracing by the dbx program.  To use:  insert call into the halt program
or whereever a traceable stop is desired. *)
var
   bogus: text; (* boghous internal file *)
begin
  writeln(output,' program crash.');
  reset(bogus);
end;
(* end module crash from delmod *)

(* begin module package.trigger *)
(* ************************************************************************ *)

(* begin module clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer. *)
begin (* initializestring *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 4.41; (@ of prgmod.p 2000 June 24 *)

(* begin module copystring *)
procedure copystring(a: string; var b: string);
(* copy string a to b *)
var l: integer; (* index to the string *)
begin
   b.length := a.length;
   for l := 1 to a.length do b.letters[l] := a.letters[l]
end;
(* end module copystring *)

(* begin module equalstring *)
function equalstring(a, b: string): boolean;
(* Test for equality between two strings at current positions.
NOTE:  A compiler bug results if one directly tests this way:
   if thedefinition^.nametag = aname
The reason is completely not clear!  I showed that the parts of the
strings were identical, but the whole was not by this test.
For this reason it is *critical to test strings with equalstring.
*)
var   index: integer; (* index to both strings *)
      equal: boolean; (* are letters in a and b the same? *)
begin (* equalstring *)
      if a.length = b.length
      then begin
         index := 1;
         repeat
            equal := (a.letters[index] = b.letters[index]);
            index := succ(index)
         until (not equal) or (index > a.length);
         equalstring := equal
      end
      else equalstring := false
end; (* equalstring *)
(* end module equalstring *)

(* begin module writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module writestring version = 4.41; (@ of prgmod.p 2000 June 24 *)

(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler);
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
(*   fillstring(s, 'this-is-the-string                                ');
the two comments make it easy to line the characters up. also, for this
example, it was assumed that the length of filler as defined by the
constant fillermax was 50. *)
var
      length: integer; (* of the string without trailing blanks *)
      index: integer; (* of s *)
begin (* fillstring *)
      clearstring(s);
      length := fillermax;
      while (length > 1) and (a[length] = ' ') do length := pred(length);
      if (length = 1) and (a[length] = ' ') then begin
         writeln(output, 'fillstring: the string is empty');
crash;
         halt
      end;

      for index := 1 to length do s.letters[index] := a[index];
      s.length := length;
      s.current := 1
end; (* fillstring *)
(* end module filler.fillstring version = 4.41; (@ of prgmod.p 2000 June 24 *)

(* begin module filler.filltrigger *)
procedure filltrigger(var t: trigger;
                      a: filler);
(* fill the trigger t *)
begin (* filltrigger *)
         fillstring(t.seek,a)
end; (* fillstring *)
(* end module filler.filltrigger version = 4.41; (@ of prgmod.p 2000 June 24 *)

(* begin module trigger.proc *)
(* this module allows one to scan a series of characters, as from
an array or a file, and to "trigger" or detect a simple string
in the series.  the advantage of the trigger is that several triggers
can "observe" a stream of characters at once, each looking for a
different thing.
some other modules required: interact.const, interact.type *)

procedure resettrigger(var t: trigger);
(* reset the trigger to ground state *)
begin (* resettrigger *)
      with t do begin
         state := 0;
         skip := false;
         found := false
      end
end; (* resettrigger *)

procedure testfortrigger(ch: char; var t: trigger);
(* look at the character ch.
   if it is part of the trigger (at the current trigger state),
       then the trigger state goes higher.
   if it is not part of the trigger then the trigger state is reset,
      skip is true and one should skip onward to find the trigger.
   if the trigger is found, found is true.
1996 Sep 12: Bug found!  In the case of a trigger "ab", the program
used to miss it for situations like "aab".  This was because at the
first a it would step up.  Then it would see the second a and recognize
that was not part of ab.  It would fail to realize that it could be
the start of a new one.  The code now accounts for that possibility. *)
begin (* testfortrigger *)
      with t do begin
         state := succ(state);
{
writestring(list,seek);
writeln(list,'testfortrigger seek.letters[',state:1,']:',
seek.letters[state],' ch:',ch);
}
         if seek.letters[state] = ch
         then begin
            skip := false;
            if state = seek.length then found := true
                                   else found := false
         end
         else begin
            (* it failed.  But wait!  It could be the beginning
            of a NEW trigger string! *)
            if seek.letters[1] = ch then begin
               state := 1;
               skip := false;
               found := false
            end
            else begin (* reset trigger *)
               state := 0;
               skip := true;
               found := false
            end
         end
      end
end; (* testfortrigger *)
(* end module trigger.proc version = 4.41; (@ of prgmod.p 2000 June 24 *)

(* begin module skipblanks *)
procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
      while (thefile^ = ' ') and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
      while (thefile^ <> ' ') and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 4.41; (@ of prgmod.p 2000 June 24 *)

(* ************************************************************************ *)
(* end module package.trigger version = 4.41; (@ of prgmod.p 2000 June 24 *)

(* begin module package.datetime *)
(* ************************************************************************ *)
(* begin module getdatetime *)
procedure getdatetime(var adatetime: datetimearray);
(* get the date and time into a single array from the system clock.
   adatetime contains the date:
       1980/06/09 18:49:11
         ye mo da ho mi se
(year, month, day, hour, minute, second).
As of 2000 February 18, the Sun Pascal compiler requires a formatting
statement.  This statement allows the date to be generated in this
standard Delila format in a single call.  Information about the
formatting statement is available on the manual page for date in Unix.
If a computer does not have this method, see the 'oldgetdatetime' routine
in delmod.p (https://alum.mit.edu/www/toms/delila/delmod.html)
for some conversion code.
*)
begin
   date(adatetime,'%Y/%m/%d %H:%M:%S');
end;
(* end module getdatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module readdatetime *)
procedure readdatetime (var thefile: text; var adatetime: datetimearray);
(* read the date and time from the file *)
var
      index: integer; (* to the udatetime *)
      (* the following is an unpacked date time array, to avoid reading
      into a packed array.  reading into a packed array is not transportable *)
      udatetime: array[1..datetimearraylength] of char;
begin
      for index:=1 to datetimearraylength do read(thefile,udatetime[index]);
      pack(udatetime, 1, adatetime);
      if (adatetime[3]='/') and (adatetime[12]=':') then begin
         writeln(output,' old datetime (only 2 year digits) read: ',
                          adatetime:datetimearraylength);
      end;
end;
(* end module readdatetime version = 7.40; {of delmod.p 2000 Feb 18} *)


(* begin module timeseed *)
procedure addtoseed(var seed, power: real; c: char);
(* add the digit represented by c to the seed at the power position *)
var
      n: integer; (* the character represented by c *)
begin (* addtoseed *)
      power := power/10;
{
writeln(output,'addtoseed, c = ',c);
writeln(output,'addtoseed, ord(c) = ',ord(c));
}
      case c of
         ' ': begin
                 writeln(output,'timeseed: error in datetime');
                 halt;
              end;
         '0': n := 0;
         '1': n := 1;
         '2': n := 2;
         '3': n := 3;
         '4': n := 4;
         '5': n := 5;
         '6': n := 6;
         '7': n := 7;
         '8': n := 8;
         '9': n := 9
      end;
(*writeln(output,'timeseed number is [',n:1,']'); (@ debug *)
      seed := seed + power*n
end; (* addtoseed *)

procedure makeseed(adatetime: datetimearray; var seed: real);
(* convert adatetime to a real number in seed, reversed order *)
var
   power: real; (* a digit of the seed such as 0.01 *)
begin
   seed := 0.0;
   power := 1.0;
   addtoseed(seed, power, adatetime[19]);
   addtoseed(seed, power, adatetime[18]);
   (* : *)
   addtoseed(seed, power, adatetime[16]);
   addtoseed(seed, power, adatetime[15]);
   (* : *)
   addtoseed(seed, power, adatetime[13]);
   addtoseed(seed, power, adatetime[12]);
   (*   *)
   addtoseed(seed, power, adatetime[10]);
   addtoseed(seed, power, adatetime[9]);
   (* / *)
   addtoseed(seed, power, adatetime[7]);
   addtoseed(seed, power, adatetime[6]);
   (* / *)
   addtoseed(seed, power, adatetime[4]);
   addtoseed(seed, power, adatetime[3]);
   addtoseed(seed, power, adatetime[2]);
   addtoseed(seed, power, adatetime[1]);
end;

procedure orderseed(adatetime: datetimearray; var seed: real);
(* convert adatetime to a real number in seed, normal order *)
var
   power: real; (* a digit of the seed such as 0.01 *)
begin
   seed := 0.0;
   power := 1.0;
   addtoseed(seed, power, adatetime[1]);
   addtoseed(seed, power, adatetime[2]);
   addtoseed(seed, power, adatetime[3]);
   addtoseed(seed, power, adatetime[4]);
   (* / *)
   addtoseed(seed, power, adatetime[6]);
   addtoseed(seed, power, adatetime[7]);
   (* / *)
   addtoseed(seed, power, adatetime[9]);
   addtoseed(seed, power, adatetime[10]);
   (*   *)
   addtoseed(seed, power, adatetime[12]);
   addtoseed(seed, power, adatetime[13]);
   (* : *)
   addtoseed(seed, power, adatetime[15]);
   addtoseed(seed, power, adatetime[16]);
   (* : *)
   addtoseed(seed, power, adatetime[18]);
   addtoseed(seed, power, adatetime[19]);
end;

procedure timeseed(var seed: real);
(* read the computer date and time.  reverse the order of the digits
and put a decimal point in front.  this gives a fraction between
zero and one that varies quite quickly, and is always unique (if the
computer has sufficient accuracy).  it is to be used as a seed to
a random number generator. *)
var
   adatetime: datetimearray; (* a date and time *)
begin (* timeseed *)
   getdatetime(adatetime);
{
writeln(output,'timeseed: adatetime: ',adatetime);
}
   makeseed(adatetime, seed);
end; (* timeseed *)
(* end module timeseed version = 7.40; {of delmod.p 2000 Feb 18} *)

(* ************************************************************************ *)
(* end module package.datetime version = 7.40; {of delmod.p 2000 Feb 18} *)


(* begin module random *)
procedure random(var seed: real);
(* the seed value is used to generate a random number in the range 0 to 1.
the seed is then altered, to provide the seed of the next random number.
in other words, the seed is the random number.  to obtain the next random
number in the pseudo-random series, simply call random again with the same
seed.  note that the series will not be truely random, since the previous
value determines the next value.  this is "pseudo-random".

an initial seed of 4.0 is suggested.  note that this is not the first
random number.  you can use different seeds to get different pseudo-random
number series.

the expected mean and standard deviation of a flat distribution is
calculated using simple integrals:

       1
mean = / x dx = 0.5
       0
           1           2
variance = / (x - mean)  dx = 1/12;  stdev = sqrt(variance) = .2886751
           0

(the integral for the variance is solved easily by substituting y = x - mean.)

   the algorithm is derived from:
pascal programs for scientists and engineers, alan r. miller
sybex, berkeley ca (1981).
which in turn was an adaptation from hp-35 applications programs *)
const
      pi = 3.141592653589; (* a value used to help mix up the result *)
var
      intermediate: real; (* an intermediate result *)
begin (* random *)
      intermediate := seed + pi;
      intermediate := exp(5.0*ln(intermediate));
      seed := intermediate - trunc(intermediate)
end; (* random *)
(* end module random version = 1.06; (@ of rndseq 1983 oct 17 *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a line (as a string) from a file not using string calls.  this lets
one obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;
         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;
         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 1.43; (@ of makman.p 2000 July 19 *)

(* begin module pluckdigit *)
function pluckdigit(number, logplace:integer): char;
(* return the digit at the place value ('logplace') position of number.
example:
      pluckdigit(13625, 3) = 3
      pluckdigit(13625, 4) = 1
This routine was taken from module numberdigit in prgmod.p, but is
modified so as not to give the sign.  Instead it gives zeros
above the digits.  'myabsolute' replaced 'absolute', which
is apparently a keyword for GPC.  The name is kept for
to keep the code looking similar to its origin.
*)
var
      place: integer; (* the exponent of logplace *)
      count: integer; (* used to make place *)
      myabsolute: integer; (* the absolute value of number *)
      acharacter: char; (* the character to be returned *)
procedure digit;
(* extract a digit at the place position *)
var
      tenplace: integer; (* ten times place *)
      z: integer; (* an intermediate value *)
      d: integer; (* the digit extracted *)
begin (* digit *)
      tenplace:=10*place;
      z:=myabsolute-((myabsolute div tenplace)*tenplace);
      if place = 1
         then d:=z
         else d:= z div place;
      case d of
         0: acharacter:='0';
         1: acharacter:='1';
         2: acharacter:='2';
         3: acharacter:='3';
         4: acharacter:='4';
         5: acharacter:='5';
         6: acharacter:='6';
         7: acharacter:='7';
         8: acharacter:='8';
         9: acharacter:='9';
      end
end; (* digit *)
begin (* pluckdigit *)
      place:=1;
      for count:=1 to logplace do place:=10*place;
      if number=0 then begin
         acharacter:='0'
      end
      else begin
         myabsolute:=number;
         if myabsolute >= place
         then digit
         else acharacter := '0'
      end;
      pluckdigit:=acharacter
end; (* pluckdigit *)
(* end module pluckdigit *)

(* begin module copymanualpage *)
procedure copymanualpage(var fin, fout: text);
(* copy a delila manual page from fin to fout *)
var
   debe, deen: trigger; (* the describe begin and end module strings *)
   vebe, veen: trigger; (* the version begin and end module strings *)
   vt: trigger; (* the version trigger *)
   v: string; (* the version line *)

   aline, (* a line of text from fin *)
   anamebegin: string; (* the name of the module at the beginning
                          of the delman.describe *)
   gotten: boolean; (* true when a trigger is found *)
begin
     (*                      1         2         3   *)
(*                  12345678901234567890123456789012 *)
   filltrigger(vt ,'version =                       ');
   filltrigger(debe,'(@ begin module describe.       ');
   filltrigger(deen,'(@ end module describe.         ');
   filltrigger(vebe,'(@ begin module version *)      ');
   filltrigger(veen,'(@ end module version           ');
(* the @ signs prevent the makman program from detecting these
strings when it looks at its own source code!  But now we have
to fix it by putting the * there: *)
   debe.seek.letters[2] := '*';
   deen.seek.letters[2] := '*';
   vebe.seek.letters[2] := '*';
   veen.seek.letters[2] := '*';

   while not eof(fin) do begin

      (* pull out a line from the file *)
      getstring(fin,aline,gotten);
{write(output,'---');writestring(output,aline); writeln(output);}

      (* scan the line for both version and definition modules *)
      resettrigger(debe);
      resettrigger(vebe);
      with aline
      do while (current<=length) and
               (not debe.found) and (not vebe.found)
      do begin
         testfortrigger(letters[current],debe);
         testfortrigger(letters[current],vebe);

         (* only accept triggers at the beginning of the line *)
         if veen.found
         then if aline.current <> veen.seek.length
              then resettrigger(veen);
         if debe.found
         then if aline.current <> debe.seek.length
              then resettrigger(debe);
         current := succ(current)
      end;
{writeln(output,'vebe.found=',vebe.found);}
{writeln(output,'debe.found=',debe.found);}

      if vebe.found
      then begin (* grab the version line *)
         getstring(fin,v,gotten);
{
         writestring(fout,aline); writeln(fout);
         writestring(fout,v); writeln(fout);
}

         (* now verify that there really is a version on this line! *)
         resettrigger(vt);
         while (v.current <= v.length) and
               (not vt.found) do begin
            testfortrigger(v.letters[v.current],vt);
            v.current := succ(v.current);
         end;
{ ;writeln(output,'vt.found=',vt.found);}

         (* now verify that there really is a version module end!
            We allow one extra line in the module (for the origin
            date), and no more. *)
         resettrigger(veen);
         resettrigger(debe);
         while (not eof(fin))
                and (not veen.found) and (not debe.found)
         do begin
            getstring(fin,aline,gotten);
            while (aline.current <= aline.length) and
                  (not veen.found) and
                  (not debe.found) do begin
               testfortrigger(aline.letters[aline.current],veen);
               testfortrigger(aline.letters[aline.current],debe);

               (* only accept triggers at the beginning of the line *)
               if veen.found
               then if aline.current <> veen.seek.length
                    then resettrigger(veen);
               if debe.found
               then if aline.current <> debe.seek.length
                    then resettrigger(debe);

               aline.current := succ(aline.current);
            end;
{
            (* show the whole version module *)
            writestring(fout,aline); writeln(fout);
}
         end;

         if (not veen.found) or (debe.found) then begin
            writeln(fout,'{The end of the version module was not found for');
            writestring(fout,v); writeln(fout,'}');
            halt
         end;
      end

      else if debe.found then begin (* grab the entire definition *)

         (* display the version grabbed earlier at the end of the module *)
         (* new location: display at top *)
         if vt.found then begin
            writestring(fout,vebe.seek); writeln(fout);
            write(fout,'{');
            writestring(fout, v);
            writeln(fout,'}');
            writestring(fout,veen.seek); writeln(fout,' *', ')');
            writeln(fout);
            resettrigger(vt);
         end
         else writeln(fout,'{no "version =" string found}');

         (* dump out the first line found *)
         writestring(fout,aline);
         writeln(fout);

         (* seek the end of the module while writing it out *)
         resettrigger(deen);
         while (not eof(fin)) and (not deen.found) do begin
            getstring(fin,aline,gotten);
            resettrigger(deen);

            with aline
            do while (current<length) and
               (not deen.skip) and
               (not deen.found)
               do begin
                  (* only accept triggers at the beginning of the line *)
                  testfortrigger(letters[current],deen);
                  if deen.found
                  then if aline.current <> deen.seek.length
                       then resettrigger(deen);
                  current := current + 1
               end;

             (* copy the line out *)
             writestring(fout,aline);
             writeln(fout);
         end;

         (* die if there was no end of the module *)
         if eof(fin) and (not deen.found) then begin
            write(fout, '{ module ');
            writestring(fout,anamebegin);
            writeln(fout,' has no end}');
            halt
         end;

      end;
   end;
end;
(* end module copymanualpage version = 1.43; (@ of makman.p 2000 July 19 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 4.43; (@ of prgmod.p 2000 July 18 *)

(* ************************************ *)
(* ************************************ *)
(* ************************************ *)

(* begin module addgood *)
procedure addgood(var g: identptr; a: filler);
(* adds a word 'a' into the goodwords list g *)
begin
   new(g^.next);
   resettrigger(g^.next^.t);
   initializestring(g^.next^.t.seek);
   fillstring(g^.next^.t.seek, a );
   g:= g^.next;
end;
(* end module addgood *)

(* begin module addbad *)
procedure addbad(var g: identptr; a: filler; ranword: string);
(* adds a word onto the badwords list *)
begin
   new(g^.next);
   resettrigger(g^.next^.t);
   initializestring(g^.next^.t.seek);
   fillstring(g^.next^.t.seek, a );
   g^.next^.newvar:= ranword;
   g:= g^.next;
end;
(* end module addbad *)

(* begin module goodlist *)
procedure goodlist(var g, start, stop: identptr; var additional: text);
(* generate a linked list of pascal terms that must not
be changed.  Read additional ones from the additional file *)

var
   c: char; (* generic character *)
   f: filler; (* filler for addgood *)
   s: string; (* string version of filler *)
   sc: string; (* string version to compare to *)
   sp: string; (* string version to of the word program *)
   x: integer; (* index to filler array *)
   isitword: boolean; (* is it a word? *)

begin

(* and array begin const div do else end eof eoln if mod not
              of or procedure program record then type var while *)

   for x:= 1 to fillermax do begin
      f[x]:= ' ';
   end;

   isitword:= false;
   fillstring(sc, 'output');
   fillstring(sp, 'program');

   new(g);
   resettrigger(g^.t);
   initializestring(g^.t.seek);
   fillstring(g^.t.seek, 'and');
   start := g;

   (* these are the reserved words on page 212 of PASCAL user manual and
   report, ISO Pascal Standard; 4th edition *)

   addgood(g, 'array');
   addgood(g, 'begin');
   addgood(g, 'case');
   addgood(g, 'const');
   addgood(g, 'div');
   addgood(g, 'do');
   addgood(g, 'downto');
   addgood(g, 'else');
   addgood(g, 'end');
   addgood(g, 'file');
   addgood(g, 'for');
   addgood(g, 'function');
   addgood(g, 'goto');
   addgood(g, 'if');
   addgood(g, 'in');
   addgood(g, 'label');
   addgood(g, 'mod');
   addgood(g, 'nil');
   addgood(g, 'not');
   addgood(g, 'of');
   addgood(g, 'or');
   addgood(g, 'packed');
   addgood(g, 'procedure');
   addgood(g, 'program');
   addgood(g, 'record');
   addgood(g, 'repeat');
   addgood(g, 'set');
   addgood(g, 'then');
   addgood(g, 'to');
   addgood(g, 'type');
   addgood(g, 'until');
   addgood(g, 'var');
   addgood(g, 'while');
   addgood(g, 'with');
   addgood(g, 'forward');

   (* these are the standard identifiers on page 211 of PASCAL user manual
   and report, ISO Pascal Standard; 4th edition *)

   addgood(g, 'abs');
   addgood(g, 'arctan');
   addgood(g, 'boolean');
   addgood(g, 'char');
   addgood(g, 'chr');
   addgood(g, 'cos');
   addgood(g, 'dispose');
   addgood(g, 'eof');
   addgood(g, 'eoln');
   addgood(g, 'exp');
   addgood(g, 'false');
   addgood(g, 'get');
   addgood(g, 'input');
   addgood(g, 'integer');
   addgood(g, 'ln');
   addgood(g, 'maxint');
   addgood(g, 'new');
   addgood(g, 'odd');
   addgood(g, 'ord');
   addgood(g, 'output');
   addgood(g, 'pack');
   addgood(g, 'page');
   addgood(g, 'pred');
   addgood(g, 'put');
   addgood(g, 'read');
   addgood(g, 'readln');
   addgood(g, 'real');
   addgood(g, 'reset');
   addgood(g, 'rewrite');
   addgood(g, 'round');
   addgood(g, 'sin');
   addgood(g, 'sqr');
   addgood(g, 'sqrt');
   addgood(g, 'succ');
   addgood(g, 'text');
   addgood(g, 'true');
   addgood(g, 'trunc');
   addgood(g, 'unpack');
   addgood(g, 'write');
   addgood(g, 'writeln');

   (* other words that should be on the list *)
   addgood(g, 'e'); (* exponentiation, due to the parsing mechanism *)


   (* add program name and file names to the good list *)
   reset(pin);

   repeat
      for x:= 1 to fillermax do begin
         f[x]:= ' ';
      end;
      x:= 0;
      isitword:= false;
      read(pin,c);
      while (c <> '(') and (c<> ' ') do begin
         x:= x+1;
         f[x]:= c;
         read(pin,c);
         isitword:= true;
      end;
      if isitword then begin
          fillstring(s,f);
          if equalstring(sp,s) then begin
          end
          else addgood(g, f);
      end;
   until c= '(';

   while (c<> ')') do begin
      for x:= 1 to fillermax do begin
         f[x]:= ' ';
      end;
      x:= 0;
      isitword:= false;
      read(pin,c);
      while (c <> '(') and (c<> ' ') and (c <> ',') and (c<> ')') do begin
         x:= x+1;
         f[x]:= c;
         read(pin,c);
         isitword:= true;
      end;
      if isitword then begin
          fillstring(s,f);
          if equalstring(sc,s) then begin
          end
          else addgood(g, f);
      end;
   end;

   (* add additional words to the good list *)

   while not eof(additional) do begin
{write(output,'additional...');}
      for x:= 1 to fillermax do begin
         f[x]:= ' ';
      end;
      x:= 0;
      isitword:= false;
      c := '.';
      while (not eoln(additional)) and (c <> ' ') do begin
         read(additional,c);
         x:= x+1;
         f[x]:= c;
{write(output,c);}
         isitword:= true;
      end;
      readln(additional);
      if isitword then begin
          fillstring(s,f);
{write(output,' => ');}
{writestring(output,s);}
          if equalstring(sp,s) then begin
          end
          else addgood(g, f);
      end;
{writeln(output);}
   end;

   stop:= g;
end;
(* end module goodlist *)


(* begin module checkforgood *)
procedure checkforgood(var g: identptr; stop: identptr; c: char;
                           isitbegin: boolean);
(* makes the check for good words *)

begin
   repeat
      testfortrigger(c, g^.t);
      if (isitbegin = false) and (g^.t.state = 1) then begin
         resettrigger(g^.t);
      end;
      g := g^.next;
   until equalstring(g^.t.seek, stop^.t.seek);

   testfortrigger(c, g^.t);
   if (isitbegin = false) and (g^.t.state = 1) then begin
      resettrigger(g^.t);
   end;
end;

(* end module checkforgood *)

(* begin module checkforbad *)
procedure checkforbad(var b: identptr; stop: identptr; c: char;
                           isitbegin: boolean);
(* makes the check for bad words *)

begin
   while not equalstring(b^.t.seek, stop^.t.seek) do begin;
      testfortrigger(c, b^.t);
      if (isitbegin = false) and (b^.t.state = 1) then begin
         resettrigger(b^.t);
      end;
      b := b^.next;
   end;

   testfortrigger(c, b^.t);
   if (isitbegin = false) and (b^.t.state = 1) then begin
      resettrigger(b^.t);
   end;
end;

(* end module checkforbad *)

(* begin module makerandom *)
procedure makerandom(var s: string;
                     var seed: real);
(* pick a random letter from the alphabet of a-z (c) and
a random number between 0 and maxnumber (ranint).
Combine these into a string s having 8 characters.
Use the previous random number seed. *)
const
  maxcharacters = 8; (* biggest number for ranint *)
  alphabetsize = 26; (* size of alphabet *)
var
  d: integer; (* digit index to ranint *)
  maxnumber: integer; (* biggest number for ranint *)
  ranint: integer; (* random integer up to maxnumber *)
begin
   maxnumber := round(exp((maxcharacters-1)*ln(10)));
   clearstring(s);

   random(seed);
   s.letters[1] := chr(trunc(alphabetsize*seed + ord('a')));

   random(seed);
   ranint:= round(seed * maxnumber);

   for d := 0 to maxcharacters-1  do begin
      s.letters[d+2] := pluckdigit(ranint,d)
      (* so what if the digits go in backwards??? *)
   end;
   s.length := maxcharacters;
{
write(output,'ranint = ',ranint:6,' "');
writestring(output,s);
writeln(output,'"');
halt;
}
end;
(* end module makerandom *)

(* begin module dumpchar *)
procedure dumpchar(var pout: text; c: char; var linelength: integer;
                       isitstring: boolean; var outputspaces: integer);
(* dump the character c out to the pout file and count it
in linelength.  Break line if it is time and it is not a string.
Do not output a space if there were previous spaces, ie if
outputspaces > 0. *)
begin
   if (linelength >= maxlength) and
      ((c = ' ') or (c = ',')) and
      (isitstring = false) then begin
      (* new code by TDS 2000 July 18 to break lines *)
      (* preserve the last character *)
if writing then {zzznnn }
      if c <> ' ' then write(pout, c);
write(pout,'{',linelength:1,'}'); {zzz}
      writeln(pout);
linenumber := succ(linenumber); {zzzlll}
write(pout,'{',linenumber:1,'}'); {zzz}
      linelength := 0;
      outputspaces := 0;
   end
   else begin
      if (c <> ' ') or isitstring then begin
if writing then {zzznnn }
         write(pout, c);
         linelength := succ(linelength);
         outputspaces := 0;
      end
      else begin
         if (outputspaces = 0) and (linelength <> 0)
         then begin
            (* this is the ONLY place that a space is output! *)
if writing then {zzznnn }
            write(pout, c);
            linelength := succ(linelength);
            outputspaces := succ(outputspaces);
         end
(*
         else begin
            write(pout, '{',outputspaces:1,'}'); {zzz  extra 2 chars: }
            linelength := succ(linelength) + 2; {zzz remove '2' when debugged}
         end;
*)

      end;
   end
end;
(* end module dumpchar *)

(* begin module dumpstring *)
procedure dumpstring(var pout: text; s: string; var linelength: integer;
                     var outputspaces: integer);
(* dump the string s to file pout, update linelength by the length
of the string *)
begin
if writing then {zzznnn }
   writestring(pout, s);
   linelength := linelength + s.length;
   outputspaces := 0;
end;
(* end module dumpstring *)

(* begin module dumpname *)
procedure dumpname(var pout: text; s: string;
                   var linelength: integer;
                   var outputspaces: integer);
(* dump the string s to pout *)
begin
if writing then {zzznnn }
   writestring(pout,s);
   linelength := linelength + s.length;
   outputspaces := 0;
end;
(* end module dumpname *)

(* begin module badlist *)
procedure badlist(var b, start, stop: identptr;
                  var isitfirst: boolean; f: filler;
                  var newvar: string;
                  var seed: real;
                  var linelength: integer;
                  var outputspaces: integer);
(* generates a linked list of variable names; write
current name out at same time (zzz poor design!) *)
var
   isitbad: boolean; (* is it already in the list *)
   s: string; (* the filler converted into a string *)
   ranword: string; (* random word *)
begin

   fillstring(s,f);
   isitbad:= false;

   makerandom(ranword, seed);

   if isitfirst = true then begin
      new(b);
      resettrigger(b^.t);
      initializestring(b^.t.seek);
      fillstring(b^.t.seek,f);
      copystring(ranword, b^.newvar);
      dumpname(pout, b^.newvar, linelength, outputspaces);
      start := b;
      stop:= b;
   end
   else begin
      b := start;
      while b <> stop do begin
         if equalstring(s,b^.t.seek) then begin
            copystring(b^.newvar, newvar);
            isitbad:= true;
            dumpname(pout,b^.newvar, linelength, outputspaces);
         end;
         b:= b^.next;
      end;
      if equalstring(s,b^.t.seek) then begin
         newvar:= b^.newvar;
         isitbad:= true;
         dumpname(pout,b^.newvar, linelength, outputspaces);
      end;

      if isitbad = false then begin
         while b <> stop do begin
            if ranword = b^.newvar then begin
               while ranword = b^.newvar do begin
                  makerandom(ranword, seed);
               end;
            end;
            b:= b^.next;
         end;

         if ranword = b^.newvar then begin
            while ranword = b^.newvar do begin
               makerandom(ranword, seed);
            end;
         end;

         addbad(b,f,ranword);
         dumpname(pout,b^.newvar, linelength, outputspaces);
         stop:= b;
      end;
   end;
   isitfirst := false;
   b := start;
end;
(* end module badlist *)

(* begin module notword *)
procedure notword(var g, b: identptr; startg, stopg: identptr;
                  var startb, stopb:  identptr;
                  var isitbegin, isitgood, isitbad: boolean;
                  var wordindex: integer;
                  var wordin: filler;
                  var isitfirst: boolean;
                  var newvar: string;
                  var seed: real;
                  var linelength: integer;
                  var outputspaces: integer);
(* if the char is not a letter do these things.
zzz do what things?
Well for one, write the word out to the (ugh!!) global
pout. *)
var
   x : integer; (* index to a for loop *)

begin

   isitbegin:= true;
   g := startg;
   b := startb;
   isitgood:= false;
   isitbad:= false;
   wordindex:= 0;

   repeat
       if g^.t.found = true then begin
         isitgood:= true;
         dumpstring(pout, g^.t.seek, linelength, outputspaces);
       end;
       resettrigger(g^.t);
       g := g^.next;
   until equalstring(g^.t.seek, stopg^.t.seek);

   if g^.t.found = true then begin
      isitgood:= true;
      dumpstring(pout, g^.t.seek, linelength, outputspaces);
    end;

   if isitgood = false then begin
      badlist(b, startb, stopb, isitfirst, wordin, newvar, seed, linelength,
              outputspaces);
   end;

   for x:= 1 to fillermax do begin
      wordin[x]:= ' ';
   end;

   resettrigger(g^.t);
   g := startg;
end;
(* end module notword *)

(* begin module findvar *)
procedure findvar(var c, prev: char;
                  var goodwords: identptr;
                  var startg: identptr;
                  var stopg: identptr;
                  var badwords: identptr;
                  var startb: identptr;
                  var stopb: identptr;
                  var isitbegin: boolean;
                  var wordin: filler;
                  var wordindex: integer;
                  var isitgood: boolean;
                  var isitbad: boolean;
                  var isitfirst: boolean;
                  var newvar: string;
                  var wasprevsymbol: boolean;
                  var wasitnumb: boolean;
                  var seed: real;
                  var linelength: integer;
                      isitstring: boolean;
                  var outputspaces: integer);

(* This procedure finds variable names to be switched.  c is the current
character read in, prev is the previous character read in.  UNFORTUNATELY
(bad program design) results are written to the global file pout. *)

begin

   if isitbegin and ((c= '0') or (c= '1') or (c= '2') or (c= '3') or
                     (c= '4') or (c= '5') or (c= '6') or (c= '7') or
                     (c= '8') or (c= '9'))
   then begin
      dumpchar(pout, c, linelength, isitstring, outputspaces);
      wasitnumb:= true;
   end
   else begin
      if (c = ' ') and (not wasprevsymbol) and (not wasitnumb)
         and (prev <> ' ') then begin
         notword(goodwords, badwords, startg, stopg, startb, stopb,
                 isitbegin, isitgood, isitbad, wordindex,wordin,
                 isitfirst, newvar, seed, linelength, outputspaces);
         dumpchar(pout, c, linelength, isitstring, outputspaces);
      end
      else begin
         if (c<> '+') and (c<> '-') and (c<> '*') and (c<> '/')
         and (c<> '<') and (c<> '=') and (c<> '>') and (c<> ':') and
         (c<> '(') and (c<> ')') and (c<> '[') and (c<> ']') and
         (c<> ',') and (c<> '.') and (c<> ';') and (c<> '''') and (c <> ' ')
         then begin

            if (c = '^') then begin
               if (isitbegin = true) then begin
                  dumpchar(pout, c, linelength, isitstring, outputspaces);
               end
               else begin
                  notword(goodwords, badwords, startg, stopg, startb, stopb,
                  isitbegin, isitgood, isitbad, wordindex,wordin,
                  isitfirst, newvar, seed, linelength, outputspaces);
                  dumpchar(pout, c, linelength, isitstring, outputspaces);
                  wasprevsymbol:= true;
                  wasitnumb:= false;
               end;
            end
            else begin
               wasprevsymbol:= false;
               wasitnumb:= false;
               goodwords := startg;
               wordindex := wordindex +1;
               wordin[wordindex]:= c;
               checkforgood(goodwords, stopg, c, isitbegin);
               isitbegin:= false;
            end;
            (* dumpchar(pout, c, linelength, isitstring, outputspaces); *)
         end
         else begin
            if not wasitnumb then begin
               if (c <> ' ') and wasprevsymbol then begin
                  wasprevsymbol:= true;
                  wasitnumb:= false;
                  dumpchar(pout, c, linelength, isitstring, outputspaces);
               end;

               if (c <> ' ') and not wasprevsymbol then begin
                  wasprevsymbol:= true;
                  wasitnumb:= false;
                  if prev <> ' ' then begin
                     notword(goodwords, badwords, startg, stopg, startb, stopb,
                             isitbegin, isitgood, isitbad, wordindex,wordin,
                             isitfirst, newvar, seed, linelength, outputspaces);
                     dumpchar(pout, c, linelength, isitstring, outputspaces);
                  end
                  else dumpchar(pout, c, linelength, isitstring, outputspaces);
               end;
            end
            else begin
               wasprevsymbol:= true;
               wasitnumb:= false;
               dumpchar(pout, c, linelength, isitstring, outputspaces);
            end;
         end;

         if eoln(pin) and (c <> ' ') and (not wasprevsymbol) then begin
            notword(goodwords, badwords, startg, stopg, startb, stopb,
                    isitbegin, isitgood, isitbad, wordindex,wordin,
                    isitfirst, newvar, seed, linelength, outputspaces);
            dumpchar(pout, ' ', linelength, isitstring, outputspaces);
            (* dumpchar(pout, ' ', linelength, isitstring, outputspaces);
               use this along with a counter that will
               stop the line at 70 characters *)
            c:= ' ';
         end;

      end;
      if (c = ' ') and wasprevsymbol
      then dumpchar(pout, c, linelength, isitstring, outputspaces);
      prev:= c;
   end;
   if eoln(pin) and wasitnumb
   then dumpchar(pout, ' ', linelength, isitstring, outputspaces);
end;
(* end module findvar *)


(* begin module pascru.themain *)
procedure themain(var pin, pout, pascrup, note, pvar: text);
(* the main procedure of the program *)
var
   c: char; (* current character *)
   prev: char; (* previous character *)
   prev2: char; (* previous character to the previous character *)
   linelength: integer; (* current length of the output line *)

   isitcurly:    boolean; (* is there a  comment with curly brackets *)
   isitparens:   boolean; (* is there a  comment with parentheses *)
   maybeparens:   boolean; (* is there a  left parens *)
   wasitparens:   boolean; (* was it a parens comment *)
   wasitcurly:   boolean; (* was it a curly comment *)

   isitvar: boolean; (* is there a var that needs to be changed *)
   isitnoes: boolean; (* are there no extra spaces *)

   isitstring: boolean; (* is it inside a string *)

(* begin module change-variable variables *)
   goodwords: identptr; (* standard pascal words *)
   startg: identptr; (* begin of goodwords linked list *)
   stopg: identptr; (* end of goodwords linked list *)

   badwords: identptr; (* not standard pascal words *)
   startb: identptr; (* begin of badwords linked list *)
   stopb: identptr; (* end of badwords linked list *)

   isitbegin: boolean; (* is it the beginning of the word *)

   wordin: filler; (* word read in *)
   wordindex: integer; (* index to filler *)

   isitgood: boolean; (* is it     one of the pascal words *)
   isitbad: boolean; (*  is it not one of the pascal words *)

   isitfirst: boolean; (* is the the first word on the list *)

   newvar: string; (* the new variable name *)

   wasprevsymbol: boolean; (* was the previous character a symbol *)
   wasitnumb: boolean; (* was the previous character a numb *)

(* end module change-variable variables *)

   outputspaces: integer; (* count of spaces just output *)

   (* parameters *)
   parameterversion: real; (* parameter version number *)
   seed: real; (* number for random generator *)
   domanualpage: char; (* do the manual page if it is 'm' *)

begin

   writeln(output,'pascru ',version:4:2);

   reset(pin);
   rewrite(pout);

   (* read first part of parameters *)
   reset(pascrup);
   readln(pascrup, parameterversion);
   if parameterversion < updateversion then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

linenumber := 1; {zzzlll}
   readln(pascrup, seed);
   if (seed <= 0.0) or (seed >= 1.0)
   then timeseed(seed);

   readln(pascrup, domanualpage);
   if domanualpage = 'm' then begin
      copymanualpage(pin,pout);
      writeln(pout);
   end;

   reset(note);
   if not eof(note) then begin
      while not eof(note) do copyaline(note, pout);
      writeln(pout);
   end;

   reset(pin);
{zzzttt}
   goodlist(goodwords, startg, stopg, pascrup);
   reset(pin);

   isitcurly:= false;
   isitparens:= false;
   maybeparens:= false;
   wasitparens:= false;
   wasitcurly:= false;

   isitvar:= false;
   isitstring:= false;

   prev:= ' ';
   linelength:= 0;
   isitbegin:= true;
   goodwords := startg;
   wordindex:= 0;

   isitgood:= false;
   isitbad:= false;
   isitfirst:= true;
   wasprevsymbol:= false;
   wasitnumb:= false;

   outputspaces := 0;

   while not eof(pin) do begin
      isitnoes:= false;
      while not eoln(pin) do begin
         read(pin,c);

         (* determine if it is inside a quotestring *)
         if (isitcurly = false) and (isitparens = false) then begin
            if c = '''' then isitstring := not isitstring
         end;

         (* remove all comments *)
         if (isitstring = false) then begin
            if (isitcurly = false) then begin
               if (prev = '(' ) and (c = '*') then begin
                  isitparens:= true;
                  maybeparens:= false;
                  wasprevsymbol:= true;
               end;
               if (prev = '*' ) and (c = ')') then begin
                  isitparens:= false;
                  wasitparens:= true;
               end;
               if c = '(' then begin
                  maybeparens := true;
                  isitnoes := true;
               end;
            end;

            if (maybeparens = true) and (c = '(') and
            (isitparens = false) and (isitcurly = false)
            and (wasitcurly = false) and (prev <> '}') and
            (c <> '}')  then begin
            end;

            if (isitparens = false) then begin
               if (c = '{') then begin
                  isitcurly:= true;
               end;
               if  (c = '}') then begin
                  isitcurly:= false;
                  wasitcurly:= true;
               end;
            end;
         end;

         if (isitstring = true) then begin
            if maybeparens = false then begin
               dumpchar(pout, c, linelength, isitstring, outputspaces);
               wasprevsymbol:= true;
            end;
         end;
        (*  else begin *)

         if (c <> ')') and (isitparens = false) then wasitparens:= false;
         if (c <> '}') and (isitcurly = false) then wasitcurly:= false;

         (* remove all indentation *)
         if (c = ' ') and (isitnoes = false) then begin
         end

         else begin
            if (isitparens = false) and (isitcurly = false) and
               (maybeparens = false) and
               (wasitparens = false) and (wasitcurly = false)
               and (isitstring = false)
               then begin
               isitnoes:= true;

               findvar(c, prev, goodwords, startg, stopg, badwords, startb,
                       stopb, isitbegin, wordin, wordindex, isitgood,
                       isitbad, isitfirst, newvar, wasprevsymbol,
                       wasitnumb, seed, linelength, isitstring, outputspaces);
            end;

            if wasitparens then begin
                isitnoes:= true;
                maybeparens:= false;
            end;

            if wasitcurly then begin
                isitnoes:= true;
                maybeparens:= false;
                if wordin[1] = ' ' then wasprevsymbol:= true;
            end;

            if (maybeparens = true) and (c <> '(') and (c <> '*') and
            (isitparens = false) and (isitcurly = false)
            and (wasitcurly = false) and
            (prev <> '}') and (c <> '}')  then begin

               findvar(prev, prev2, goodwords, startg, stopg, badwords, startb,
                       stopb, isitbegin, wordin, wordindex, isitgood,
                       isitbad, isitfirst, newvar, wasprevsymbol,
                       wasitnumb, seed, linelength, isitstring, outputspaces);

               findvar(c, prev, goodwords, startg, stopg, badwords, startb,
                       stopb, isitbegin, wordin, wordindex, isitgood,
                       isitbad, isitfirst, newvar, wasprevsymbol,
                       wasitnumb, seed, linelength, isitstring, outputspaces);

               isitnoes:= true;
               maybeparens:= false;
            end;

            if (c = '(') and (prev = '(') and (isitcurly = false)
            and (isitparens = false) then begin

               findvar(prev, prev2, goodwords, startg, stopg, badwords, startb,
                       stopb, isitbegin, wordin, wordindex, isitgood,
                       isitbad, isitfirst, newvar, wasprevsymbol,
                       wasitnumb, seed, linelength, isitstring, outputspaces);

               findvar(c, prev, goodwords, startg, stopg, badwords, startb,
                       stopb, isitbegin, wordin, wordindex, isitgood,
                       isitbad, isitfirst, newvar, wasprevsymbol,
                       wasitnumb, seed, linelength, isitstring, outputspaces);

               isitnoes:= true;
               maybeparens:= false;
               c:= ' ';
            end;

            if eoln(pin) and (c = '(') then begin

               findvar(c, prev, goodwords, startg, stopg, badwords, startb,
                       stopb, isitbegin, wordin, wordindex, isitgood,
                       isitbad, isitfirst, newvar, wasprevsymbol,
                       wasitnumb, seed, linelength, isitstring, outputspaces);
               isitnoes:= true;
               maybeparens:= false;
            end;

         end;
         prev2:= prev;
         prev:= c;
      end;

      (* condense the code *)
      if (isitcurly = false) and (isitparens = false)  then begin
         c := ' ';

         findvar(c, prev, goodwords, startg, stopg, badwords, startb,
                 stopb, isitbegin, wordin, wordindex, isitgood,
                 isitbad, isitfirst, newvar, wasprevsymbol,
                 wasitnumb, seed, linelength, isitstring, outputspaces);
      end;
      readln(pin);
   end;
   writeln(pout);

   rewrite(pvar);
   badwords:= startb;
   repeat
      writestring(pvar, badwords^.newvar);
      write(pvar,' ');
      writestring(pvar, badwords^.t.seek);
      writeln(pvar);
      badwords:= badwords^.next;
   until badwords = nil;

end;
(* end module pascru.themain *)

begin
   themain(pin, pout, pascrup, note, pvar);
1: end.
