program reffea(reffeap, db, reffeatures, output);
(* reffea: extract references from GenBank entries as lister features

  Tom Schneider
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.02; (* of reffea.p 1997 January 12
origin 1997 January 12 *)
(* end module version *)

(* begin module describe.exon *)
(*
name
   reffea: extract references from GenBank entries as lister features

synopsis
   reffea(reffeap: in, db: in,
       reffeatures: out output: out)

files
   reffeap: parameters to control the program, one per line:

     1:  limit region somehow? zzz would need
     to have a list of accession numbers and ranges.
     need rules about how overlaps are handled
     (include only whole or with partials)

   db: a set of GenBank entries containing references:

REFERENCE   5  (bases 38597 to 38672)

   reffeatures: Locations of references in the format for the Lister
      program:

define "J02459.ref.5" "-" "<]" "[>" 0 [length-1]
@ J02459 [start] 1 "J02459.ref.5" ""

      where [length-1] is the length of the reference, minus 1 and
      [start] is the first base of the reference.

   output: messages to the user

description

   Locate references on a lister map.

examples

documentation

see also
   lister.p, exon.p

author
   Thomas Dana Schneider

bugs

   The code was created from exon.p and can be trimmed a lot.
   The parameter file contains stuff that is unnecessary
   (it's identical to exonp at the moment)

   a required parameter: version number

   a required parameter: references to include
     (in a list that ends in 0)
     or maybe better:  reference ranges to include or exclude
    include by +, exclude by -.

   parameter: 

technical notes

*)
(* end module describe.exon *)

(* begin module interact.const *)
      maxstring = 150; (* the maximum string *)
(* end module interact.const version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module my.filler.const *)
      fillermax = 20; (* the size of the filler array for a string *)
(* end module my.filler.const *)
(* from filler.const version = 4.13; (@ of prgmod.p 1994 sep 5 *)
 
type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module filler.type *)
      (* the following is an array used to fill a string.
      it is convenient to have it much shorter than the maxstring, so that
      it is easy to fill the string using procedure fillstring.
      the user must declare the value of constant fillermax. *)
      filler = packed array[1..fillermax] of char;
(* end module filler.type version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module trigger.type *)
      trigger = record (* an object to be searched for *)
         seek: string; (* the characters looked for *)
         state: integer; (* how close to triggering we are *)
         skip: boolean; (* trigger not found- skip the line *)
         found: boolean (* the trigger was found *)
      end;
(* end module trigger.type version = 4.16; (@ of prgmod.p 1996 August 12 *)

var
   reffeap, db, reffeatures:
   text; (* files used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module interact.writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module interact.writestring version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler);
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
(*   fillstring(s, 'this-is-the-string                                ');
the two comments make it easy to line the characters up. also, for this
example, it was assumed that the length of filler as defined by the
constant fillermax was 50. *)
var
      length: integer; (* of the string without trailing blanks *)
      index: integer; (* of s *)
begin (* fillstring *)
      clearstring(s);
      length := fillermax;
      while (length > 1) and (a[length] = ' ') do length := pred(length);
      if (length = 1) and (a[length] = ' ') then begin
         writeln(output, 'fillstring: the string is empty');
         halt
      end;

      for index := 1 to length do s.letters[index] := a[index];
      s.length := length;
      s.current := 1
end; (* fillstring *)
(* end module filler.fillstring version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module filler.filltrigger *)
procedure filltrigger(var t: trigger;
                      a: filler);
(* fill the trigger t *)
begin (* filltrigger *)
         fillstring(t.seek,a)
end; (* fillstring *)
(* end module filler.filltrigger version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module trigger.proc *)
(* this module allows one to scan a series of characters, as from
an array or a file, and to "trigger" or detect a simple string
in the series.  the advantage of the trigger is that several triggers
can "observe" a stream of characters at once, each looking for a
different thing.
some other modules required: interact.const, interact.type *)

procedure resettrigger(var t: trigger);
(* reset the trigger to ground state *)
begin (* resettrigger *)
      with t do begin
         state := 0;
         skip := false;
         found := false
      end
end; (* resettrigger *)

procedure testfortrigger(ch: char; var t: trigger);
(* look at the character ch.
   if it is part of the trigger (at the current trigger state),
       then the trigger state goes higher.
   if it is not part of the trigger then the trigger state is reset,
      skip is true and one should skip onward to find the trigger.
   if the trigger is found, found is true. *)
begin (* testfortrigger *)
      with t do begin
         state := succ(state);

{
            writestring(output,seek);
            writeln(output,'testfortrigger seek.letters[',state:1,']:"',
                           seek.letters[state],'" ch:"',ch,'"');
writeln(output,'ord(ch) = ',ord(ch):1);
writeln(output,'ord(seek.letters[state]) = ',ord(seek.letters[state]):1);
}

(*         if debugging then begin
            writestring(list,seek);
            writeln(list,'testfortrigger seek.letters[',state:1,']:',
                           seek.letters[state],' ch:',ch);
         end;*)

         if seek.letters[state] = ch
         then begin
            skip := false;
            if state = seek.length then found := true
                                   else found := false
         end
         else begin (* reset trigger *)
            state := 0;
            skip := true;
            found := false
         end
      end
end; (* testfortrigger *)
(* end module trigger.proc version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module copyline *)
procedure copyline(var fin, fout: text);
(* copy a line from file fin to file fout
but DO NOT CARRIAGE RETURN on the fout.
Carriage return on the fin. *)
begin (* copyline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
end; (* copyline *)
(* end module copyline version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module skipblanks *)
procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
      while (thefile^ = ' ') and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
      while (thefile^ <> ' ') and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module grabtoken *)
procedure grabtoken(var thefile: text; var thestring: string);
(* skip any blanks and then grab the next token from the file *)
var
   c: char; (* a character in thefile *)
   done: boolean; (* done finding the name *)
begin
   skipblanks(thefile);
   done := false;
   with thestring do begin
      length := 0;
      while not done do begin
         if eoln(thefile)
         then done := true
         else begin
            read(thefile,c);
            if c = ' '
            then done := true
            else begin
               length := succ(length);
               letters[length] := c;
            end
         end
      end
   end
end;
(* end module grabtoken version = 4.16; (@ of prgmod.p 1996 August 12 *)

(* begin module exon.themain *)
procedure themain(var reffeap, db, reffeatures: text);
(* the main procedure of the program *)
const
   exonmax = 10000; (* largest number of exons the program can handle *)
type
   exonrecord = record
      (* a record of an exon *)
      thestart: integer; (* the start *)
      thestop:  integer; (* the stop *)
      cds: integer; (* which cds of the locus it is in *)
      complement: boolean; (* whether it is on the complementary strand *)
      getthestart: boolean; (* whether to get the start *)
      getthestop: boolean; (* whether to get the stop *)
      printasexon: boolean; (* if true, the start or stop represent an exon.
         otherwise it must be an intron. *)
   end;
var
   accession: trigger; (* trigger to find the ACCESSION pattern *)
   accname: string; (* the current accession name *)
   allowalternative: boolean; (* allow alternate splicing *)
   atcomplement: boolean; (* we are at a complement *)
   c: char; (* a character in db *)
   cds: trigger; (* trigger to find the CDS pattern *)
   cdscount: integer; (* count of cds in current entry *)
   closeparen: trigger; (* trigger to find the end of the join pattern *)
   complement: trigger; (* trigger to find the end of the complement pattern *)
   completeunit: boolean; (* true when a unit is complete *)
   comma: trigger; (* trigger to find the end of the comma pattern *)
   debug: boolean; (* true when debugging the program *)
   features: trigger; (* trigger to find the ACCESSION pattern *)
   getexons: boolean; (* true if we are to get exon features *)
   getintrons: boolean; (* true if we are to get intron features *)
   dotcount: integer; (* count number of dots in a unit of a CDS *)
   e: integer; (* index for fealist *)
   entryend: trigger; (* a mark for the end of an entry *)
   exon: trigger; (* trigger to find the exon feature *)
   intron: trigger; (* trigger to find the intron feature *)
   fealist: array[1..exonmax] of exonrecord; (* features in a locus *)
   feacount: integer; (* current number of features listed in fealist *)
   firstnumber: boolean; (* true when the first number is about to be read *)
   geneprediction: trigger; (* a locus contains non-experimental junk *)
   good: boolean; (* the end point is good unless it is found to be outside
     the sequence *)
   hold: integer; (* for reversing start and stop when atcomplement is true *)
   join: trigger; (* trigger to find the join pattern *)
   killentry: boolean; (* kill entries that contain bad phrases *)
   killputative: boolean; (* kill putative entries *)
   killnotexperimental: boolean; (* kill not_experimental entries *)
   killgeneprediction: boolean; (* kill gene prediction entries *)
   killunpublished: boolean; (* kill Unpublished entries *)
   killpseudo: boolean; (* kill pseudo containing entries *)
   length: integer; (* length of an exon *)
   linenumber: integer; (* count of the lines in db *)
   locus: trigger; (* trigger to find the LOCUS pattern *)
   locuscount: integer; (* count of locus *)
   noends: boolean; (* true when printing ends of the CDS *)
   putative: trigger; (* a locus contains non-experimental junk *)
   notexperimental: trigger; (* a locus contains non-experimental junk *)
   numbers: set of '0'..'9';
   organism: trigger; (* trigger to find organism key word *)
   printon: boolean; (* if true, print the data passing by *)
   relorient: integer; (* relative orientation of site to piece *)
   start: integer; (* the start base of an exon *)
   stop: integer; (* the stop base of an exon *)
   theAfromrange, theAtorange: integer; (* from/to range: acceptor delila inst *)
   theDfromrange, theDtorange: integer; (* from/to range: donor delila inst *)
   totalfeacount: integer; (* count of exons in entire db *)
   totalentrycount: integer; (* count of entrys in entire db *)
   unitcount: integer; (* true when the end of a  unit has been found *)
   unitend: boolean; (* true when the end of a  unit has been found *)
   infeatures: boolean; (* true if inside feature secion of an entry *)
   unpublished: trigger; (* a locus contains junk *)
   pseudo: trigger; (* a locus contains pseudo genes *)

   reference: trigger; (* trigger to find the REFERENCE pattern *)
   refnumber: integer; (* number of a reference *)
{zzz}

procedure grabspecies(var db, f: text);
(* get the species name from db into f. *)
var
   c: char; (* a character read from db *)
begin
{
writeln(output,'grabspecies');
}
   resettrigger(organism);
   reset(db);
   while (not eof(db)) and (not organism.found) do begin
      if eoln(db)
      then readln(db)
      else begin
         read(db, c);
         testfortrigger(c, organism);
         if organism.found then begin
            skipblanks(db);
            read(db,c);
            write(f,c,'.');
            skipnonblanks(db);
            skipblanks(db);
            copyline(db,f);
         end;
      end;
   end;
   if not organism.found then begin
      writeln(output,'ORGANISM line missing in db');
      halt
   end;
end;

procedure nextline;
(* move to the next line *)
begin
  if eoln(db) then begin
     readln(db);
{writeln(output); }
     linenumber := succ(linenumber);
     if debug then if printon then writeln(output);
  end
end;

procedure nextcharacter(var f: text; var c: char);
(* read the next character c from file f *)
begin
    read(f,c);
    if debug then if printon then write(output,c);
end;

procedure readaninteger(var f: text; first: char; var n: integer);
(* read an integer from f, protecting ourselves from stupidity.
first is the first character of the number *)
begin
{write(output,'readaninteger: ');}
   n := ord(first)- ord('0');
   while f^ in numbers do begin
      nextcharacter(f, c);
      n := 10*n + ord(c)- ord('0');
   end;
{writeln(output,' becomes ',n:1);}
end;

procedure sign(var f: text; x: integer);
(* write the sign of x *)
begin
   if x < 0 then write(f,'-')
            else write(f,'+');
end;

procedure printexons;
(* Print the exons listed in the fealist, then clear the list. *)
var
   currentexons: integer; (* number of exons printed right now *)
   e: integer; (* index for fealist *)
   firstexon: integer; (* the first exon to print *)
   lastexon: integer; (* the last exon to print *)
procedure thename(var f: text);
(* write the exon name to file f *)
begin
   writestring(f,accname);
   write(f, '.CDS.');
   write(f, start:1,'-',stop:1);
end;
begin
if debug then if eoln(db) then writeln(output);
if debug then writeln(output);
if debug then write(output,'****************************');
if debug then writeln(output, '---> ', feacount:4,' ',start:4,' ',stop:4,' ',killentry);


   firstexon := 1;
   lastexon := feacount;

   currentexons := lastexon - firstexon + 1;
   if (currentexons > 0) and (not killentry) then begin
      totalfeacount := totalfeacount + currentexons;
      totalentrycount := totalentrycount + 1;
      for e := firstexon to lastexon do begin
         start := fealist[e].thestart;
         stop  := fealist[e].thestop;

         if fealist[e].printasexon then begin
            if fealist[e].complement
            then length := start - stop + 1
            else length := stop - start + 1;
         end;

         if fealist[e].complement then relorient := -1
                                   else relorient := +1;

         if fealist[e].printasexon then begin

            (* write reffeatures file here with this format:
               define "J02833exon467-615" "-" "<]" "[>" 0 148
            *)
            write(reffeatures, 'define "');
            thename(reffeatures);
            write(reffeatures, '" "-" "<]" "[>" 0 ',(length-1):1);
            writeln(reffeatures);

            (* call it immediately;
               @ J02833 467 +1 "J02833exon467-615" ""
            *)
            write(reffeatures, '@ ');
            writestring(reffeatures,accname);
            write(reffeatures, ' ',start:1,' ',relorient:1,' "');
            thename(reffeatures);
            write(reffeatures, '" ""');
            writeln(reffeatures);
         end;

      end;
   end;
   feacount := 0; (* zap the entries whether or not they were printed *)
end;

procedure addfeature(featureisexon: boolean; getstart, getstop: boolean);
(* see if the feature already exists in the list for this
locus.  If not, add it to the list.  mark whether to
get the start and stops.  This allows ends of CDS to be removed. 
When featureisexon is false then the feature is not a cds or exon (ie it's an
intron), and we should not mark it as an exon.  Only instructions for the
ends are given.
*)
var
   good: boolean; (* was the feature unique? *)
begin
{writeln(output,'addfeature');}
if debug then writeln(output);
if debug then writeln(output,'addfeature');
if debug then writeln(output,'    good unit: ',start:1,'..',stop:1);
if debug then writeln(output,'    feacount: ',feacount:1);
{if debug then writeln(output, 'checking for duplicate exons **********');}
   e := 1;
   good := true;
   if featureisexon
   then write(output,'exon ',start:1,'-',stop:1)
   else write(output,'intron ',start:1,'-',stop:1);
   while e <= feacount do begin
if debug then write(output, '    exon ', e:4,' ',
      fealist[e].thestart:1,'..',fealist[e].thestop:1);
      (* if either the start OR the stop is the same
         as any previous one, toss it out.  This
         not only eliminates duplicates but also
         drops alternative splicing.  In 4167 exons
         only 8 were dropped for this reason. *)
      if 
         (allowalternative and
          ((fealist[e].thestart = start) and
           (fealist[e].thestop = stop)))
         or
         ((not allowalternative) and
          ((fealist[e].thestart = start) or
           (fealist[e].thestop = stop)))
      then begin
{write(output, 'allowalternative: ', allowalternative);}
{
         write(output,'exon ',start:1,'-',stop:1,
                      ' was dropped because it is ');
}
         write(output,' was dropped because it is ');
         if ((fealist[e].thestart = start) and
             (fealist[e].thestop = stop))
         then write(output,'duplicate')
         else write(output,'alternative');

{
writeln(output);
writeln(output,'fealist[e].thestart = ', fealist[e].thestart);
writeln(output,'               start = ', start);
writeln(output,' fealist[e].thestop = ', fealist[e].thestop);
writeln(output,'                stop = ', stop);
}

         good := false;
         e := feacount;
      end
      else if ((fealist[e].thestart = start) or
               (fealist[e].thestop = stop))
         then write(output,' alternative'); (* mark alternatives *)
      e := succ(e)
   end;

   if good then begin 
      feacount := succ(feacount);
      if feacount > exonmax then begin
         writeln(output,'exonmax = ',exonmax:1,
                        ' exceeded');
         halt
      end;
      (* record the exon in the list *)
      fealist[feacount].complement := atcomplement;
      fealist[feacount].cds := cdscount;
      fealist[feacount].printasexon := featureisexon;

      if featureisexon then begin
         fealist[feacount].thestart := start;
         fealist[feacount].thestop  := stop;
         fealist[feacount].getthestart := getstart;
         fealist[feacount].getthestop  := getstop;
      end
      else begin (* for an intron, reverse the order of the recording: *)
         fealist[feacount].thestart := stop;
         fealist[feacount].thestop  := start;
         fealist[feacount].getthestart := getstop;
         fealist[feacount].getthestop  := getstart;
      end
   end;
   writeln(output);
end;

procedure dojoin;
(* do the join function to identify the parts of a CDS. *)
var
   getstart, getstop: boolean; (* whether to get the start or stop
      of an exon.  Don't ever get them if they are at the end of a CDS. *)
begin
if debug then writeln(output,'[JOIN FOUND]');
   printon := true;
   resettrigger(join);

{printon := false;}
   (* go through the join list *)
   resettrigger(closeparen);
   unitcount := 0;
   firstnumber := true;
   good := true;
   dotcount := 0;
   unitend := false;
   completeunit := false;
   repeat
      nextline;
      if db^=' '
      then nextcharacter(db,c) (* skip all blanks *)
      else begin
         nextline;
         nextcharacter(db,c);
         if (c = ',') or (c = ')')
         then begin (* end of the unit *)
            unitend := true;
            unitcount := succ(unitcount);
{write(output,'   unitend, unitcount = ',unitcount:1);}
         end
         else begin (* inside of a unit *)
            if c in numbers then begin
               if firstnumber
               then begin
                  readaninteger(db, c, start);

                  if start = stop + 1 then begin
                     killentry := true;
                     writestring(output,accname); (* report the name to output *)
                     writeln(output,' zero length intron containing entry!',
                                    ' DELETED');
                  end;

                  completeunit := false; (* until stop has been read *)
                  firstnumber := false;
               end
               else begin
                  readaninteger(db, c, stop);
                  completeunit := true;
               end
            end
            else if c = '.' then begin
               dotcount := succ(dotcount);
            end
            else begin (* any other junk: kill *)
{writeln(output,'unit died because of ',c);}
               good := false;
            end
         end; (* end of unit read *)
{writeln(output,'dotcount=', dotcount:1);}
{writeln(output,'HERE good=', good, ' unitend=',unitend);}

         if unitend and completeunit and good
         then if dotcount <> 2 then good := false;

         if noends and unitend then begin
            (* kill the end ones *)
            if unitcount = 1 then good := false;
            if {unitend and} (c = ')') then good := false;
         end;

{
write  (output,' unitend=',unitend);
write  (output,' completeunit=',completeunit);
writeln(output,' good=',good);
}

         if unitend and completeunit and good
         then begin (* record the unit *)
            if atcomplement then begin
               (* reverse start and stop *)
               hold := start;
               start := stop;
               stop := hold;
            end;

            (* Do another test.  Two cases:

               1) If we are at the low end of the CDS,
               and it is the exon start, this is
               probably not an acceptor and should be removed.

               2) If we are at the high end of the CDS,
               and it is the exon stop, this is
               probably not a donor and should be removed.
            *)

            getstart := true;
            getstop := true;
            (* test for low end of the CDS *)
            if unitcount = 1 then getstart := false;

            (* test for high end of the CDS *)
            if c = ')' then getstop := false;

            addfeature(true, getstart, getstop);
         end;

         if unitend then begin (* reeset for next time *)
{writeln(output,'UNITEND');}
            unitend := false;
            good := true;
            dotcount := 0;
            firstnumber := true;
            completeunit := false;
         end;
      end;

      testfortrigger(c,closeparen);
   until closeparen.found; (* join list finishes *)
   resettrigger(closeparen);
   atcomplement := false
end;

procedure dofeature(featureisexon: boolean);
(* get the feature information.  Cases of "<" or ">" are
not included since they are off the end of the piece.
     exon            143..251
                     /gene="hMLH1"
                     /number=14
When featureisexon is false then the feature is not a cds or exon (ie it's an
intron), and we should not mark it as an exon.  Only instructions for the
ends are given.
*)
var
   getstart, getstop: boolean; (* whether to get the start or stop
      of an exon. *)
begin
{writeln(output,'dofeature *************************');}
   start := 0;
   stop := 0;
   getstart := false;
   getstop := false;
   skipblanks(db);
   read(db,c);
   if c in numbers then begin (* ie, it's not "<" *)
      readaninteger(db, c, start);
      getstart := true;
{writeln(output,' start =',start:1);}
   end
   else if c = '<' then begin
      (* get the number but keep getstart false *)
      read(db,c);
      if c in numbers
      then readaninteger(db, c, start)
      else start := 0;
   end;;
   (* skip the ".." *)
   read(db,c);
   if c = '.' then begin
      read(db,c);
{writeln(output,' c1="',c,'"');}
      if c = '.' then begin
{writeln(output,' c2="',c,'"');}
         read(db,c);
{writeln(output,' c3="',c,'"');}
         if c = '>' then begin
            read(db,c); (* skip on *)
            if c in numbers
            then readaninteger(db, c, stop)
            else stop := maxint;
         end
         else if c in numbers then begin (* ie, it's not ">" *)
            readaninteger(db, c, stop);
            getstop := true;
{writeln(output,' stop =',stop:1);}
         end;
      end;
   end;
   readln(db);
{writeln(output,'start: ',getstart);}
{writeln(output,' stop: ',getstop);}
   if getstart or getstop then addfeature(featureisexon, getstart, getstop);
{writeln(output,'***********************************');}
end;

procedure readparameters;
(* read the parameters *)
begin
   reset(reffeap);

   readln(reffeap,c);
   if c = 'n' then noends := true
              else noends := false;

   readln(reffeap,c);
   if c = 'd' then debug := true
              else debug := false;

   readln(reffeap, theAfromrange, theAtorange);

   readln(reffeap, theDfromrange, theDtorange);

   read(reffeap,c);
   if c = 'e' then getexons := true
              else getexons := false;
   read(reffeap,c);
   if c = 'i' then getintrons := true
              else getintrons := false;
   readln(reffeap);

   readln(reffeap,c);
   if c = 'a' then allowalternative := true
              else allowalternative := false;

   read(reffeap,c);
   if c = 'P' then killputative := true
              else killputative := false;

   read(reffeap,c);
   if c = 'N' then killnotexperimental := true
              else killnotexperimental := false;

   read(reffeap,c);
   if c = 'G' then killgeneprediction := true
              else killgeneprediction := false;

   read(reffeap,c);
   if c = 'U' then killunpublished := true
              else killunpublished := false;

   read(reffeap,c);
   if c = 'S' then killpseudo := true
              else killpseudo := false;

end;

procedure writeparameters(var f: text);
(* read the parameters to file f *)
procedure no(body: boolean);
begin
   write(f,'* ');
   if body then write(f,'no ')
           else write(f,'   ');
end;
begin
   writeln(f ,'***********************************************');
   writeln(f ,'**************** parameters *******************');
   no(noends);
   writeln(f,'exons on the ends of CDS included.');

   writeln(f, '* the ACCEPTOR from range: ', theAfromrange:3);
   writeln(f, '* the ACCEPTOR   to range: ', theAtorange:3);
   writeln(f, '* the DONOR    from range: ', theDfromrange:3);
   writeln(f, '* the DONOR      to range: ', theDtorange:3);

   no(not getexons);
   writeln(f ,'exon features      included.');

   no(not getintrons);
   writeln(f ,'intron features    included.');

   no(not allowalternative);
   writeln(f ,'alternative exons  included.');

   no(not killputative);
   writeln(f ,'"putative"         entries will be killed.');

   no(not killnotexperimental);
   writeln(f ,'"not_experimental" entries will be killed.');

   no(not killgeneprediction);
   writeln(f ,'"gene prediction"  entries will be killed.');

   no(not killunpublished);
   writeln(f ,'"Unpublished"      entries will be killed.');

   no(not killpseudo);
   writeln(f ,'"pseudo"           entries will be killed.');

   writeln(f ,'***********************************************');

end;

procedure killtest(    c: char;
                   var t: trigger;
                       n: string;
                       killt: boolean;
                   var killentry: boolean);
(* Test whether the character c triggers trigger t.
n is the name of the entry.  killt is whether the entry should be killed
if the trigger t was found.  killentry is whether the entry has already been
killed. *)
begin
   testfortrigger(c,t);
   if t.found then begin
      if n.length > 0
      then writestring(output,n) (* report the name to output *)
      else write(output,'The next entry');
      write(output,' will be');
      if killt then begin
         write(output, ' KILLED');
         killentry := true;
      end
      else write(output, ' KEPT');
      write(output, '.  It contains: "');
      writestring(output,t.seek);
      writeln(output, '".');
      resettrigger(t);
   end;
end;

begin (* themain *)
   writeln(output,'ref ',version:4:2);

   readparameters;

   writeparameters(output);

   if debug then writeln(output,'DEBUGGING');

     (*                                1         2         3         4         5 *)
     (*                       12345678901234567890123456789012345678901234567890 *)
filltrigger(      locus,     'LOCUS               ');
filltrigger(   organism,     'ORGANISM            ');
filltrigger(  accession,     'ACCESSION           ');
filltrigger(  reference,     'REFERENCE           ');
{zzz}
filltrigger(   features,     'FEATURES            ');
filltrigger(        cds,     'CDS                 ');
filltrigger(       exon,     'exon                ');
filltrigger(     intron,     'intron              ');
filltrigger(       join,     'join(               ');
filltrigger( closeparen,     ')                   ');
filltrigger(      comma,     ',                   ');
filltrigger( complement,     'complement(         ');
filltrigger(   putative,     'putative            ');
filltrigger(notexperimental, 'not_experimental    ');
filltrigger(geneprediction,  'gene prediction     ');
filltrigger(unpublished,     'Unpublished         ');
filltrigger(pseudo,          'pseudo              ');
filltrigger(entryend,        '//                  ');

   resettrigger(reference);
{zzz}

   resettrigger(cds);
   resettrigger(exon);
   resettrigger(intron);
   resettrigger(locus);
   resettrigger(accession);
   resettrigger(features);
   resettrigger(join);
   resettrigger(closeparen);
   resettrigger(complement);
   resettrigger(comma);
   resettrigger(putative);
   resettrigger(notexperimental);
   resettrigger(geneprediction);
   resettrigger(unpublished);
   resettrigger(pseudo);
   resettrigger(entryend);

   locuscount := 0;
   cdscount := 0;
   totalentrycount := 0;
   totalfeacount := 0;
   feacount := 0;
   infeatures := false;
   linenumber := 1; (* we are already at the first line *)
   numbers:=['0','1','2','3','4','5','6','7','8','9'];
   killentry := false;

   rewrite(reffeatures);
   writeln(reffeatures,'* ref: reffeatures ',version:4:2);

   reset(db);

   while not eof(db) do begin
      while not eoln(db) do begin
         nextcharacter(db, c);

         testfortrigger(c,locus);
         if locus.found then begin
            killentry := false;
            locuscount := succ(locuscount);
            cdscount := 0;
            clearstring(accname);
            infeatures := false;
            atcomplement := false;
            stop := -maxint; (* be sure initial intron length is large *)
            start := -maxint; (* just to be clean *)
if debug then writeln(output);
if debug then writeln(output,'LOCUS ',locuscount:1,' ======================');
         end;

         testfortrigger(c,entryend);
         if entryend.found then begin
            printexons; (* print previously found exons *)
         end;

{write(output,c);}
         testfortrigger(c,accession);
         if accession.found then begin
            (* capture the accession string *)
            grabtoken(db,accname);
            writestring(output,accname); (* report the name to output *)
            writeln(output);
            if accname.length <= 0 then begin
               writeln(output,'MISSING ACCESSION STRING');
            end;
            resettrigger(accession);
if debug then write(output,'ACCESSION ');
if debug then writestring(output, accname);
if debug then writeln(output,' ======================');
         end;

         (* test for bad entry markers and kill them: *)
         if not killentry then begin
            killtest(c,putative,       accname,killputative,       killentry);
            killtest(c,notexperimental,accname,killnotexperimental,killentry);
            killtest(c,geneprediction ,accname,killgeneprediction ,killentry);
            killtest(c,unpublished    ,accname,killunpublished    ,killentry);
            killtest(c,pseudo         ,accname,killpseudo         ,killentry);
         end;

         testfortrigger(c,reference);
         if reference.found then begin
            (* read the reference, which has this form:
            REFERENCE   6  (bases 37945 to 38027)
            *)

            skipblanks(db);
            (* read reference number *)
            read(db,c); (* first character *)
            if c in numbers then begin
               readaninteger(db, c, refnumber);
            end
            else begin
               writeln(output,'a numeral does not follow the REFERENCE!');
               halt
            end;

            skipblanks(db);
            (* read reference number *)
            read(db,c); (* first character *)
            if c <> '(' then begin
               writeln(output,'REFERENCE number is not followed by a "("');
               halt
            end;

            read(db,c); (* 'base' starts ones we want *)
            if c = 'b' then begin

               skipnonblanks(db); (* skip the "(bases" *)

               skipblanks(db);
               (* read reference start *)
               read(db,c); (* first character *)
               if c in numbers then begin
                  readaninteger(db, c, start);
               end
               else begin
                  writeln(output,'a numeral does not follow the REFERENCE!');
                  halt
               end;
   
               skipblanks(db);
               read(db,c); (* first character *)
               if c <> 't' then begin
                  writeln(output,'first REFERENCE number is not followed by "to"');
                  halt
               end;

               skipnonblanks(db); (* finish the "to" off *)

               skipblanks(db);
                  (* read reference stop *)
               read(db,c); (* first character *)
               if c in numbers then begin
                  readaninteger(db, c, stop);
               end
               else begin
                  writeln(output,'a numeral does not follow the "to"!');
                  halt
               end;

               length := stop-start + 1;

write  (output,'reference number ',refnumber:1);
write  (output,' start ',start:1);
write  (output,' stop ',stop:1);
writeln(output,' length ',length:1);
{zzzqqq}
                (* write the definitions and calls in this form:
                   define "J02459.ref.5" "-" "<]" "[>" 0 [length-1]
                   @ J02459 [start] 1 "J02459.ref.5" ""
                *)
               write(reffeatures,'define');

               (* name of reference *)
               write(reffeatures,' "');
               writestring(reffeatures,accname);
               write(reffeatures, '.ref.',refnumber:1,'"');

               writeln(reffeatures,
                       ' "-" "<]" "[>" 0 ',
                       (length-1):1
                       );

               (* Write the call to the definition: *)

               write(reffeatures,'@ ');
               writestring(reffeatures,accname);
               write(reffeatures,' ',start:1);

               (* orientation *)
               if stop > start
               then write(reffeatures,' 1')
               else write(reffeatures,' -1');

               (* name of reference *)
               write(reffeatures,' "');
               writestring(reffeatures,accname);
               write(reffeatures, '.ref.',refnumber:1,'"');
               write(reffeatures, ' ""');
               writeln(reffeatures);

               totalentrycount := succ(totalentrycount);
               totalfeacount := succ(totalfeacount);
            end;

         end;

         testfortrigger(c,features);

(* zzz kill feature functions *)
         features.found := false;

         if features.found then begin
{writeln(output,'FEATURE found'); }
            infeatures := true;
         end;

         testfortrigger(c,cds);
         if cds.found then if infeatures then begin
            skipblanks(db);
            cdscount := succ(cdscount);
if debug then writeln(output);
if debug then writeln(output,'CDS ',cdscount:1,  ' ======================');

            while not eoln(db) do begin
               nextcharacter(db, c);

               testfortrigger(c,complement);
               if complement.found
               then atcomplement := true;

               if atcomplement then begin
                  testfortrigger(c,closeparen);
                  if closeparen.found then begin
                     atcomplement := false
                  end;
               end;

               begin (* unique case of a single exon, eg M35391 *)
                  testfortrigger(c,join);
                  if join.found then begin
                     dojoin
                  end
                  else begin (* If it's a number then just grab the two numbers *)
                     if c in numbers then begin

                        atcomplement := false;
                        skipblanks(db);
                        readaninteger(db, c, start);

                        read(db, c);
                        if c <> '.' then begin
                           writeln(output,'error in reading CDS "',c,
                                          '" found instead of "."');
                           halt
                        end;

                        read(db, c);
                        if c <> '.' then begin
                           writeln(output,'error in reading CDS "',c,
                                          '" found instead of "."');
                           halt
                        end;

                        read(db, c);
                        if not (c in numbers) then begin
                           if c = '>' then begin
                              read(db,c); (* skip on *)
                              if c in numbers
                              then readaninteger(db, c, stop)
                              else stop := maxint;
                           end
                           else begin
                              writeln(output,'error in reading CDS "',c
                                            ,'" is not a number');
{zzz}
                              halt
                           end
                        end
                        else readaninteger(db, c, stop);
{
writeln(output,'start ',start:1);
writeln(output,'stop ',stop:1);
}
                        addfeature(true, true, true);
                     end
                  end
               end

            end
         end;

         if infeatures then if getexons then begin
            testfortrigger(c,exon);
            if exon.found then begin
               dofeature(true);
               resettrigger(exon);
            end;
         end;

         if infeatures then if getintrons then begin
            testfortrigger(c,intron);
            if intron.found then begin
               dofeature(false);
               resettrigger(intron);
            end;
         end;

      end;
      nextline;
   end;

   printexons; (* finish off previous ref list *)

   writeln(output,totalentrycount:1,' entrys have REFERENCES');
   writeln(output,totalfeacount:1,' REFERENCES processed');
end;
(* end module ref.themain *)

begin
   themain(reffeap, db, reffeatures);
1: end.
