program rila(ribl, rilap,
             headlatex, labelatex, sequlatex, senulatex,
             freqlatex, ribllatex, taillatex,
             output);
(* rila: reformat the ribl table and a sequence into latex format

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.44; (* of rila.p 2013 Mar 12
2013 Mar 12, 1.43: clean up
2013 Mar 12, 1.42: make frequency table
2012 Mar 12, 1.41: tighten \hspace{\fboxsep} code into a command (failed!)
2012 Mar 12, 1.40: use \hspace{\fboxsep} for all non-fbox cases to align!
2012 Mar 12, 1.39: absorb boldface from rila.p diverged line 1.32
                   ~/slides/ev/figs/ev-html/ribl-weightmatrix/rebuild
2007 Sep 06, 1.38: zap wrapping
2007 Sep 06, 1.37: no wrapping in latex file
2005 May 23, 1.36: add patent message
2004 Mar 19, 1.35: upgrade documentation
2002 May 14, 1.34: cleanup
2002 May 14, 1.33: bring in new ribl reading functions from prgmod
2002 May 14, 1.30: upgrade for new ribl format
1996 Oct 24, 1.29: previous version
origin 1992 August 15 *)
(* end module version *)

(* begin module describe.rila *)
(*
name
   rila: reformat the ribl table into latex format

synopsis
   rila(ribl: in, rilap: in,
        headlatex: out, labelatex: out, senulatex: out,
        sequlatex: out, freqlatex: out, ribllatex: out,
        taillatex: out,
        output: out)

files
   ribl: output of ri program

   rilap:  parameters to control the program.

      1: two integers that define the range of ribl to convert.

      2: the lowest Ri value to print; lower values are given
        a negative infinity symbol.

      3: a sequence to put boxes around.  If there is no sequence,
        then no boxes will be used.

      4: number of decimal places for Ri values.

      5: capitalize: if the first character is u or c, it the sequence
      and bases are uppercase.  otherwise, they are lower case.

      6: if the first character is h, \hline is put before and after
         the labelatex, riblatex and senulatex.

      7: if the first character is b, make a box around each Ri(b,l)
         value corresponding to the sequence.

      8: if the first character is b, make everything boldface.

   headlatex: start of table format for LaTeX

   labelatex: label line format for LaTeX

   sequlatex: table format for LaTeX of the sequence as letters

   senulatex: table format for LaTeX of the sequence as numbers

   freqlatex: table format for LaTeX of the frequencies of bases

   ribllatex: table format for LaTeX of the weight matrix with the
      weights for the sequence boxed.

   taillatex: end of table format for LaTeX

   output: messages to the user

description
   Read the ribl and reformat it so it can be used in a LaTeX table.  Also
   generate the sequence matrix, and box the sequence on the ribl table.  The
   program also supplies a table header, a label line and a tail.  These can
   be concatenated together.  The user may can insert whatever they want
   between parts or do other things to the table.

examples

   This UNIX command concatenates the parts together:

rila
cat \
version   \
headlatex \
    morehead  \
labelatex \
sequlatex \
senulatex \
    midfigure \
labelatex \
ribllatex \
taillatex \
> tabular.tex

   The user creates two files, morehead and midfigure.

   Here is a rilap:

-3 6     two integers that define the range of ribl to convert.
-10      the lowest Ri value to print; lower values are given
caggtctgca
1        number of decimal places for Ri values
c        u or c = upper case sequence and bases
h        h = hlines given
b        b = box the sequence on the ribl matrix
b        b = boldface everything

rilap: parameters for rila version 1.39 or higher

documentation

   See Figure 1 of

@article{Schneider.Ri,
author = "T. D. Schneider",
title = "Information Content of Individual Genetic Sequences",
journal = "J. Theor. Biol.",
volume = "189",
number = "4",
pages = "427-441",
note = "\htmladdnormallink
{https://alum.mit.edu/www/toms/paper/ri/}
{https://alum.mit.edu/www/toms/paper/ri/}",
year = "1997"}

see also

   {program that makes the ribl file: } ri.p

   https://alum.mit.edu/www/toms/paper/ri/

author
   Thomas Dana Schneider

bugs

   A bug in LaTex will not allow one to read the taillatex in without getting
   extra junk on the end of the table.  You can either concatenate the parts
   of the table together and read it all at once or you can supply your own
   '\end{tabular}'.

technical notes

*)
(* end module describe.rila *)

(* begin module ribl.const *)
   minribl = -10001; (* lowest  ribl matrix from allowed *)
   maxribl = +10001; (* highest ribl matrix to   allowed *)
   defnegativeinfinity = -1000; (* default for negative infinity
                                   for the Ri(b,l) table *)
(* end module ribl.const version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module string.const *)
      maxstring = 2000; (* the maximum string *)
(* end module string.const version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

type

(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      previous: stringptr; (* the previous string in a series *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module wave.type *)
      waveptr = ^waveparam; (* to link several wave definitions together *)
      waveparam = record (* parameters to define a cosine wave *)
         (* define a cosine wave: *)
         extreme: char; (* h or l, the high or low extreme to be defined *)
         wavelocation: real; (* the location in bases of the extreme *)
         wavebit: real; (* the location in bits of the extreme *)
         waveamplitude: real; (* the amplitude of the wave in bits *)
         wavelength: real; (* the wave length of the wave in bases *)
{ 2000 Nov 1.  This is difficult because the length of a curve is not
simply multiplied by a scale factor.
         BASEdashon: real; (* dashon interval in bases *)
         BASEdashoff: real; (* dashon interval in bases *)
         BASEdashoffset: real; (* dashon interval in bases *)
}
         dashon: real; (* the size of on dashes in cm.
                          dashon <= 0 means no dashes *)
         dashoff: real; (* the size of on dashes in cm *)
         dashoffset: real; (* the size of off dashes in cm *)
         thickness: real; (* thickness of the wave in cm.  <= 0 means default *)
         next: waveptr; (* the next wave *)
      end;
(* end module wave.type version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module base.type *)
   (* define the four nucleotide bases *)
   base = (a,c,g,t);
(* end module base.type version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module scan.type *)
   (* define the ribl matrix.  Source library is prgmod.p
      REQUIRES ribl.const module
      REQUIRES base.type or book.type module
      REQUIRES wave.type module for wave definition! *)
   rblarray = array[a..t, minribl..maxribl] of real; (* real(B,L) *)
   ribltype = record
      riblname: string; (* name of the weight matrix *)
      riblheader: stringptr; (* misc header information, starting with '*' *)
      data: rblarray; (* real(B,L) *)
      numbers: array[a..t, minribl..maxribl] of integer; (* n(b,l) *)
      frombase, tobase: integer; (* range of the matrix *)
      mean, stdev: real; (* mean and standard deviation of distribution *)
      consensus: real; (* Ri value of consensus sequence *)
      anticonsensus: real; (* Ri value of anticonsensus sequence *)
      averageRi: real; (* average Ri for random sequence *)
      n: integer; (* number of sequences used to create matrix *)
      symmetry: char; (* the symmetry of the Riblmatrix *)
      waves: waveptr; (* cosine wave definitions for the matrix *)
      cmperbase, cmperbit: real; (* required for wave dash definition *)
      Ribound: real; (* the Ri boundary for this definition *) 
      Zbound: real; (* the Z boundary for this definition *) 
      Pbound: real; (* the probability boundary for this definition *) 
   end;
   ribltypeptr = ^ribltype; (* pointer to ribltype so it does
      not need to be created unless actually used *)
(* end module scan.type version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

var
   (* files used by this program: *)
   ribl: text;
   headlatex: text;
   labelatex: text;
   sequlatex: text;
   senulatex: text;
   freqlatex: text;
   ribllatex: text;
   taillatex: text;
   rilap: text;

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module skipblanks *)

(* 2003 July 31: tab is considered a blank character *)

function isblank(c: char): boolean;
(* is the character c blank or tab? *)
const
   tab = 9; (* tab character *)
begin
   isblank := (c = ' ') or (ord(c) = tab)
end;

procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
   while isblank(thefile^) and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
   while (not isblank(thefile^)) and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module clearstring *)
(* These modules clear strings in various ways *)
(* ---- *)
procedure emptystring(var ribbon: string);
(* empty the contents of the string but do NOT
remove the pointer.  This is useful for clearing
one string within a linked list of them. *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
   with ribbon do begin
      for index := 1 to maxstring do letters[index] := ' ';
      length := 0;
      current := 0;
   end
end; (* emptystring *)
(* ---- *)
procedure clearstring(var ribbon: string);
(* empty the string and remove the pointer.
As of 2010 Dec 04 there is a previous pointer that
allows proper removal of the linked list using disposestring.
This clearstring routine is for the initial clearing
of the variables. *)
begin (* clearstring *)
   with ribbon do begin
      emptystring(ribbon);
      previous := nil;
      next := nil;
   end
end; (* clearstring *)
(* ---- *)
procedure disposestring(var ribbon: stringptr);
(* remove the linked list from the string in revere order *)
var
   t: stringptr; (* a link in the string *)
begin
   t := ribbon;
   (* find the end *)
   while t^.next <> nil do t := t^.next;
   (* move backwards and dispose links *)
   while t^.previous <> nil do begin
      t := t^.previous;
      dispose(t^.next);
   end;
   dispose(t);
end; (* disposestring *)
(* ---- *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer.  This is now deprecated, do not use it
since clearstring still clears the next pointer. *)
begin (* initializestring *)
   writeln(output,'remove initializestring routine!');
   writeln(output,'replace it with clearstring routine!');
   halt; (* to force deprecation *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module detabstring *)
procedure detabstring(var s: string);
(* Tabs are converted to blanks.  This allows the tab to be a special "solid"
kind of blank, but the string written won't vary with position on the line
depending on the tabstops. *)
const
   tabcharacter = 9; (* the ordinal of the tab character *)
var
   c: char; (* a character write out *)
   i: integer; (* index to the string s *)
begin
   with s do for i := 1 to length do begin
      c := letters[i];
{
for testing:
      if ord(c) = tabcharacter then c := 'T'; (* tab converted to blank *)
}
      if ord(c) = tabcharacter then c := ' '; (* tab converted to blank *)
      letters[i] := c
   end;
end;
(* end module detabstring version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module readstringline *)
procedure readstringline(var afile: text; var buffer: string);
(* get a line (as a string) from a file. *)
var   index: integer; (* of buffer *)
begin (* getstring *)
   clearstring(buffer);
   index := 0;
   while (not eoln(afile)) and (index < maxstring) do begin
      index := succ(index);
      read(afile, buffer.letters[index])
   end;
   if not eoln(afile) then begin
      writeln(output, 'readstringline: a line exceeds maximum string size (',
                      maxstring:1,')');
      halt
   end;
   buffer.length := index;
   buffer.current := 1;
   readln(afile);
end; (* readstringline *)
(* end module readstringline version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module writestring version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module readquotestring *)
procedure readquotestring(var afile: text; var s: string);
(* read the quoted string from afile, remove the quotes.  Trailing
blanks are converted to tab characters. *)
const
   tabcharacter = 9; (* the ordinal of the tab character *)
var
   c: char; (* a character read *)
   i: integer; (* index to the string s *)
begin
   emptystring(s);
   skipblanks(afile);
   with s do begin
      read(afile,c);
      if c <> '"' then begin
         writeln(output, 'quote string expected but " missing');
         writeln(output, 'character found instead:"',c,'"');
         halt
      end;
      if eoln(afile) then begin
         writeln(output, 'missing end of quote string " string');
         halt
      end;
      repeat
         read(afile,c);
         if c <> '"' then begin
            length := succ(length);
            letters[length] := c;
         end;
      until (c = '"')
            or eoln(afile) or (length = maxstring);
      if length = maxstring then begin
         writeln(output, 'A quote string exceeded ',maxstring:1,
                         ' characters.');
         writeln(output, 'Use a shorter string or increase constant',
                         ' maxstring.');
         halt
      end;
      if c <> '"' then begin
         writeln(output, 'The second " mark of a string is missing.');
         halt
      end;
      (* convert all blanks to tab characters *)
      for i := 1 to length do begin
         if letters[i] = ' '
         then letters[i] := chr(tabcharacter); (* set to tab character *)
      end;
   end;
end;
(* end module readquotestring version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module writequotestring *)
procedure writequotestring(var afile: text; s: string);
(* Write the string s to afile, surrounded by double quote marks.  Tabs are
converted to blanks.  This allows the tab to be a special "solid" kind of
blank, but the string written won't vary with position on the line depending
on the tabstops. *)
begin
   detabstring(s);
   write(afile, '"');
   writestring(afile, s);
   write(afile, '"');
end;
(* end module writequotestring version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module equalstring *)
function equalstring(a, b: string): boolean;
(* Test for equality between two strings at current positions.
NOTE:  A compiler bug results if one directly tests this way:
   if thedefinition^.nametag = aname
The reason is completely not clear!  I showed that the parts of the
strings were identical, but the whole was not by this test.
For this reason it is *critical to test strings with equalstring.
*)
var   index: integer; (* index to both strings *)
      equal: boolean; (* are letters in a and b the same? *)
begin (* equalstring *)
      if a.length = b.length
      then begin
         index := 1;
         repeat
            equal := (a.letters[index] = b.letters[index]);
            index := succ(index)
         until (not equal) or (index > a.length);
         equalstring := equal
      end
      else equalstring := false
end; (* equalstring *)
(* end module equalstring version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module grabtoken *)
procedure grabtoken(var thefile: text; var thestring: string);
(* skip any blanks and then grab the next token from the file *)
var
   c: char; (* a character in thefile *)
   done: boolean; (* done finding the name *)
begin
   skipblanks(thefile);
   done := false;
   with thestring do begin
      length := 0;
      while not done do begin
         if eoln(thefile)
         then done := true
         else begin
            read(thefile,c);
            if c = ' '
            then done := true
            else begin
               length := succ(length);
               letters[length] := c;
            end
         end
      end
   end
end;
(* end module grabtoken version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module capitalize *)
function capitalize(c: char): char; 
(* convert the character c to upper case *) 
var 
   n: integer; (* c is the n'th letter of the alphabet *)
begin 
   n := ord(c); 
   if (n >= ord('a')) and 
      (n <= ord('z'))   
   then c := chr( n - ord('a') + ord('A'));
   capitalize := c
end; 
(* end module capitalize version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module decapitalize *)
function decapitalize(c: char): char; 
(* convert the character c to lower case *) 
var 
   n: integer; (* c is the n'th letter of the alphabet *)
begin 
   n := ord(c); 
   if (n >= ord('A')) and 
      (n <= ord('Z'))   
   then c := chr( n - ord('A') + ord('a'))
   else c := chr(n);
   decapitalize := c
end; 
(* end module decapitalize version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module readheader *)
procedure readheader(var afile: text; ch: char; var theheader: stringptr);
(* Read the header from a file.  Only read lines that begin with ch.
2004 Dec 15: Handle end of file (eof) gracefully. *)
var
   h: stringptr; (* pointer to the riblheader *)
begin
   if eof(afile) then begin
      clearstring(theheader^);
   end
   else begin
      if afile^ = ch then begin
         new(theheader);
         h := theheader;
         while afile^='*' do begin
            readstringline(afile, h^);
            if not eof(afile) then begin
               if afile^='*' then begin
                  new(h^.next);
                  h := h^.next;
               end
               else h^.next := nil;
            end
         end
      end
      else begin
         theheader := nil
      end;
   end;
end;
(* end module readheader version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module readwaveparameters *)
procedure readawaveparameter(var afile: text; var wp: waveparam);
{
                                 cmperbase, cmperbit: real);
}
(* Read one wave parameter from a file. *)
{
2000 Nov 1:  The following experimental code is not implemented:
See the coscurve program
(https://alum.mit.edu/www/toms/delila/coscurve.html) for details
about the conversion from bases to cm.  The code now requires
cm per base and cmperbit.
const
   lover2pi = 1.2160067233; (* L/2pi for L being the length of
       the cosine function from 0 to 2pi *)
var
   multiplier: real; (* factor to multiply by to convert from bases to cm. *)
}
begin
   readln(afile,wp.extreme);
   readln(afile,wp.wavelocation);
   readln(afile,wp.wavebit);
   readln(afile,wp.waveamplitude);
   readln(afile,wp.wavelength);
   if wp.wavelength <= 0.0 then begin
      writeln(output,'wave parameters: wavelength must be positive');
      halt
   end;
   (* one may skip the dash and thickness parameters
      of the last wave definition.  This allows
      backwards compatability, but should not be
      used in general. *)
   if eof(afile)
   then begin
      wp.dashon := 0;
      wp.dashoff := 0;
      wp.dashoffset := 0;
      wp.thickness := 0
   end
   else begin
      (* implement new method to read dash information
         without breaking previous wave parameter files *)
      if afile^ <> 'd'
      then begin (* default to old method *)
         readln(afile,wp.dashon);
         wp.dashoff := wp.dashon;
         wp.dashoffset := 0;
      end
      else begin
         get(afile); (* skip the 'd' *)
         (* reading the parameters directly in cm would require
            that the user compute or guess the values.
            So read in as bases and compute cm.
            Unfortunately the length of a curve is not the simple
            scale factor.  Make the user guess for now.
         *)
         readln(afile,wp.dashon, wp.dashoff, wp.dashoffset)

{
experimental code to account for actual wave length:
         readln(afile, wp.BASEdashon, wp.BASEdashoff, wp.BASEdashoffset);
         multiplier    := 1;
         multiplier    := lover2pi
 * wp.waveamplitude (* in bits *)
 * cmperbit (* converts to cm *)
 / wp.wavelength (* 10.6 bp/turn *)
* cmperbase
;
multiplier    := lover2pi;
multiplier    := 1.0;
multiplier    := 1.0/lover2pi;
         multiplier := lover2pi
                       * wp.waveamplitude (* in bits *)
                       * cmperbit (* converts to cm *)
                       * wp.wavelength (* 10.6 bp/turn *)
                       * cmperbase;
         multiplier := 1.0
         wp.dashon     := multiplier * wp.BASEdashon;
         wp.dashoff    := multiplier * wp.BASEdashoff;
         wp.dashoffset := multiplier * wp.BASEdashoffset;
with wp do begin
   writeln(output, 'cmperbase  = ',cmperbase:5:2);
   writeln(output, 'cmperbit   = ',cmperbit:5:2);
   writeln(output, 'dashon     = ',dashon:5:2);
   writeln(output, 'dashoff    = ',dashoff:5:2);
   writeln(output, 'dashoffset = ',dashoffset:5:2);
   writeln(output, 'waveamplitude   = ',wp.waveamplitude:5:2);
   writeln(output, 'wavelength      = ',wp.wavelength:5:2);
end;
}

      end;
      if eof(afile)
      then wp.thickness := 0
      else readln(afile,wp.thickness);
   end;
end;

procedure readwaveparameters(var afile: text; var w: waveptr);
{
                                 cmperbase, cmperbit: real);
}
(* read from afile the wave parameters w.  The routine is done either at end
of file, when a period ('.') appears or at a blank line.  This makes it
compatable with different reading mechanisms. *)
var
   done: boolean; (* done reading *)
   morethanone: boolean; (* more than one wave? *)
   p: waveptr; (* one of the definitions of a cosine wave *)
procedure waystoend;
(* if end of file, a blank line or a period, consider the job done *)
var
   clear: boolean; (* not eof, so clear to test for comment *)
begin
   if eof(afile) then done := true;
   clear := false;
   if not done (* skip any number of comments *)
   then while (not clear) and (not done) do begin
      if eof(afile) then begin
         done := true;
         clear := true;
      end
      else if afile^ = '*'
           then readln(afile)
           else clear := true
   end;
   if not done then if eoln(afile) then done := true;
   if not done then if afile^='.' then begin
      readln(afile);
      done := true;
   end;
end;
begin
   done := false;
   morethanone := false;
   w := nil;
   waystoend;
   if not done then begin
      new(w);
      p := w;
      while not done do begin
         waystoend;
         if done
         then p^.next := nil
         else begin
            if afile^='*'
            then readln(afile) (* skip any number of comments *)
            else begin
               if morethanone then begin
                  new(p^.next);
                  p := p^.next
               end;
{
               readawaveparameter(afile, p^, cmperbase, cmperbit);
}
               readawaveparameter(afile, p^);
               (* if we ever come back again there will be more than one: *)
               morethanone := true;
            end
         end
      end
   end
   else w := nil;
end;
(* end module readwaveparameters version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module getriblmatrix *)
procedure getriblmatrix(var afile: text; var matrix: ribltypeptr);
(* Get the matrix from a file, with the defining coordinate limits,
followed by the mean and standard deviation and other data.
The entire matrix is read in. *)
const
   clearvalue = 5.0; (* If a large bit position is ever seen (for DNA), we
      KNOW there is a bug in the code! *)
var
   b: base; (* a base in the matrix *)
   l: integer; (* a coordinate in the matrix *)
   thel: integer; (* the l value in the file (should match l) *)

   teststring: string; (* string to test for 'ribl ' at matrix start *) 
   riblstart: string; (* word at the start of the matrix *)

procedure testforend;
(* test for end of file *)
begin
   if eof(afile) then begin
      writeln(output,'part or all of the ribl matrix is missing');
      halt;
   end
end;
procedure skip;
(* skip comment lines *)
begin
   testforend;
   while afile^='*' do begin
     readln(afile);
     testforend
   end
end;
begin
   new(matrix);
   with matrix^ do begin

      (* read name of the matrix *)
      grabtoken(afile, riblstart);
      teststring.letters[1] := 'r'; 
      teststring.letters[2] := 'i';
      teststring.letters[3] := 'b';
      teststring.letters[4] := 'l';
      teststring.length := 4; 
 
      if not equalstring(riblstart, teststring) then begin
         writeln(output,'The ribl matrix must begin with ''ribl "name"''');   
         writequotestring(output,riblstart);
         writeln(output, ' seen instead');
         halt
      end;
 
      skipblanks(afile); (* skip any spaces *)
      readquotestring(afile, matrix^.riblname);
      readln(afile);

      readheader(afile,'*',riblheader);

      readln(afile,frombase,tobase);

      if frombase < minribl then begin
         writeln(output,'Matrix is too large for program.');
         writeln(output,'frombase (=',frombase:1,')');
         writeln(output,'minribl (=',minribl:1,')');
         writeln(output,'Decrease minribl or increase frombase');
         halt
      end;

      if tobase > maxribl then begin
         writeln(output,'Matrix is too large for program.');
         writeln(output,'tobase (=',tobase:1,')');
         writeln(output,'maxribl (=',maxribl:1,')');
         writeln(output,'Increase maxribl or decrease tobase');
         halt
      end;

      (* clear unused matrix for safety *)
      for l := minribl to frombase-1
      do for b := a to t do matrix^.data[b,l] := clearvalue;
      for l := tobase+1 to maxribl
      do for b := a to t do matrix^.data[b,l] := clearvalue;

      for l := frombase to tobase do begin
         testforend;
         for b := a to t do read(afile,matrix^.data[b,l]);
         read(afile,thel);
         if thel <> l then begin
            writeln(output, 'positions in ribl do not match from-to range');
            halt;
         end;
         for b := a to t do read(afile,matrix^.numbers[b,l]);
         readln(afile)
      end;

      skip; readln(afile,mean);
      skip; readln(afile,stdev);
      skip; readln(afile,consensus);
      skip; readln(afile,anticonsensus);
      skip; readln(afile,averageRi);
      skip; readln(afile,n);
      skip; readln(afile,symmetry);
      skip; readln(afile,Ribound);
      skip; readln(afile,Zbound);
      skip; readln(afile,Pbound);
{this is a guess - set the converters to neutral for now
      cmperbase := 1.0;
      cmperbit := 1.0;
      readwaveparameters(afile, waves, cmperbase, cmperbit);
}
      readwaveparameters(afile, waves);
   end;
end;
(* end module getriblmatrix version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module PatentMessage *)
procedure PatentMessage(var f: text);
(* Write the patent message to file f
version = 3.02 of PatentMessage_module 2017 Jul 31
*)
procedure makebar;
begin
writeln(f,
'**********************************************************************');
end;
begin
makebar;
writeln(f,
'* By downloading this code you agree to the Source Code Use License: *');
writeln(f,
'* https://alum.mit.edu/www/toms/Source_Code_Use_License_Schneider.txt *');
writeln(f,
'* Contact: https://alum.mit.edu/www/toms/contacts.html                *');
makebar;
end;
(* end module PatentMessage version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module rila.themain *)
procedure themain(var ribl, rilap,
           headlatex, labelatex, sequlatex, senulatex,
           freqlatex, ribllatex, taillatex:
           text);
(* the main procedure of the program *)
const
   wid = 1; (* width of output numbers - set to 1 to save space *)
var
   askedfrom: integer; (* the from requested *)
   askedto: integer; (* the to requested *)
   b: base; (* a base in the matrix *)
   caps: boolean; (* if true, capitalize bases and sequence *)
   closebox: boolean; (* if true, close the fbox *)
   column: integer; (* column in the input ribl array.  This is a ROW
      on the output table! *)
   count: integer; (* count of items written out *)
   decimals: integer; (* number of decimals to print Ri values *)
   dobox: boolean; (* if true, put boxes around Ri(b,l) that match the sequence *)
   dobold: boolean; (* if true, make everything bold face *)
   hline: boolean; (* if true, put in \hline around parts *)
   l: integer; (* a coordinate in the matrix *)
   length: integer; (* length of the sequence *)
   matrix: ribltypeptr; (* the ribl matrix *)
   negativeinfinity: real ; (* negative infinity for the Ri(b,l) table *)
   sequence: array[1..20] of char; (* nucleotide sequence in lower case *)
   thebase: char; (* character representing current base *)

   wraplimit: integer; (* number of items in latex lines before wrapping.
      2007 Sep 06: set to high, can be parameter later. *)

   sum: integer; (* sum of n(b,l) *)
   bsum: base; (* a base in the n(bsum,l) matrix *)

procedure startfiles;
(* start the files *)
begin
   writeln(output,'rila ',version:4:2);

   PatentMessage(output);

   rewrite(headlatex);
   rewrite(labelatex);
   rewrite(sequlatex);
   rewrite(senulatex);
   rewrite(freqlatex);
   rewrite(ribllatex);
   rewrite(taillatex);
   writeln(headlatex,'% rila ',version:4:2,' head');
   writeln(labelatex,'% rila ',version:4:2,' label');
   writeln(sequlatex,'% rila ',version:4:2,' sequence as letters');
   writeln(senulatex,'% rila ',version:4:2,' sequence as numbers');
   writeln(freqlatex,'% rila ',version:4:2,' f(b,l) matrix');
   writeln(ribllatex,'% rila ',version:4:2,' Ri(b,l) matrix');
   writeln(taillatex,'% rila ',version:4:2,' tail');
end;

procedure readparameters;
(* read the parameters *)
var
   p: integer; (* index to position *)
begin
   reset(rilap);
   if eof(rilap) then begin
      writeln(output,'Missing data in rilap');
      halt
   end;

   readln(rilap,askedfrom,askedto);
   if (askedto < minribl) or (askedfrom > maxribl) then begin
      writeln(output,'cannot handle this much data; use smaller range or',
                     ' increase constants minribl or maxribl');
      halt
   end;

   readln(rilap,negativeinfinity);

   (* read sequence *)
   if not eof(rilap) then begin
     length := 0;
     while not eoln(rilap) do begin
       length := length + 1;
       read(rilap,sequence[length]);
       sequence[length] := decapitalize(sequence[length]);
       write(output,sequence[length]);
       if not(sequence[length] in ['a','c','g','t']) then begin
          writeln(output,'sequence must be characters acgt only');
          halt
       end;
     end;
     writeln(output);
   end;
   if length > 0 then begin
      write(labelatex,'% sequence read in: ');
      for p := 1 to length do write(labelatex,sequence[p]);
      writeln(labelatex,' is of length ',length:1);
   end;

   readln(rilap,decimals);

   if (rilap^ = 'c') or (rilap^ = 'u') then caps := true
                                       else caps := false;
   readln(rilap);

   if rilap^='h' then hline := true
                 else hline := false;
   readln(rilap);

   if rilap^='b' then dobox := true
                 else dobox := false;
   readln(rilap);

   dobold := false;
   if not eof(rilap) then begin
      if not eoln(rilap) then begin
         if rilap^='b' then dobold := true
                       else dobold := false;
         readln(rilap);
      end;
   end;

end;

procedure db(var f: text);
(* do bold begin to file f *)
begin
  if dobold then
  write(f,'\textbf{');
end;

procedure de(var f: text);
(* do bold end to file f *)
begin
  if dobold then
  write(f,'}');
end;

procedure dm(var f: text);
(* do bold math to file f *)
begin
  if dobold then
  write(f,'\mathbf{');
end;
procedure setupheaders;
(* set up the headers *)
var
   position: integer; (* index to position *)
begin

   writeln(labelatex,'% ');
   writeln(labelatex,'% Ri(b,l) TABLE: -------------------------');

   (* copy the header *)
   reset(ribl);
   write(labelatex,'% ');
   copyaline(ribl,output); (* get the name *)
   while ribl^='*' do begin
      write(labelatex,'% ');
      copyaline(ribl,labelatex);
   end;

   if askedfrom < matrix^.frombase then begin
      writeln(output,'requested from, ',
                     askedfrom:1,
                     ', is out of range of the ribl',
                     matrix^.frombase:1,
                     ' to ',
                     matrix^.tobase:1);
      halt
   end;

   if askedto > matrix^.tobase then begin
      writeln(output,'requested to, ',
                     askedto:1,
                     ', is out of range of the ribl',
                     matrix^.frombase:1,
                     ' to ',
                     matrix^.tobase:1);
      halt
   end;

   writeln(labelatex,'% ----------------------------------------');
   writeln(labelatex,'% ');

   writeln(labelatex,'% from ',askedfrom:1,' to ',askedto:1);

(*
   writeln(labelatex,'base $b$ & \multicolumn{',
                      length:1,'}{|c|}{position $l$} \\');
*)
   db(labelatex); write(labelatex,'base '); de(labelatex);
   write(labelatex,'$');
   dm(labelatex); write(labelatex,'b'); de(labelatex);
   write(labelatex,'$');
   write(labelatex,' & \multicolumn{', length:1,'}{|c|}{');
   db(labelatex); write(labelatex,'position '); de(labelatex);
   write(labelatex,'$');
   dm(labelatex); write(labelatex,'l'); de(labelatex);
   writeln(labelatex, '$} \\');

   write(headlatex,'\begin{tabular}{|c|');
   for position := 1 to length do write(headlatex,'r');
   writeln(headlatex,'|}');

   for position := 1 to length do begin
      write(sequlatex,' & ');
      db(sequlatex);
      if caps
      then write(sequlatex,   capitalize(sequence[position]))
      else write(sequlatex, decapitalize(sequence[position]));
      de(sequlatex);
   end;
   writeln(sequlatex,' \\');

end;

begin

   startfiles;
   readparameters;
   wraplimit := 100;
   if hline then writeln(labelatex,'\hline');

   reset(ribl);
   getriblmatrix(ribl, matrix);

   setupheaders;
(*
   This command doesn't work for some reason
   write(ribllatex,'\newcommand{\fakespace}{\hspace{\fboxsep}}');
*)
   (* Using this instead of fbox will align the content with fboxes!! *)

   (* transpose and write out data *)
   (* do table header *)
   write(freqlatex,' ');
   write(ribllatex,' ');
   write(senulatex,' ');
   count := 0;

   for l := askedfrom to askedto do begin 

      write(freqlatex,' & ','$ ');
      dm(freqlatex);
      write(freqlatex,l:wid);
      de(freqlatex);
      write(freqlatex,'$');
      write(freqlatex,'\hspace{\fboxsep}'); (* align with fboxes!! *)

      write(ribllatex,' & ','$ ');
      dm(ribllatex);
      write(ribllatex,l:wid);
      de(ribllatex);
      write(ribllatex,'$');
(*
      write(ribllatex,'\hspace{\fboxsep}'); (# align with fboxes!! #)
      write(ribllatex,'\fakespace'); (# align with fboxes!! #)
This fails for some reason, so don't use the command.
*)
      write(ribllatex,'\hspace{\fboxsep}'); (* align with fboxes!! *)

      write(senulatex,' & ','$ ');
      dm(senulatex);
      write(senulatex,l:wid);
      de(senulatex);
      write(senulatex,'$');

      count := count + 1;
   end;
   writeln(freqlatex,' \\');
   writeln(ribllatex,' \\');
   writeln(senulatex,' \\');

   (* now do tables *)
   closebox := false;
   if hline then writeln(senulatex,'\hline');
   if hline then writeln(freqlatex,'\hline');
   if hline then writeln(ribllatex,'\hline');
   for column := 1 to 4 do begin
     case column of
       1: thebase := 'a';
       2: thebase := 'c';
       3: thebase := 'g';
       4: thebase := 't';
     end;

     case column of
       1: b := a;
       2: b := c;
       3: b := g;
       4: b := t;
     end;

     if caps then thebase := capitalize(thebase)
             else thebase := decapitalize(thebase);
     db(senulatex);
     writeln(senulatex,thebase);
     de(senulatex);
     db(senulatex);
     writeln(freqlatex,thebase);
     writeln(ribllatex,thebase);
     de(senulatex);
     count := 0;

     for l := askedfrom to askedto do begin 
         count := count + 1;
         write(freqlatex,' & ');
         write(ribllatex,' & ');
         write(senulatex,' & ');

         if (count <= length) and (sequence[count]=decapitalize(thebase))
            and dobox
         then begin
           write(freqlatex,'\fbox{');
           write(ribllatex,'\fbox{');
           closebox := true;
         end
         else begin
           write(freqlatex,'\mbox{');
           write(ribllatex,'\mbox{');
           closebox := false;
         end;

         write(freqlatex,'$');
         dm(freqlatex);
         sum := 0;
         for bsum := a to t
         do sum := sum + matrix^.numbers[bsum,l];
         write(freqlatex,(matrix^.numbers[b,l]/sum):wid:decimals);
         (* NB: this gets done 3 extra times and so is wasteful 
            but so what??  The matrix it tiny! *)
         de(freqlatex);
         write(freqlatex,'$');

         write(ribllatex,'$');
         dm(ribllatex);
         if matrix^.data[b,l] <= negativeinfinity
         then write(ribllatex,'-\infty')
         else if matrix^.data[b,l] < 0
              then write(ribllatex,matrix^.data[b,l]:wid:decimals)
              else write(ribllatex,'+',matrix^.data[b,l]:wid:decimals);
         de(ribllatex);
         write(ribllatex,'$');

         if closebox
         then write(freqlatex,'}')
         else write(freqlatex,'\hspace{\fboxsep}}'); (* align mboxes!! *)

         if closebox
         then write(ribllatex,'}')
         else write(ribllatex,'\hspace{\fboxsep}}'); (* align mboxes!! *)

         write(senulatex,'$');
         dm(senulatex);
         if (count <= length) and (sequence[count]=decapitalize(thebase))
         then write(senulatex,'1')
         else write(senulatex,'0');
         de(senulatex);
         write(senulatex,'$');

     end;
     writeln(freqlatex,' \\');
     writeln(ribllatex,' \\');
     writeln(senulatex,' \\');
   end;
   if hline then writeln(senulatex,'\hline');
   if hline then writeln(freqlatex,'\hline');
   if hline then writeln(ribllatex,'\hline');
   if hline then writeln(labelatex,'\hline');

   writeln(taillatex,'\end{tabular}');
end;
(* end module rila.themain *)

begin
   themain(ribl, rilap, headlatex, labelatex, sequlatex, senulatex,
           freqlatex, ribllatex, taillatex);
1: end.
