program spec(csdata, baseline, xyin, output);
(* spec: analyse two spectra from the camspec

  Tom Schneider
  network address: toms@alum.mit.edu


  1992
 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.10; (* of spec.p 1992 June 15
origin 1992 May 15 *)
(* end module version *)

(* begin module describe.spec *)
(*
name
   spec: analyse two spectra from the camspec

synopsis
   spec(csdata: in, baseline: in, xyin: out, output: out)

files
   csdata:  contains one spectrum from the Camspec to be used
     as the data.
   baseline:  contains one spectrum from the Camspec to be used
     as the baseline.
   xyin: input to xyplo program
   output: messages to the user

description
   Analysis of spectra produced by the camspec.

   Setup:  Establish communications with the Camspec.

   Give the T command.  If it does
   not respond with '2', use the A command and repeat T until it does.
   This sets the mode to transmittance.

   Then set to 450 nm with           450G
   Check that it is set with         U
   Set to 100% transmittance with    B
   Check that it is 100% with        V
   You can use the O command to do both checks at once.

   Use T and A to set the mode to 0, for reporting in absorbance.
   'AT' should do it.

   Obtain the spectrum with
      xl h
   where x is the interval:
      L 5nm
      M 1nm
      N 0.5nm
   and l is the lower wavelength, h is the high.  Eg,
      L400 700
   scans from 400 to 700 nm with steps of 5nm.

examples

documentation

see also
   xyplo.p

author
   Thomas Dana Schneider

bugs

technical notes
   The spectrum is set to -maxint at both ends so that when multiple
   xyin spectra are concatenated, the return lines all run below the graph
   (where you won't see them with xyplo).  Xyplo will object, but ignore it.

   The camspec sends absorbance data multiplied by 100.  This number is
   in constant 'correctionfactor'.

*)
(* end module describe.spec *)

(* begin module const.spec *)
   correctionfactor = 100.00; (* factor by which to divide the data
      from the camspec to get absorbtion *)
   cswid = 10; (* width of data numbers put to xyin *)
   csdec = 5; (* decimals of data numbers put to xyin *)
(* end module const.spec *)

var
   csdata, baseline, xyin: text; (* a file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module spec.skip *)
procedure skip(var f: text; n: integer);
(* skip n characters in file f *)
var i: integer; (* counts the skips *)
begin
   for i := 1 to n do get(f)
(*
   for i := 1 to n do begin
      write(output,'(',f^,')'); get(f)
   end;
*)
end;
(* end module spec.skip *)

(* begin module spec.tostart *)
procedure tostart(var f: text;
                  var interval: real;
                  var low: integer;
                  var high: integer);
(* move file f to the start of the spectrum and obtain the
interval, low and high nm of the spectrum *)
var
   done: boolean; (* have we found the start? *)
begin
   reset(f);
   done := false;
   while not done do begin
      if eof(f) then begin
         writeln(output,'no spectrum found!');
         halt
      end;
      if not eoln(f) then begin
         if f^='S' then begin
            get(f);
            if f^='c' then begin
               get(f);
               if f^='a' then begin
                  get(f);
                  if f^='n' then begin
{
writeln(output,'found Scan');
while 1 = 1 do begin write(output,f^); get(f); end;
}
                     skip(f,4); (* ' at ' *)
                     read(f,interval);
                     skip(f,14); (* 'nm intervals' *)
                     read(f,low,high);
                     done := true
                  end
               end
            end
         end
      end;
      readln(f)
   end;

   readln(f);
end;
(* end module spec.tostart *)

(* begin module spec.showdata *)
procedure showdata(var f: text;
                  var interval: real;
                  var low: integer;
                  var high: integer);
(* show the data to file f: interval, low and high nm of the spectrum *)
begin
   writeln(f,'* ',interval:3:1,' interval');
   writeln(f,'* ',low:3,' low nm');
   writeln(f,'* ',high:3,' high nm');
   writeln(f,'* ');
end;
(* end module spec.showdata *)

(* begin module spec.readdata *)
procedure readdata(var f: text; var nm, data: real);
(* read the wavelength (nm) and absorbance (data) from f *)
var
   c: char; (* a character used to skip the comma between the data *)
begin
   readln(f,nm,c,data);
   if c <> ',' then begin
      writeln(output,'Misread: "',c,'" should have been a comma.');
      halt
   end;
   data := data/correctionfactor
end;
(* end module spec.readdata *)

(* begin module spec.themain *)
procedure themain(var csdata, baseline, xyin: text);
(* the main procedure of the program *)
var
   bdata: real; (* the data for baseline *)
   bhigh: integer; (* baseline high nm *)
   binterval: real; (* baseline interval *)
   blow: integer; (* baseline low nm *)
   bnm: real; (* the nm for baseline *)

   ddata: real; (* the data for data *)
   dhigh: integer; (* data high nm *)
   dinterval: real; (* data interval *)
   dlow: integer; (* data low nm *)
   dnm: real; (* the nm for data *)

   done: boolean; (* when done reading the data *)

   zerobaseline: boolean; (* if true, then set the baseline to zero
      because there is no baseline file *)
begin
   writeln(output,'spec ',version:4:2);

   reset(csdata);
   reset(baseline);
   rewrite(xyin);

   tostart(csdata, dinterval, dlow, dhigh);

   if eof(baseline)
   then begin
      zerobaseline := true;
      binterval := dinterval;
      blow := dlow;
      bhigh := dhigh;
   end
   else begin
      zerobaseline := false;
      tostart(baseline, binterval, blow, bhigh);
   end;

   if (binterval <> dinterval) or
      (blow <> dlow) or
      (bhigh <> dhigh)
   then begin
      writeln(output,'intervals and range of spectrum must match');
      halt
   end;

   writeln(xyin,'* spec ',version:4:2);
   if zerobaseline then begin
      writeln(xyin,'* baseline: set to zero');
   end
   else begin
      writeln(xyin,'* baseline:');
      showdata(xyin, binterval, blow, bhigh);
   end;
   writeln(xyin,'* csdata:');
   showdata(xyin, dinterval, dlow, dhigh);

   writeln(xyin,'*');
   writeln(xyin,'* columns: nm, data-baseline, baseline, data');
   writeln(xyin,'*');

   (* put out zero point: *)
   writeln(xyin,blow:5,' ',(-maxint):cswid,' ',maxint:cswid,' ',0:cswid);

   done := false;
   while not done do begin

      if eof(baseline) and not zerobaseline then begin
         write(output,'unexpected end of baseline file');
         halt
      end;

      if eof(csdata) then begin
         write(output,'unexpected end of csdata file');
         halt
      end;

      if (baseline^='S') and not zerobaseline then begin
         if baseline^<>'S' then begin
            writeln(output,'baseline ended early');
            halt
         end;
         done := true
      end;

      if csdata^='S' then begin
         if csdata^<>'S' then begin
            writeln(output,'csdata file ended early');
            halt
         end;
         done := true
      end

      else begin (* read the data *)
         readdata(csdata,dnm,ddata);
         if zerobaseline
         then begin
            bnm := dnm;
            bdata := 0.0;
         end
         else readdata(baseline,bnm,bdata);

         if bnm <> dnm then begin
            writeln(output,'misread: ',
                           ' baseline is at ',bnm:5:1,
                           ' but ',
                           ' data is at ',dnm:5:1);
            halt
         end;
         writeln(xyin, dnm:5:1,
                       ' ',(ddata-bdata):cswid:csdec,
                       ' ',bdata:cswid:csdec,
                       ' ',ddata:cswid:csdec)
      end
   end;

   (* put out zero point for redraw: *)
   writeln(xyin,bhigh:5,' ',(-maxint):cswid,' ',maxint:cswid,' ',0:cswid);
end;
(* end module spec.themain *)

begin
   themain(csdata, baseline, xyin);
1: end.
