program subsymvec(symvecin, subsymvecp, symvecout, output);
(* subsymvec: extract a subset of a symvec file

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.01; (* of subsymvec.p 2000 Dec 4
2000 Dec 4, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.subsymvec *)
(*
name
   subsymvec: extract a subset of a symvec file

synopsis
   subsymvec(symvecin: in, subsymvecp: in, symvecout: out, output: out)

files

   symvecin:  a symbolic vector, symvec, eg. from alpro or dalvec.

   symvecout:  a symbolic vector, symvec

   subsymvecp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      range:  two integers that define the range of the symvec to use.

   output: messages to the user

description

   Extract a subset of a symvec.

examples

documentation

see also

   {sources of symvec files:} alpro.p dalvec.p

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.subsymvec *)

(* The next module is taken from module dalvec.const
to get the infofield, infodecim and nfield values.  The other
parameters don't matter (yet).  *)
(* begin module dalvec.const *)
   infofield = 8; (* size of field for printing information in bits *)
   infodecim = 5; (* number of decimal places for printing information *)

   nfield = 4; (* size of field for printing n, the number of sites *)
   mapmax = 26; (* the number of bases or amino acids to be sorted *)

   pwid = 8; (* width in character places to print PostScript numbers *)
   pdec = 5; (* decimal places to print PostScript numbers *)
   pnum = 100000; (* 10^pdec for testing when something will round to zero *)

   cxmove = 30; (* units to move a character horizontally *)
(* end module dalvec.const *)

var
   symvecin, (* file used by this program *)
   subsymvecp, (* file used by this program *)
   symvecout: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline from prgmod 4.54 *)

(* begin module subsymvec.themain *)
procedure themain(var symvecin, subsymvecp, symvecout: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   s: integer; (* number of symbols in the alphabet *)
   sindex: integer; (* index for s *)

   fromrange: integer; (* lower bound of symvec to use *)
     torange: integer; (* upper bound of symvec to use *)

   r: real; (* information *)
   v: real; (* information variance *)
   n: integer; (* number of sequences at this position *)
   l: integer; (* number of symbols in the alphabet *)

   startup: boolean; (* have we got past the header? *)
{zzz}
begin
   writeln(output,'subsymvec ',version:4:2);
   reset(subsymvecp);
   readln(subsymvecp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
   readln(subsymvecp, fromrange, torange);
   write  (output,'* extraction from: ', fromrange:1);
   writeln(output,            ' to: ',   torange:1);

   reset(symvecin);
   rewrite(symvecout);
   writeln(symvecout,'* subsymvec ',version:4:2);
   write  (symvecout,'* extraction from: ', fromrange:1);
   writeln(symvecout,            ' to: ',   torange:1);
   startup := true;

   while not eof(symvecin) do begin
      if symvecin^ = '*'
      then copyaline(symvecin, symvecout)
      else begin
         if startup then begin
            readln(symvecin, s);
            writeln(output    ,s:1, ' number of symbols');
            writeln(symvecout ,s:1, ' number of symbols');
            startup := false
         end
         else begin
{
writeln(output,'processing');
}
            readln(symvecin, l, n, r, v);

{
               writeln(symvecout,      l: nfield,
                                  ' ', n: nfield,
                                  ' ', r: infofield:infodecim,
                                  ' ', v: (infofield+3));
}

            if (l >= fromrange) and (l <= torange) then begin
               (* copy the position *)

               writeln(symvecout,      l: nfield,
                                  ' ', n: infofield,
                                  ' ', r: infofield:infodecim,
                                  ' ', v: (infofield+3));

               for sindex := 1 to s
               do copyaline(symvecin, symvecout)
            end
            else begin (* skip the position *)
               for sindex := 1 to s
               do readln(symvecin)
            end;
         end
      end;

   end;

end;
(* end module subsymvec.themain *)

begin
   themain(symvecin, subsymvecp, symvecout);
1: end.
