program truncround(afile, truncroundp, xyin, output);
(* truncround: plot trunc and round functions

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu (use only if first address fails)
  https://alum.mit.edu/www/toms/

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.01; (* of truncround.p 2010 Feb 06
2010 Feb 06, 1.01: adjust z function to be the right thing for genhis
2010 Feb 05, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.truncround *)
(*
name
   truncround: plot trunc and round functions

synopsis
   truncround(afile: in, truncroundp: in, xyin: out, output: out)

files

   afile:  not used

   xyin:  x and y columns of results

   truncroundp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      xmin: minimum x value
      xmax: maximum x value
      yrd:  displacement for: y is the round of x
      ytd:  displacement for: y is the trunc of x
      yzd:  displacement for: x shift of round
      zxd:  x shift of round

   output: messages to the user

description

   Test trunc vs round functions.

   For xyin create three data sets:

     yt := trunc(x) + ytd; { trunc displaced vertically }
     yr := round(x) + yrd; { round displaced vertically }
     yz := round(x+zxd) - 1.0 + yzd; { round shifted horiz. and vertically }

   The program is used for testing the binning function of genhis.

examples

   truncroundp:

1.00  version of truncround that this parameter file is designed for.
-5.0  xmin
+5.0  xmax
-3.0  yrd
+3.0  ytd
 0.0  yzd
+0.5  zxd

   xyplop:

3 6       zerox zeroy         graph coordinate center
x -8 8     zx min max          (character, real, real) if zx='x' then set xaxis
y -8 8     zy min max          (character, real, real) if zy='y' then set yaxis
8 8   2 2 xinterval yinterval xsubintervals ysubintervals: axis intervals
8 8       xwidth    ywidth    width of numbers in characters
1 1       xdecimal  ydecimal  number of decimal places
15 15     xsize     ysize     size of axes in cm
x
y = trunc(x)+3; y = round(x)-3; y = round(x+0.5)-1;
c         zc                  'c' crosshairs, axXyYnN
n 2       zxl base            if zxl='l' then make x axis log to the given base
n 2       zyl base            if zyl='l' then make y axis log to the given base
          *********************************************************************
1 2       xcolumn   ycolumn   columns of xyin that determine plot location
0         symbol column       the xyin column to read symbols from
0  0      xscolumn  yscolumn  columns of xyin that determine the symbol size
0 0 0     hue saturation brightness   columns for color manipulation
          *********************************************************************
c         symbol-to-plot      c(circle)bd(dotted box)x+Ifgpr(rectangle)
0         symbol-flag         character in xyin that indicates that this symbol
0.15      symbol sizex        side in cm on the x axis of the symbol.
0.15      symbol sizey        as for the x axis, get size from yscolumn
nl 0.125  no connection (example for connection is c- 0.125 for dashed 0.125 cm)
n  0.125  linetype  size      linetype l.-in and size of dashes or dots
          *********************************************************************
.
          **** version 8.50 of xyplop: uses cm for distances
          **** more parameters
p 1.5 0.5 1.5 1.0 edgecontrol (p=page),edgeleft,edgeright,edgelow,edgehigh (cm)
9.00      version of xyplo that this parameter file is designed for.

documentation

see also

   genhis.p

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.truncround *)

var
   afile, (* file used by this program *)
   truncroundp, (* file used by this program *)
   xyin: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module truncround.themain *)
procedure themain(var afile, truncroundp, xyin: text);
(* the main procedure of the program *)
const
   wid = 5; (* width    of numbers in xyin *)
   dec = 2; (* decimals of numbers in xyin *)
var
   parameterversion: real; (* parameter version number *)
   x: real; (* x, independent variable *)
   xmin, xmax: real; (* range of x *)
   yr: real; (* round of x *)
   yrd: real; (* displacement of yr *)
   yt: real; (* trunc of x *)
   ytd: real; (* displacement of yt *)
   yz: real; (* trunc of x *)
   yzd: real; (* displacement of yz *)
   zxd: real; (* displacement of x for the z graph *)
begin
   writeln(output,'truncround ',version:4:2);
   reset(truncroundp);
   readln(truncroundp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   rewrite(xyin);
   writeln(xyin,'* truncround ',version:4:2);

   readln(truncroundp, xmin); { := -5.0;}
   readln(truncroundp, xmax); { := +5.0;}
   readln(truncroundp, yrd);  { := -3.0;}
   readln(truncroundp, ytd);  { := +3.0;}
   readln(truncroundp, yzd);  { := +0.0;}
   readln(truncroundp, zxd);  { := +0.5;}

{
   xmin := -5.0;
   xmax := +5.0;
   yrd  := -3.0;
   ytd  := +3.0;
   yzd  := +0.0;
   zxd  := +0.5;
}

   x := xmin;
   while x < xmax do begin

       yt := trunc(x) + ytd; { trunc displaced vertically }
       yr := round(x) + yrd; { round displaced vertically }
       yz := round(x+zxd) - 1.0 + yzd; { round shifted horiz. and vertically }

       writeln(xyin, x:wid:dec, ' ', yt:wid:dec, ' t');
       writeln(xyin, x:wid:dec, ' ', yr:wid:dec, ' r');
       writeln(xyin, x:wid:dec, ' ', yz:wid:dec, ' z');
       x := x + 0.1;
   end;

end;
(* end module truncround.themain *)

begin
   themain(afile, truncroundp, xyin);
1: end.
